
#include "ACM_Reader.h"
using namespace Juna::Sound;

#include <mmreg.h>
#include <msacm.h>


#pragma comment(lib, "msacm32.lib")


ACM_FileReader::ACM_FileReader(void)
  : File(),
	AverageByte(0),BlockAlign(0),
	hAcmStream(NULL),SrcBuffer(NULL),DestBuffer(NULL),
	SrcBufferSize(0),DestBufferSize(0),
	SrcBufferRemain(0),DestBufferRemain(0),DestRemainStart(0),
	SrcReadEnd(0),ReadPos(0),SrcFormat(NULL),
	SoundFormat(),Error(false),RefCounter(1)
{
	::memset(&ACMStreamHdr,0,sizeof(ACMSTREAMHEADER));
	ACMStreamHdr.cbStruct = sizeof(ACMSTREAMHEADER);
}

ACM_FileReader *ACM_FileReader::Create(void)
{
	return new ACM_FileReader();
}

void ACM_FileReader::Release(void)
{
	RefCounter--;
	if (RefCounter <= 0)
		delete this;
}



namespace {
class auto_ptr
{
	void *ptr;
public:
	explicit auto_ptr(void *p = 0) : ptr(p) {}
	~auto_ptr() {operator delete(ptr);}
	void *get(void) {return ptr;}
	void release(void) {ptr = 0;}
private:
	auto_ptr(auto_ptr &src);
	auto_ptr &operator=(const auto_ptr &rhs);
};

bool Test_ACM_Wave(const char *filename,DWORD offset,
				   unsigned long &doffset,unsigned long &dsize,unsigned long &factsize,
				   WAVEFORMATEX **wfex,bool formatSuggest)
{
	Juna::IO::InputFile file(filename);
	if (!file.isOpen())
		return false;
	if (!file.Seek(offset))
		return false;
	DWORD head[3] = {0};

	if (file.Read(head,sizeof(DWORD) * 3) < sizeof(DWORD) * 3)
		return false;
//t@Cwb_`FbN
	if (head[0] != 'R'+('I'<<8)+('F'<<16)+('F'<<24) ||
		head[2] != 'W'+('A'<<8)+('V'<<16)+('E'<<24))
		return false;

	if (file.Read(head,sizeof(DWORD) * 2) < sizeof(DWORD) * 2)
		return false;
	while (head[0] != 'f'+('m'<<8)+('t'<<16)+(' '<<24))
	{
		if (!file.Seek(head[1] + file.Tell()))
			return false;
		if (file.Read(head,sizeof(DWORD) * 2) < sizeof(DWORD) * 2)
			return false;
	}


//fmt^CvɑΉ
	int size;
	if (head[1] == sizeof(PCMWAVEFORMAT))
		size = sizeof(WAVEFORMATEX);
	else
		size = head[1];
	auto_ptr ptr(operator new(size));
	if (!ptr.get())
		return false;
	WAVEFORMATEX *wfx = static_cast<WAVEFORMATEX *>(ptr.get());

	if (file.Read(wfx,head[1]) < head[1])
		return false;
	if (head[1] == sizeof(PCMWAVEFORMAT))
		wfx->cbSize = 0;
	else if (head[1] != sizeof(WAVEFORMATEX) + wfx->cbSize)
		return false;

	if (file.Read(head,sizeof(DWORD) * 2) < sizeof(DWORD) * 2)
		return false;
	while (head[0] != 'd'+('a'<<8)+('t'<<16)+('a'<<24))
	{
//fact`NT[`
		if (factsize && head[0] == 'f'+('a'<<8)+('c'<<16)+('t'<<24))
		{
			if (file.Read(&factsize,head[1]) < head[1])
				return false;
		}
		else
		{
			if (!file.Seek(head[1] + file.Tell()))
				return false;
		}
		if (file.Read(head,sizeof(DWORD) * 2) < sizeof(DWORD) * 2)
			return false;
	}
	doffset = file.Tell();
	dsize = head[1];

	if (formatSuggest)
	{
		WAVEFORMATEX destwfx = {0};
		destwfx.wFormatTag = WAVE_FORMAT_PCM;
		if (::acmFormatSuggest(NULL,wfx,&destwfx,sizeof(destwfx),
							   ACM_FORMATSUGGESTF_WFORMATTAG))
			return false;
	}
	if (wfex)
	{
		*wfex = wfx;
		ptr.release();
	}
	return true;
}
}//namespace ""


bool ACM_FileReader::OpenWaveFile
	(const char *filename,DWORD offset,DWORD bufsize,bool fact)
{
	Close();
	WAVEFORMATEX *wfx;
	unsigned long doffset,dsize,factsize;
	if (!Test_ACM_Wave(filename,offset,doffset,dsize,factsize,&wfx,false))
		return false;
	auto_ptr ptr(wfx);
	return Open_data(filename,doffset,dsize,wfx,bufsize,fact ? factsize : 0);
}


bool ACM_FileReader::Open_data
	(const char *filename,unsigned long offset,unsigned long size,
	 WAVEFORMATEX *srcwfx,long srcbufsize,DWORD factsize)
{
	Close();
	if (!File.Open(filename,offset,size))
		return false;

	SrcFormat = static_cast<WAVEFORMATEX *>(operator new(sizeof(WAVEFORMATEX) + srcwfx->cbSize));
	::memcpy(SrcFormat,srcwfx,sizeof(WAVEFORMATEX) + srcwfx->cbSize);
	WAVEFORMATEX destwfx;
	destwfx.cbSize = 0;
	destwfx.wFormatTag = WAVE_FORMAT_PCM;
	if (::acmFormatSuggest(NULL,srcwfx,&destwfx,sizeof(destwfx),ACM_FORMATSUGGESTF_WFORMATTAG))
		goto ERR;

	if (::acmStreamOpen(&hAcmStream,NULL,srcwfx,&destwfx,NULL,0,0,0))
		goto ERR;

	if (::acmStreamSize(hAcmStream,srcbufsize,&DestBufferSize,ACM_STREAMSIZEF_SOURCE))
		goto ERR;
	if (::acmStreamSize(hAcmStream,DestBufferSize,&SrcBufferSize,ACM_STREAMSIZEF_DESTINATION))
		goto ERR;
	SrcBuffer = static_cast<BYTE *>(operator new(SrcBufferSize));
	DestBuffer = static_cast<BYTE *>(operator new(DestBufferSize));

	SrcBufferRemain = 0;
	DestBufferRemain = 0;
	DestRemainStart = 0;

	ACMStreamHdr.pbSrc = SrcBuffer;
	ACMStreamHdr.cbSrcLength = SrcBufferSize;
	ACMStreamHdr.pbDst = DestBuffer;
	ACMStreamHdr.cbDstLength = DestBufferSize;
	if (::acmStreamPrepareHeader(hAcmStream,&ACMStreamHdr,0))
		goto ERR;


	SoundFormat.set(destwfx.nSamplesPerSec,destwfx.wBitsPerSample,
					(unsigned char)destwfx.nChannels);
	SrcReadEnd = 0;
	ReadPos = 0;
	AverageByte = srcwfx->nAvgBytesPerSec;
	BlockAlign = srcwfx->nBlockAlign;

	if (factsize)
	{
		Length = factsize;
	}
	else if (AverageByte % srcwfx->nSamplesPerSec == 0)	//vɒʏWavt@Cz
	{
		Length = size / (AverageByte / srcwfx->nSamplesPerSec);
	}
	else if (AverageByte < 0xFFFF && srcwfx->nSamplesPerSec < 0xFFFF)	//32bitɎ܂
	{
		Length = size / AverageByte * srcwfx->nSamplesPerSec;
		Length += (size % AverageByte) * srcwfx->nSamplesPerSec / AverageByte;
	}
	else	//r[ȈkȂABɂȂ邯
	{
		Length = size / AverageByte * srcwfx->nSamplesPerSec;
		Length += ((size % AverageByte) * 1024 / AverageByte) * srcwfx->nSamplesPerSec / 1024;
	}

	return true;

ERR:;
	Close();
	return false;
}


void ACM_FileReader::Close(void)
{
	if (ACMStreamHdr.fdwStatus & ACMSTREAMHEADER_STATUSF_PREPARED) {
		ACMStreamHdr.cbSrcLength = SrcBufferSize;
		::acmStreamUnprepareHeader(hAcmStream,&ACMStreamHdr,0);
	}
	if (hAcmStream) {
		::acmStreamClose(hAcmStream,0);
		hAcmStream = NULL;
	}
	if (SrcBuffer) {
		operator delete(SrcBuffer);
		SrcBuffer = NULL;
	}
	if (DestBuffer) {
		operator delete(DestBuffer);
		DestBuffer = NULL;
	}
	if (SrcFormat) {
		operator delete(SrcFormat);
		SrcFormat = NULL;
	}
	File.Close();

	Length = AverageByte = BlockAlign = 0;
	SrcBufferSize = DestBufferSize = 0;
	SrcBufferRemain = DestBufferRemain = DestRemainStart = 0;
	SrcReadEnd = ReadPos = 0;
	SoundFormat.reset();
	Error = false;
}


unsigned long ACM_FileReader::Read(void *dest,unsigned long size)
{
	DWORD readbyte = 0;
	do {
		if (DestBufferRemain > 0)
		{
			if (readbyte + DestBufferRemain < size)
			{
				::memcpy((BYTE *)dest + readbyte,
					   DestBuffer + DestRemainStart,DestBufferRemain);
				readbyte += DestBufferRemain;
				DestBufferRemain = 0;
				if (SrcReadEnd && SrcBufferRemain == 0)
				{
					DestRemainStart = 0;	//Ô
					break;
				}
			}
			else
			{
				DWORD lastread = size - readbyte;
				::memcpy((BYTE *)dest + readbyte,DestBuffer + DestRemainStart,lastread);
				readbyte += lastread;
				DestBufferRemain -= lastread;
				DestRemainStart += lastread;
				break;
			}
		}
		DestRemainStart = 0;
		DWORD ascflag = 0;
		if (SrcReadEnd == 0)
		{
			unsigned long read;
			read = File.Read(SrcBuffer + SrcBufferRemain,SrcBufferSize - SrcBufferRemain);
			if (read < SrcBufferSize - SrcBufferRemain)
			{
				if (read == 0 && !File.isEnd())
				{
					Error = true;
					break;
				}
				SrcReadEnd = 1;
			}
			SrcBufferRemain += read;
			ascflag = ACM_STREAMCONVERTF_BLOCKALIGN;
		}

//		if (SrcBufferRemain == 0)
//		{
//		}
		ACMStreamHdr.cbSrcLength = SrcBufferRemain;
		if (::acmStreamConvert(hAcmStream,&ACMStreamHdr,ascflag))
		{
			Error = true;
			break;
		}

//̓obt@ŎgĂȂΐ擪܂ňړ
		::memmove(SrcBuffer + ACMStreamHdr.cbSrcLengthUsed,SrcBuffer,SrcBufferRemain - ACMStreamHdr.cbSrcLengthUsed);
		SrcBufferRemain -= ACMStreamHdr.cbSrcLengthUsed;

		DestBufferRemain = ACMStreamHdr.cbDstLengthUsed;
	} while (DestBufferRemain > 0);

	ReadPos += readbyte / SoundFormat.sample_size();
	return readbyte;
}

unsigned long ACM_FileReader::Seek(unsigned long samples)
{
	if (samples > Length)
		samples = Length;
	long pos; 
	const long freq = SoundFormat.frequency;
	if (AverageByte % freq == 0)	//vɒʏWavt@Cz
	{
		pos = samples * (AverageByte / freq);
	}
	else if (AverageByte < 0xFFFF && freq < 0xFFFF)	//32bitɎ܂
	{
		pos = samples / freq * AverageByte;
		pos += (samples % freq) * AverageByte / freq;
	}
	else	//r[ȈkȂABɂȂ邯ǁ@قƂsamples̑傫𒲂ׂׂ
	{
		pos = samples / freq * AverageByte;
		pos += ((samples % freq) * 1024 / freq) * AverageByte / 1024;
	}

	pos -= pos % BlockAlign;
	if (!File.Seek(pos))
	{
		Error = true;
		return (unsigned long)-1;
	}

	if (AverageByte % freq == 0)	//vɒʏWavt@Cz
	{
		ReadPos = pos / (AverageByte / freq);
	}
	else if (AverageByte < 0xFFFF && freq < 0xFFFF)	//32bitɎ܂
	{
		ReadPos = pos / AverageByte * freq;
		ReadPos += (pos % AverageByte) * freq / AverageByte;
	}
	else	//r[ȈkȂABɂȂ邯ǁ@قƂAverageBytȇ傫𒲂ׂׂ
	{
		ReadPos = pos / AverageByte * freq;
		ReadPos += ((pos % AverageByte) * 1024 / AverageByte) * freq / 1024;
	}

//	acmStreamReset(hAcmStream,0);	//Zbgˁ[
	::acmStreamUnprepareHeader(hAcmStream,&ACMStreamHdr,0);
	::acmStreamClose(hAcmStream,0);
	WAVEFORMATEX destwfx;
	destwfx.cbSize = 0;
	destwfx.wFormatTag = WAVE_FORMAT_PCM;
	::acmFormatSuggest(NULL,SrcFormat,&destwfx,sizeof(destwfx),ACM_FORMATSUGGESTF_WFORMATTAG);
	::acmStreamOpen(&hAcmStream,NULL,SrcFormat,&destwfx,NULL,0,0,0);
	::acmStreamPrepareHeader(hAcmStream,&ACMStreamHdr,0);
//hCoĂobt@ZbgAƃ}Vȕ@͂Ȃ̂H

	{
//		Error = true;
//		return (unsigned long)-1;
	}

	SrcBufferRemain = 0;
	DestBufferRemain = 0;
	SrcReadEnd = 0;
	return ReadPos;
}

