
#ifndef Angle_H
#define Angle_H

#include <math.h>

static const float Pi = 3.1415926535897932384626433832795f;
static const float Pi2 = Pi * 2;

class RadianAngle
{
public:
	RadianAngle(void) {}
	RadianAngle(float radian) : Radian(radian) {}

	float radian(void) const {return Radian;}
	float degree(void) const {return Radian * (360 / Pi2);}

	float sin(void) const {return ::sin(Radian);}
	float cos(void) const {return ::cos(Radian);}
	float tan(void) const {return ::tan(Radian);}

	RadianAngle &SetAtan(float x) {Radian = ::atan(x);return *this;}
	RadianAngle &SetAtan(float y,float x) {Radian = ::atan2(y,x);return *this;}

	RadianAngle &SetNormal(void)
	{
		if (Radian > Pi2)
			Radian = ::fmod(Radian,Pi2);
		else if (Radian < 0)
			Radian = ::fmod(Radian,Pi2) + Pi2;
		return *this;
	}
	RadianAngle &SetNormal2(void)
	{
		Radian += Pi;
		SetNormal();
		Radian -= Pi;
		return *this;
	}
	RadianAngle Normalize(void) const
	{
		if (Radian > Pi2)
			return RadianAngle(::fmod(Radian,Pi2));
		else if (Radian < 0)
			return RadianAngle(::fmod(Radian,Pi2) + Pi2);
		return *this;
	}
	RadianAngle Normalize2(void) const
	{
		RadianAngle r(Radian + Pi);
		r.SetNormal();
		return r - Pi;
	}
	RadianAngle &SetNegative(void) {Radian = -Radian;return *this;}

	RadianAngle &operator+=(RadianAngle rhs) {Radian += rhs.Radian;return *this;}
	RadianAngle &operator-=(RadianAngle rhs) {Radian -= rhs.Radian;return *this;}
	RadianAngle operator+(RadianAngle rhs) const {return RadianAngle(Radian + rhs.Radian);}
	RadianAngle operator-(RadianAngle rhs) const {return RadianAngle(Radian - rhs.Radian);}

	RadianAngle operator-(void) const {return RadianAngle(-Radian);}
	RadianAngle operator+(void) const {return Normalize();}

	RadianAngle &operator*=(float rhs) {Radian *= rhs;return *this;}
	RadianAngle &operator/=(float rhs) {Radian /= rhs;return *this;}
	RadianAngle operator*(float rhs) const {return RadianAngle(Radian * rhs);}
	RadianAngle operator/(float rhs) const {return RadianAngle(Radian / rhs);}

private:
	float Radian;
};

#endif
