
#ifndef Game_Attack_H
#define Game_Attack_H

#include "Juna/intrusive_list.hpp"
#include "TypeDef.h"


class I_Attacked;
class I_Attacker;

class AttackParam;

class I_AttackedCreator;
class I_AttackerCreator;

class AttackManager;

typedef Juna::intrusive_list<I_Attacked> AttackedList;
typedef Juna::intrusive_list<I_Attacker> AttackerList;


class I_Attacked : public Juna::intrusive_list_node
{
public:
	virtual const I_Shape &attackedShape(void) const = 0;

private: friend class I_Attacker;
	virtual void Damaged(uint32 power,const Vector &vector,
						 const I_Attacker &attacker) = 0;
protected:
	I_Attacked(I_AttackedCreator &creator);
	~I_Attacked() {unlink();}
private:
	using Juna::intrusive_list_node::unlink;
};

class I_Attacker : public Juna::intrusive_list_node
{
private: friend class AttackManager;
	virtual void Attack(const AttackParam &param) = 0;

protected:
	void Damage(I_Attacked &target,uint32 power,const Vector &vector)
		{target.Damaged(power,vector,*this);}

protected:
	I_Attacker(I_AttackerCreator &creator);
	~I_Attacker() {unlink();}
private:
	using Juna::intrusive_list_node::unlink;
};

class AttackParam
{
public:
	typedef AttackedList list;

	const list &Enemy;
	const list &Friend;

	AttackParam(const list &enemy,const list &friend_)
		: Enemy(enemy),Friend(friend_) {}
private:
	void operator=(AttackParam &);
};


class I_AttackedCreator
{
private: friend class I_Attacked;
	void Create(I_Attacked &attacked) {Creature.push_back(attacked);}
private: friend class AttackManager;
	AttackedList Creature;
};
class I_AttackerCreator
{
private: friend class I_Attacker;
	void Create(I_Attacker &attacker) {Creature.push_back(attacker);}
private: friend class AttackManager;
	AttackerList Creature;
};

inline I_Attacked::I_Attacked(I_AttackedCreator &creator) {creator.Create(*this);}
inline I_Attacker::I_Attacker(I_AttackerCreator &creator) {creator.Create(*this);}



class AttackManager
{
public:
	void AttackPhase(const AttackManager &enemy)
	{
		AttackParam ap(enemy.Attacked,Attacked);
		for (AttackerList::iterator it = Attacker.begin();it != Attacker.end();++it)
			it->Attack(ap);
	}
	void RealizeCreature(void)
	{
		Attacked.splice(Attacked.end(),dCreator.Creature);
		Attacker.splice(Attacker.end(),rCreator.Creature);
	}
	const AttackedList &attacked(void) const {return Attacked;}
	const AttackerList &attacker(void) const {return Attacker;}

private:
	AttackedList Attacked;
	AttackerList Attacker;
public:
	I_AttackedCreator dCreator;
	I_AttackerCreator rCreator;
};

#endif
