
#include "BulletMLObject.h"

#include "GameDefine.h"

#include "Attack.h"

#include "VisualizeParam.h"

#include <math.h>



#include "libBulletML/bulletmlparser.h"
#include "libBulletML/bulletmlparser-tinyxml.h"
#include "libBulletML/bulletmlrunner.h"

//#define STATIC
//static linkłgAŗpӂāBTCYł
#ifdef STATIC
	#ifndef _DEBUG
		#pragma comment(lib , "libBulletML/libbulletml.lib")
	#else
		#pragma comment(lib , "libBulletML/libbulletml_d.lib")
	#endif
#else
	#ifndef _DEBUG
		#pragma comment(lib , "libBulletML/bulletml.lib")
	#else
		#pragma comment(lib , "libBulletML/bulletml_d.lib")
	#endif
#endif

using namespace Juna;

BulletMLManager::BulletMLManager(void) : ObjectCount(0)
{
	const char * const xml[] = {
		"bml/test.xml",
	};
	for (int i = 0;i < 1;i++)
	{
		Parsers.push_back(new ::BulletMLParserTinyXML(xml[i]));
		Parsers.back()->build();
	}

}
BulletMLManager::~BulletMLManager()
{
	for (ParserContainer::iterator it = Parsers.begin();it != Parsers.end();++it)
	{
//DLLlibNƂŕ傪o
//炭DLL̊JɂɓBĂ܂H
		delete *it;
		*it = 0;
	}
}

BulletMLObject *BulletMLManager::Create(BulletMLParser *parser,
	Generator &generator,const Point &position,RadianAngle direction)
{
	ObjectCount++;
	return new BulletMLObject(generator,position,direction,parser);
}


BulletMLManager &BulletMLManager::GetInstance(void)
{
	static BulletMLManager instance;
	return instance;
}

BulletMLObject::VoidController BulletMLObject::Controller;

void BulletMLObject::Break(void)
{
	BulletMLManager::GetInstance().ObjectCount--;
	delete this;
}
bool BulletMLObject::Update(void)
{
	if (LifeCounter <= 0)
	{
		return false;
	}
	UpdatePosition();
	return true;
}

void BulletMLObject::Observe(const ObserveParam &param)
{
	const ObserveParam::list &enemy = param.Enemy;
	if (!enemy.empty())
		Target = enemy.front().position();

	if (isEnd())
		reset(Parser);
	TurnCounter++;
	run();
}

void BulletMLObject::Damaged(uint32 power,const Vector &vector,
							 const I_Attacker &attacker)
{
	LifeCounter -= power;
}

void BulletMLObject::Visualize(const VisualizeParam &param) const
{
	using Juna::D2D::Color;
	param.Screen.Blit(position().x,position().y,
		param.Sprite,Rect(0,0,Size(64,64)),
		direction().radian(),1.0f,1.0f,
		Color(255,128,128,255));
}


namespace {
bool OutOfX(const Point &pos)
{
	return pos.x > StgViewRight || pos.x < StgViewLeft;
}

bool OutOfY(const Point &pos)
{
	return pos.y > StgViewBottom || pos.y < StgViewTop;
}
bool OutOfBottom(const Point &pos)
{
	return pos.y > StgViewBottom;
}

bool OutOfField(const Point &pos)
{
	return OutOfX(pos) || OutOfY(pos);
}
	static const float HitRange = 4;
	static const int AttackPower = 1;
}

class BulletMLShot : public I_Entity, public I_Visual, public I_Observer, public I_Attacker,
					 public ::BulletMLRunner
{
public:
	BulletMLShot(::Generator &generator,
				 const Point &position,RadianAngle direction,float speed,
				 BulletMLState *state)
	  : I_Entity(generator.Entity),
		I_Visual(generator.Visual,VL_Top),
		I_Observer(generator.Observer),
		I_Attacker(generator.Attacker),
		BulletMLRunner(state),
		Generator(generator),
		Position(position), Direction(direction),
		Velocity(direction.cos() * speed,direction.sin() * speed),Speed(speed),
		Range(HitRange), Power(AttackPower),Alive(true),
		Target(0,0),Turn(0),Accel(0,0)
	{}

	BulletMLShot(::Generator &generator,
				 const Point &position,RadianAngle direction,float speed,
				 BulletMLParser *parser)
	  : I_Entity(generator.Entity),
		I_Visual(generator.Visual,VL_Top),
		I_Observer(generator.Observer),
		I_Attacker(generator.Attacker),
		BulletMLRunner(parser),
		Generator(generator),
		Position(position), Direction(direction),
		Velocity(direction.cos() * speed,direction.sin() * speed),Speed(speed),
		Range(HitRange), Power(AttackPower),Alive(true),
		Target(0,0),Turn(0),Accel(0,0)
	{}

	~BulletMLShot() {}

	virtual bool Update(void)
	{
//		if (OutOfField(Position) || !Alive)
		if (OutOfX(Position) || OutOfBottom(Position) || !Alive)
			return false;
		Position += Velocity;
		return true;
	}
	virtual void Break(void) {delete this;}
	virtual void Attack(const AttackParam &param)
	{
		struct Circle : public Juna::D2::Intersection::I_Circle
		{
			Circle(const BulletMLShot &shot) : shot(shot) {}
			virtual Data GetCircleData(void) const
			{
				return Data(shot.position(),shot.range());
			}
			const BulletMLShot &shot;
		};
		Circle c(*this);

		typedef AttackParam::list list;
		const list &enemy = param.Enemy;
		for (list::const_iterator it = enemy.begin();it != enemy.end();++it)
		{
			if (it->attackedShape().Test(c))
			{
				Damage(*it,Power,Vector(0,0));
				Alive = false;
			}
		}
	}
	virtual void Observe(const ObserveParam &param)
	{
		const ObserveParam::list &enemy = param.Enemy;
		if (!enemy.empty())
			Target = enemy.front().position();

		Turn++;
		run();
	}

	virtual void Visualize(const VisualizeParam &param) const
	{
		param.Screen.Blit(Position.x,Position.y,
			param.Sprite,Rect(128,0,Size(16,16)),Direction.radian());
	}

	const Point &position(void) const {return Position;}
	float range(void) const {return Range;}


private:
	virtual double getBulletDirection() {return Direction.radian() * (360 / Pi2) + 90;}
	virtual double getAimDirection()
	{
		return ::atan2(Target.y - Position.y,Target.x - Position.x) * (360 / Pi2) + 90;
	}
	virtual double getBulletSpeed() {return Speed;}
	virtual double getDefaultSpeed() {return Speed;}
	virtual double getRank() {return 0.5;}
	virtual void createSimpleBullet(double direction, double speed);
	virtual void createBullet(BulletMLState* state,
							  double direction, double speed)
	{
		direction -= 90;
		new BulletMLShot(Generator,Position,float(direction * (Pi2 / 360)),float(speed),state);
	}
	virtual int getTurn() {return Turn;}
	virtual void doVanish() {Alive = false;}

	virtual void doChangeDirection(double direction)
	{
		direction -= 90;
		Direction = float(direction * (Pi2 / 360));
	}
	virtual void doChangeSpeed(double speed)
	{
		Speed = float(speed);
		Velocity.Set(Direction.cos() * Speed,Direction.sin() * Speed);
	}
	virtual void doAccelX(double ax) {Velocity.x += float(ax) - Accel.x;Accel.x = float(ax);}
	virtual void doAccelY(double ay) {Velocity.y += float(ay) - Accel.y;Accel.y = float(ay);}
	virtual double getBulletSpeedX() {return Velocity.x;}
	virtual double getBulletSpeedY() {return Velocity.y;}
//	virtual double getRand() { return (double)rand() / RAND_MAX; }

private:

	Generator &Generator;
	Point Position;
	RadianAngle Direction;
	Vector Velocity;
	float Speed;
	float Range;
	uint32 Power;
	bool Alive;

	Point Target;
	int Turn;
	Vector Accel;
};

void BulletMLObject::createBullet(BulletMLState* state,
								  double direction, double speed)
{
	direction -= 90;
	new BulletMLShot(Generator,position(),float(direction * (Pi2 / 360)),float(speed),state);
}

class BulletMLSimpleShot : public I_Entity, public I_Visual, public I_Attacker
{
public:
	BulletMLSimpleShot(::Generator &generator,
					   const Point &position,double direction,float speed)
	  : I_Entity(generator.Entity),
		I_Visual(generator.Visual,VL_Top),
		I_Attacker(generator.Attacker),
		Generator(generator),
		Position(position), Direction(float(direction * (Pi2 / 360))),
		Velocity(Direction.cos() * speed,Direction.sin() * speed),
		Range(HitRange), Power(AttackPower), Alive(true)
	{}

	~BulletMLSimpleShot() {}

	virtual bool Update(void)
	{
		if (OutOfField(Position) || !Alive)
			return false;
		Position += Velocity;
		return true;
	}
	virtual void Break(void) {delete this;}
	virtual void Attack(const AttackParam &param)
	{
		struct Circle : public Juna::D2::Intersection::I_Circle
		{
			Circle(const BulletMLSimpleShot &shot) : shot(shot) {}
			virtual Data GetCircleData(void) const
			{
				return Data(shot.position(),shot.range());
			}
			const BulletMLSimpleShot &shot;
		};
		Circle c(*this);

		typedef AttackParam::list list;
		const list &enemy = param.Enemy;
		for (list::const_iterator it = enemy.begin();it != enemy.end();++it)
		{
			if (it->attackedShape().Test(c))
			{
				Damage(*it,Power,Vector(0,0));
				Alive = false;
			}
		}
	}

	virtual void Visualize(const VisualizeParam &param) const
	{
		param.Screen.Blit(Position.x,Position.y,
			param.Sprite,Rect(128,0,Size(16,16)),Direction.radian());
	}

	const Point &position(void) const {return Position;}
	float range(void) const {return Range;}

private:

	Generator &Generator;
	Point Position;
	RadianAngle Direction;
	Vector Velocity;
	float Range;
	uint32 Power;
	bool Alive;
};

void BulletMLObject::createSimpleBullet(double direction, double speed)
{
	direction -= 90;
	new BulletMLSimpleShot(Generator,position(),direction,float(speed));
}


void BulletMLShot::createSimpleBullet(double direction, double speed)
{
	direction -= 90;
	new BulletMLSimpleShot(Generator,position(),direction,float(speed));
}
