
#ifndef Game_BulletML_Object_H
#define Game_BulletML_Object_H


#include "TypeDef.h"

#include "Automaton.h"


#include "Generator.h"

#include "libBulletML/bulletmlrunner.h"

#include <vector>
#include <string>

class BulletMLParser;
class BulletMLObject;

class BulletMLManager
{
public:
	typedef std::vector<BulletMLParser *> ParserContainer;
private:
	BulletMLManager(void);
	BulletMLManager(BulletMLManager&);
public:
	~BulletMLManager();


	BulletMLObject *Create(::BulletMLParser *parser,
						   Generator &generator,const Point &position,RadianAngle direction);


	const ParserContainer &GetParsers(void) const {return Parsers;}

	int objectCount(void) const {return ObjectCount;};

	static BulletMLManager &GetInstance(void);
private:
	ParserContainer Parsers;

	friend class BulletMLObject;
	int ObjectCount;
};

class BulletMLObject : public base_Automaton, public Juna::D2::Intersection::I_Circle,
					   public BulletMLRunner
{
	class VoidController : public I_UnitController
	{
	public:
		virtual void UnitControl(I_Unit &,const ObserveParam &) {}
		virtual void UnitBreak(I_Unit &) {}
	};
	static VoidController Controller;
public:
	BulletMLObject(::Generator &generator,const Point &position,RadianAngle direction,
				   BulletMLParser *parser,float bullet_speed = 4)
	  : base_Automaton(Controller,generator,position,direction,Vector(0,0)),
		BulletMLRunner(parser),
		Parser(parser), Speed(DefaultSpeed), DefaultSpeed(bullet_speed),
		Target(0,0),TurnCounter(0),
		LifeCounter(1000)
	{}
	~BulletMLObject() {}

public:
	virtual bool Update(void);
	virtual void Break(void);
	virtual void Visualize(const VisualizeParam &param) const;

	virtual void Observe(const ObserveParam &param);

	virtual const I_Shape &attackedShape(void) const {return *this;}

	virtual void Damaged(uint32 power,const Vector &vector,
						 const I_Attacker &attacker);

	virtual void Move(RadianAngle ) {}
	virtual void Turn(RadianAngle ) {}
	virtual void Act(unsigned int ) {}

	virtual Data GetCircleData(void) const {return Data(position(),8);}
private:
	virtual double getBulletDirection() {return direction().radian() * (360 / Pi2) + 90.0;}
	virtual double getAimDirection()
	{
		double angle = ::atan2(Target.y - position().y,Target.x - position().x) * (360 / Pi2);
		return angle + 90.0;
	}
	virtual double getBulletSpeed() {return Speed;}
	virtual double getDefaultSpeed() {return DefaultSpeed;}
	virtual double getRank() {return 0.5;}
	virtual void createSimpleBullet(double direction, double speed);
	virtual void createBullet(BulletMLState* state,
							  double direction, double speed);
	virtual int getTurn() {return TurnCounter;}
	virtual void doVanish() {LifeCounter = 0;}

/*
	virtual void doChangeDirection(double) {}
	virtual void doChangeSpeed(double) {}
	virtual void doAccelX(double) {}
	virtual void doAccelY(double) {}
	virtual double getBulletSpeedX() { return velocity().x;}
	virtual double getBulletSpeedY() { return velocity().y;}
	virtual double getRand() { return (double)rand() / RAND_MAX; }
*/

private:
	BulletMLParser *Parser;
	float Speed;
	float DefaultSpeed;
	Point Target;
	int TurnCounter;

	int LifeCounter;
};


#endif
