
#include "Enemy.h"
#include "Allocator.h"
#include "VisualizeParam.h"
#include "GameDefine.h"
#include "Sounder.h"

#include "Shot.h"


#include <new>
#include <math.h>

Enemy *Enemy::Create(::Generator &generator,int32 life_counter,
					 const Point &position,RadianAngle angle,const Vector & velocity)
{
	I_Allocator &a = generator.Allocator;
	assert(a.blockSize() >= sizeof(Enemy));
	if (a.full())
		return 0;
	return new(a.Allocate()) Enemy(generator,life_counter,position,angle,velocity);

//	return new Enemy(generator,life_counter,position,angle,velocity);
}

void Enemy::Break(void)
{
	I_Allocator &a = Generator.Allocator;
	this->~Enemy();
	a.Deallocate(this);

//	delete this;
}

bool OutOfX(const Point &pos)
{
	return pos.x > StgViewRight || pos.x < StgViewLeft;
}
bool OutOfY(const Point &pos)
{
	return pos.y > StgViewBottom || pos.y < StgViewTop;
}

bool OutOfField(const Point &pos)
{
	return OutOfX(pos) || OutOfY(pos);
}

bool Enemy::Update(void)
{
	UpdatePosition();
	if (LifeCounter > 0)
	{
		return true;
	}
	else
	{
		Sounder::instance.PlaySE1();
		return false;
	}
}
void Enemy::Damaged(uint32 power,const Vector &vector,
					const I_Attacker &attacker)
{
	LifeCounter -= power;
	Velocity += vector;
}

void Enemy::Observe(const ObserveParam &param)
{
	Controller->UnitControl(*this,param);
}

void Enemy::Visualize(const VisualizeParam &param) const
{
	using Juna::D2D::Color;
	param.Screen.Blit(position().x,position().y,
		param.Sprite,Rect(0,0,Size(64,64)),
		direction().radian(),1.0f,1.0f,
		Color(255,255,128,128));
}

void Enemy::Move(RadianAngle direction)
{
	Velocity.Set(direction.cos(),direction.sin());
}
void Enemy::Turn(RadianAngle direction)
{
	AngleV = direction - this->direction();
}
void Enemy::Act(unsigned int command)
{
	switch (command)
	{
	case 0:
	default:
		Shot::Create(Generator,position(),direction().radian(),
					 Vector(direction().cos() * 3,direction().sin() * 3));
		break;
	}
}


void Enemy::DefController::UnitControl(I_Unit &controlled,const ObserveParam &observed)
{
	RadianAngle a = controlled.direction();
	if (OutOfX(controlled.position()))
	{
		a += Pi + Pi / 2;
		a.SetNormal();
		a -= Pi;
		a = -a;
		a -= Pi / 2;
	}
	if (OutOfY(controlled.position()))
	{
		a.SetNormal2();
		a = -a;
	}
	controlled.Turn(a);
	controlled.Move(a);
	ActionCounter++;
	if (ActionCounter >= 60)
	{
		controlled.Act(0);
		ActionCounter = 0;
	}
}
