
#ifndef Juna_Input_H
#define Juna_Input_H


struct IDirectInputDevice8;

namespace Juna {
namespace Input {

class I_GameController
{
protected:
	I_GameController(void)
		: X_center_rate(0.5f), Y_center_rate(0.5f) {ClearInput();}
	~I_GameController() {}
public:

	float x_axis(void) const {return X_axis;}
	float y_axis(void) const {return Y_axis;}

	bool buttonState(int index) const {return Button[index];}
	bool buttonPressed(int index) const {return (Button[index]) & !LastButton[index];}
	bool buttonReleased(int index) const {return !Button[index] & (LastButton[index]);}

	bool left(void) const {return X_axis < -X_center_rate;}
	bool right(void) const {return X_axis > X_center_rate;}
	bool up(void) const {return Y_axis < -Y_center_rate;}
	bool down(void) const {return Y_axis > Y_center_rate;}

	void SetCenterRangeX(float rate) {X_center_rate = rate;}
	void SetCenterRangeY(float rate) {Y_center_rate = rate;}

	virtual unsigned buttonNum(void) const = 0;
//͏ԂXV܂
	virtual bool GetInput(void) = 0;
//OGetInput()ɂXV𖳌ɂB
//	virtual void CancelInput(void) = 0;

	void ClearInput(void)
	{
		X_axis = Y_axis = 0;
		for (int i = 0;i< 32;i++)
			Button[i] = LastButton[i] = false;
	}
protected:
	float X_axis,Y_axis;
	float X_center_rate,Y_center_rate;

	bool Button[32];
	bool LastButton[32];
};



class DIGameController : public I_GameController
{
public:
	DIGameController(void);
	~DIGameController();

	virtual unsigned buttonNum(void) const {return ButtonNum;}
//͏ԂXV܂
	virtual bool GetInput(void);
//OGetInput()ɂXV𖳌ɂB
	void CancelInput(void);

	bool Acquire(void);
	bool UnAcquire(void);

	bool Initialize(int id = -1);
	void Terminalize(void);

private:
	IDirectInputDevice8 *Device;
	unsigned ButtonNum;

	long X_center,Y_center;
	float X_range,Y_range;
};

class APIGameController : public I_GameController
{
public:
	APIGameController(unsigned int id = 0);
	~APIGameController() {}

	virtual unsigned buttonNum(void) const {return ButtonNum;}
//͏ԂXV܂
	virtual bool GetInput(void);
//OGetInput()ɂXV𖳌ɂB
	void CancelInput(void);

	void Reset(unsigned int id = 0);

private:
	unsigned int JoyID;
	unsigned ButtonNum;

	unsigned int X_center,Y_center;
	float X_range,Y_range;
};



enum KeyCode
{
KC_ESCAPE = 0x01,
KC_1,KC_2,KC_3,KC_4,KC_5,
KC_6,KC_7,KC_8,KC_9,KC_0,

KC_MINUS,KC_EQUALS,KC_BACK,KC_TAB,

KC_Q,KC_W,KC_E,KC_R,KC_T,KC_Y,KC_U,KC_I,KC_O,KC_P,

KC_LBRACKET,KC_RBRACKET,KC_RETURN,KC_LCONTROL,

KC_A,KC_S,KC_D,KC_F,KC_G,KC_H,KC_J,KC_K,KC_L,

KC_SEMICOLON,KC_APOSTROPHE,KC_GRAVE,KC_LSHIFT,KC_BACKSLASH,

KC_Z,KC_X,KC_C,KC_V,KC_B,KC_N,KC_M,

KC_COMMA,KC_PERIOD,KC_SLASH,KC_RSHIFT,KC_MULTIPLY,
KC_LMENU,KC_SPACE,KC_CAPITAL,

KC_F1,KC_F2,KC_F3,KC_F4,KC_F5,
KC_F6,KC_F7,KC_F8,KC_F9,KC_F10,

KC_NUMLOCK,KC_SCROLL,

KC_NUMPAD7,KC_NUMPAD8,KC_NUMPAD9,KC_SUBTRACT,
KC_NUMPAD4,KC_NUMPAD5,KC_NUMPAD6,KC_ADD,
KC_NUMPAD1,KC_NUMPAD2,KC_NUMPAD3,
KC_NUMPAD0,KC_DECIMAL,
KC_OEM_102 = 0x56,    /* <> or \| on RT 102-key keyboard (Non-U.S.) */
KC_F11,KC_F12,
KC_F13 = 0x64,KC_F14,KC_F15,
KC_KANA = 0x70,
KC_ABNT_C1 = 0x73,
KC_CONVERT = 0x79,KC_NOCONVERT =0x7B,KC_YEN = 0x7D,
KC_ABNT_C2 = 0x7E,

KC_NUMPADEQUALS = 0x8D,
KC_PREVTRACK = 0x90,
KC_AT,KC_COLON,KC_UNDERLINE,KC_KANJI,
KC_STOP,KC_AX,KC_UNLABELED,

KC_NEXTTRACK = 0x99,    /* Next Track */
KC_NUMPADENTER = 0x9C,KC_RCONTROL,
KC_MUTE = 0xA0,KC_CALCULATOR,KC_PLAYPAUSE,KC_MEDIASTOP = 0xA4,
KC_VOLUMEDOWN = 0xAE,KC_VOLUMEUP = 0xB0,
KC_WEBHOME = 0xB2,
KC_NUMPADCOMMA = 0xB3,KC_DIVIDE = 0xB5,
KC_SYSRQ = 0xB7,
KC_RMENU = 0xB8,KC_PAUSE = 0xC5,KC_HOME = 0xC7,
KC_UP = 0xC8,KC_PRIOR,
KC_LEFT = 0xCB,KC_RIGHT = 0xCD,KC_END = 0xCF,
KC_DOWN = 0xD0,KC_NEXT,KC_INSERT,KC_DELETE,

KC_LWIN = 0xDB,KC_RWIN,KC_APPS,KC_POWER,KC_SLEEP,

KC_WAKE = 0xE3,
KC_WEBSEARCH = 0xE5,KC_WEBFAVORITES,KC_WEBREFRESH,KC_WEBSTOP,
KC_WEBFORWARD,KC_WEBBACK,KC_MYCOMPUTER,KC_MAIL,KC_MEDIASELECT,

KC_BACKSPACE = KC_BACK,
KC_LALT = KC_LMENU,
KC_RALT = KC_RMENU,
KC_CAPSLOCK = KC_CAPITAL,
KC_NUMPADPLUS = KC_ADD,
KC_NUMPADMINUS = KC_SUBTRACT,
KC_NUMPADSTAR = KC_MULTIPLY,
KC_NUMPADSLASH = KC_DIVIDE,
KC_NUMPADPERIOD = KC_DECIMAL,
KC_LEFTARROW = KC_LEFT,
KC_UPARROW = KC_UP,
KC_RIGHTARROW = KC_RIGHT,
KC_DOWNARROW = KC_DOWN,
KC_PGUP = KC_PRIOR,
KC_PGDN = KC_NEXT,

KC_MAX = 256
};


//L[{[h
class Keyboard
{
public:
	Keyboard(void);
	~Keyboard();

	//L[{[h
	bool keyState(int keycode) const {return KeyState[keycode] != 0;}
	bool keyPressed(int keycode) const {return (KeyState[keycode] != 0) & !LastKeyState[keycode];}
	bool keyReleased(int keycode) const {return !KeyState[keycode] & (LastKeyState[keycode] != 0);}

//͏ԂXV܂
	bool GetInput(void);
//OGetInput()ɂXV𖳌ɂB
	void CancelInput(void);

	bool Initialize(void);
	void Terminalize(void);

private:
	IDirectInputDevice8 *Device;
	unsigned char KeyState[256];
	unsigned char LastKeyState[256];
};

}//namespace Juna
}//namespace Input

#endif
