
#include "2d_intersection.h"

namespace Juna {
namespace D2 {
namespace Intersection {

InfiniteShape InfiniteShape::Instance;
VoidShape VoidShape::Instance;



inline bool Point_LineDividedArea(
	const Point &p,
	Number_Type la,Number_Type lb,Number_Type lc,bool lless)
{
	if (lless)
	{
		if (la * p.x + lb * p.y + lc <= 0)
			return true;
	}
	else
	{
		if (la * p.x + lb * p.y + lc >= 0)
			return true;
	}
	return false;
}

bool Circle_LineDividedArea(
	const Point &cc,Number_Type cr,
	Number_Type la,Number_Type lb,Number_Type lc,bool lless)
{
	if (lless)
	{
		if (la * cc.x + lb * cc.y + lc <= 0)
			return true;
	}
	else
	{
		if (la * cc.x + lb * cc.y + lc >= 0)
			return true;
	}
	if (la == 0)
	{
		Number_Type y = -lc / lb;
		return Point_Line_Distance2(cc,Point(0,y),Point(1,y)) <= square(cr);
	}
	else if (lb == 0)
	{
		Number_Type x = -lc / la;
		return Point_Line_Distance2(cc,Point(x,0),Point(x,1)) <= square(cr);
	}

	Number_Type y = (la * cc.x + lc) / -lb;
	Number_Type x = (lb * cc.y + lc) / -la;
	return Point_Line_Distance2(cc,Point(cc.x,y),Point(x,cc.y)) <= square(cr);
}

//
bool LineDividedArea_LineDividedArea(
	Number_Type l1a,Number_Type l1b,Number_Type l1c,bool l1less,
	Number_Type l2a,Number_Type l2b,Number_Type l2c,bool l2less)
{
	if (l1a == 0 || l2a == 0)
	{
		if (l1a != l2a)
			return true;
		if (l1less  == l2less)
			return true;
		Number_Type y1 = -l1c / l1b;
		Number_Type y2 = -l2c / l2b;
		if ((y1 < y2 && l1less) || (y2 < y1 && l2less))
			return false;
		return true;
	}
	if (l1b == 0 || l2b == 0)
	{
		if (l1b != l2b)
			return true;
		Number_Type x1 = -l1c / l1a;
		Number_Type x2 = -l2c / l2a;
		if ((x1 < x2 && l1less) || (x2 < x1 && l2less))
			return false;
		return true;
	}
	Number_Type t1 = l1a / l1b;
	Number_Type t2 = l2a / l2b;

	if (t1 != t2)
		return true;

	Number_Type y1 = -l1c / l1b;
	Number_Type y2 = -l2c / l2b;
	if ((y1 < y2 && l1less) || (y2 < y1 && l2less))
		return false;
	return true;
}


Number_Type	Point_Segment_Distance2(
	const Point &p,
	const Point &s1,const Point &s2)
{
	Number_Type a = square(p.x - s1.x) + square(p.y - s1.y);
	Number_Type b = square(p.x - s2.x) + square(p.y - s2.y);
	Number_Type c = square(s1.x - s2.x) + square(s1.y - s2.y);
	if (a > b)
	{
		if (a >= b + c)
			return b;
	}
	else
	{
		if (b >= a + c)
			return a;
	}
	Number_Type r = (a + b -(a-b)/(c*2)*(a-b) - c/2) / 2;
//vZ덷̂Ń}CiXɂȂĂ܂
	return r * (r > 0);
//	return (a/2 + b/2 -(a-b)*(a-b)/(c*2)/2 - c/4);
}

bool Circle_Segment(
	const Point &cc,Number_Type cr,
	const Point &s1,const Point &s2,Number_Type shw)
{
	return Point_Segment_Distance2(cc,s1,s2) < square(cr + shw);
}


bool LineDividedArea_Segment(
	Number_Type la,Number_Type lb,Number_Type lc,bool lless,
	const Point &s1,const Point &s2,Number_Type shw)
{
	return Circle_LineDividedArea(s1,shw,la,lb,lc,lless) ||
		   Circle_LineDividedArea(s2,shw,la,lb,lc,lless);
}

bool Segment_Segment(
	const Point &s1p1,const Point &s1p2,Number_Type s1hw,
	const Point &s2p1,const Point &s2p2,Number_Type s2hw)
{
	Number_Type range2 = square(s1hw + s2hw);
	if (Point_Segment_Distance2(s1p1,s2p1,s2p2) < range2 ||
		Point_Segment_Distance2(s1p2,s2p1,s2p2) < range2 ||
		Point_Segment_Distance2(s2p1,s1p1,s1p2) < range2 ||
		Point_Segment_Distance2(s2p2,s1p1,s1p2) < range2)
	{
		return true;
	}

	Number_Type d1x = s1p1.x - s1p2.x;
	Number_Type d1y = s1p1.y - s1p2.y;
	Number_Type d2x = s2p1.x - s2p2.x;
	Number_Type d2y = s2p1.y - s2p2.y;
	if ((d1x * (s2p1.y - s1p1.y) + d1y * (s1p1.x - s2p1.x)) *
		(d1x * (s2p2.y - s1p1.y) + d1y * (s1p1.x - s2p2.x)) < 0 &&
		(d2x * (s1p1.y - s2p1.y) + d2y * (s2p1.x - s1p1.x)) *
		(d2x * (s1p2.y - s2p1.y) + d2y * (s2p1.x - s1p2.x)) < 0)
	{
		return true;
	}
	return false;
}


namespace {

bool IntersectionTest(const I_Circle &c1,const I_Circle &c2)
{
	I_Circle::Data cd1 = c1.GetCircleData();
	I_Circle::Data cd2 = c2.GetCircleData();
	return Circle_Circle(cd1.center,cd1.radius,cd2.center,cd2.radius);
}

bool IntersectionTest(const I_Circle &c,const I_LineDividedArea &lda)
{
	I_Circle::Data cd = c.GetCircleData();
	I_LineDividedArea::Data ldad = lda.GetLineDividedAreaData();
	return Circle_LineDividedArea(
				cd.center,cd.radius,
				ldad.a,ldad.b,ldad.c,ldad.less);
}
bool IntersectionTest(const I_Circle &c,const I_Segment &s)
{
	I_Circle::Data cd = c.GetCircleData();
	I_Segment::Data sd = s.GetSegmentData();
	return Circle_Segment(
				cd.center,cd.radius,
				sd.edge1,sd.edge2,sd.half_width);
}

bool IntersectionTest(const I_LineDividedArea &lda1,const I_LineDividedArea &lda2)
{
	I_LineDividedArea::Data ldad1 = lda1.GetLineDividedAreaData();
	I_LineDividedArea::Data ldad2 = lda2.GetLineDividedAreaData();
	return LineDividedArea_LineDividedArea(
				ldad1.a,ldad1.b,ldad1.c,ldad1.less,
				ldad2.a,ldad2.b,ldad2.c,ldad2.less);
}

bool IntersectionTest(const I_LineDividedArea &lda,const I_Segment &s)
{
	I_LineDividedArea::Data ldad = lda.GetLineDividedAreaData();
	I_Segment::Data sd = s.GetSegmentData();

	return LineDividedArea_Segment(
				ldad.a,ldad.b,ldad.c,ldad.less,
				sd.edge1,sd.edge2,sd.half_width);
}

bool IntersectionTest(const I_Segment &s1,const I_Segment &s2)
{
	I_Segment::Data sd1 = s1.GetSegmentData();
	I_Segment::Data sd2 = s2.GetSegmentData();
	return Segment_Segment(sd1.edge1,sd1.edge2,sd1.half_width,
						   sd2.edge1,sd2.edge2,sd2.half_width);
}

}//anonymous namespace


bool I_Circle::Test(const I_Circle &other) const
	{return IntersectionTest(*this,other);}
bool I_Circle::Test(const I_LineDividedArea &other) const
	{return IntersectionTest(*this,other);}
bool I_Circle::Test(const I_Segment &other) const
	{return IntersectionTest(*this,other);}

bool I_LineDividedArea::Test(const I_Circle &other) const
	{return IntersectionTest(other,*this);}
bool I_LineDividedArea::Test(const I_LineDividedArea &other) const
	{return IntersectionTest(*this,other);}
bool I_LineDividedArea::Test(const I_Segment &other) const
	{return IntersectionTest(*this,other);}

bool I_Segment::Test(const I_Circle &other) const
	{return IntersectionTest(other,*this);}
bool I_Segment::Test(const I_LineDividedArea &other) const
	{return IntersectionTest(other,*this);}
bool I_Segment::Test(const I_Segment &other) const
	{return IntersectionTest(*this,other);}


namespace {
template<typename T>
bool UnitedShapeTest(const I_UnitedShape &this_,const T &other)
{
	int n = this_.GetPartsNum();
	for (int i = 0;i < n;i++)
	{
		if (this_.GetPart(i).Test(other))
			return true;
	}
	return false;
}
}// anonymous namespace

bool I_UnitedShape::Test(const I_Circle &other) const
	{return UnitedShapeTest(*this,other);}
bool I_UnitedShape::Test(const I_LineDividedArea &other) const
	{return UnitedShapeTest(*this,other);}
bool I_UnitedShape::Test(const I_Segment &other) const
	{return UnitedShapeTest(*this,other);}
bool I_UnitedShape::Test(const I_CombinationShape &other) const
	{return UnitedShapeTest(*this,other);}

}//	namespace Intersection
}//	namespace D2
}//	namespace Juna


