
#ifndef Juna_2D_Intersection_Test_H
#define Juna_2D_Intersection_Test_H
//񎟌

#include "2d_coordinate.h"

namespace Juna {
namespace D2 {
namespace Intersection {

//gp鐔ľ^
typedef float Number_Type;
typedef D2::Point<Number_Type> Point;

//	{IȌvZ
inline Number_Type square(Number_Type x) {return x * x;}

inline bool Circle_Circle(const Point &c1c,Number_Type c1r,
						  const Point &c2c,Number_Type c2r)
	{return square(c1c.x - c2c.x) + square(c1c.y - c2c.y) < square(c1r + c2r);}

inline Number_Type	Point_Line_Distance2(const Point &p,
										 const Point &lp1,const Point &lp2)
{
	Number_Type a = square(p.x - lp1.x) + square(p.y - lp1.y);
	Number_Type b = square(p.x - lp2.x) + square(p.y - lp2.y);
	Number_Type c = square(lp1.x - lp2.x) + square(lp1.y - lp2.y);

	Number_Type r = (a + b -(a-b)/(c*2)*(a-b) - c/2) / 2;
//vZ덷̂Ń}CiXɂȂĂ܂
	return r * (r > 0);
//	return (a/2 + b/2 -(a-b)*(a-b)/(c*2)/2 - c/4);
}



class I_Shape
{
private:
	friend class VoidShape;		//
	friend class InfiniteShape;	//

	friend class I_Circle;			//~
	friend class I_LineDividedArea;	//ŕꂽ̈
	friend class I_Segment;			//

	friend class I_CombinationShape;	//gݍ킹

	~I_Shape() {}
public:

	virtual bool Test(const I_Shape &other) const = 0;

//_ufBXpb`
	virtual bool Test(const VoidShape &) const {return false;}
	virtual bool Test(const InfiniteShape &) const {return true;}

	virtual bool Test(const I_Circle &other) const = 0;
	virtual bool Test(const I_LineDividedArea &other) const = 0;
	virtual bool Test(const I_Segment &other) const = 0;

	virtual bool Test(const I_CombinationShape &other) const = 0;
};

//	falseƂȂۂ̌`
class VoidShape : public I_Shape
{
public:
	virtual bool Test(const I_Shape &) const {return false;}
	virtual bool Test(const InfiniteShape &) const {return false;}

	virtual bool Test(const I_Circle &) const {return false;}
	virtual bool Test(const I_LineDividedArea &) const {return false;}
	virtual bool Test(const I_Segment &) const {return false;}
	virtual bool Test(const I_CombinationShape &) const {return false;}

	static VoidShape Instance;
};
//	VoidShapeȊO͏trueƂȂ閳̌`
class InfiniteShape : public I_Shape
{
public:
//	Void̏ꍇfalsẻ\邽ߔ
	virtual bool Test(const I_Shape &other) const {return other.Test(*this);}

	virtual bool Test(const I_Circle &) const {return true;}
	virtual bool Test(const I_LineDividedArea &) const {return true;}
	virtual bool Test(const I_Segment &) const {return true;}
	virtual bool Test(const I_CombinationShape &) const {return true;}

	static InfiniteShape Instance;
};

//gݍ킹`
class I_CombinationShape : public I_Shape
{
protected:
	~I_CombinationShape() {}
public:
	virtual bool Test(const I_Shape &other) const {return other.Test(*this);}
};

//~
class I_Circle : public I_Shape
{
protected:
	~I_Circle() {}
public:
	virtual bool Test(const I_Shape &other) const {return other.Test(*this);}

	virtual bool Test(const I_Circle &other) const;
	virtual bool Test(const I_LineDividedArea &other) const;
	virtual bool Test(const I_Segment &other) const;
	virtual bool Test(const I_CombinationShape &other) const {return other.Test(*this);}

	struct Data
	{
		Point center;
		Number_Type radius;
		Data(const Point &c,Number_Type r) : center(c),radius(r) {}
	};
//~̏
	virtual Data GetCircleData(void) const = 0;
};

//ŕꂽ̈
class I_LineDividedArea : public I_Shape
{
protected:
	~I_LineDividedArea() {}
public:
	virtual bool Test(const I_Shape &other) const {return other.Test(*this);}

	virtual bool Test(const I_Circle &other) const;
	virtual bool Test(const I_LineDividedArea &other) const;
	virtual bool Test(const I_Segment &other) const;
	virtual bool Test(const I_CombinationShape &other) const {return other.Test(*this);}

	struct Data
	{
		Number_Type a,b,c;
		bool less;
		Data(Number_Type a,Number_Type b,Number_Type c,bool less)
			: a(a),b(b),c(c),less(less) {}
	};
//̏
	virtual Data GetLineDividedAreaData(void) const = 0;
//Q__ԂĂꍇ͑ǂ0div
};

//iJvZj
class I_Segment : public I_Shape
{
protected:
	~I_Segment() {}
public:
	virtual bool Test(const I_Shape &other) const {return other.Test(*this);}

	virtual bool Test(const I_Circle &other) const;
	virtual bool Test(const I_LineDividedArea &other) const;
	virtual bool Test(const I_Segment &other) const;
	virtual bool Test(const I_CombinationShape &other) const {return other.Test(*this);}

	struct Data
	{
		Point edge1,edge2;
		Number_Type half_width;	//i̓񕪂̈̒lj
		Data(const Point &e1,const Point &e2,Number_Type hw)
			: edge1(e1),edge2(e2),half_width(hw) {}
	};
//̏
	virtual Data GetSegmentData(void) const = 0;
};



//	ėpaW
class I_UnitedShape : public I_CombinationShape
{
protected:
	I_UnitedShape() {}
public:
	virtual bool Test(const I_Circle &other) const;
	virtual bool Test(const I_LineDividedArea &other) const;
	virtual bool Test(const I_Segment &other) const;
	virtual bool Test(const I_CombinationShape &other) const;

//ĂI_Shape̐
	virtual int GetPartsNum(void) const = 0;
//ĂIShapë
	virtual const I_Shape &GetPart(int index) const = 0;
	//Ԓl̗ĹAGetPart()ĂяoŎĂ\ȂB
};


}//	namespace Intersection
}//	namespace D2
}//	namespace Juna

#endif

