
#ifndef Juna_BlockHeap_Hpp
#define Juna_BlockHeap_Hpp

#include <stddef.h>

namespace Juna {

class BlockHeap
{
public:
	BlockHeap(void) : Heap(0),Next(0),BlockSize(0),Capacity(0),AllocNum(0) {}
	BlockHeap(size_t blocksize,size_t capacity,size_t block_align = 4)	//2^n
		: Heap(0) {Initialize(blocksize + (block_align-1) & ~(block_align-1),capacity);}

	~BlockHeap() {operator delete(Heap);}

	//TCYύXčď
	void Initialize(size_t blocksize,size_t capacity)
	{
		operator delete(Heap);
		Heap = operator new(blocksize * capacity);
		BlockSize = blocksize;
		Capacity = capacity;
		Reset();
	}
	//TCYύXȂōď
	void Reset(void)
	{
		Next = Heap;
		for (size_t i = 0;i < Capacity-1;i++)
		{
			void **head = static_cast<void **>(add_address(Heap,i * BlockSize));
			*head = add_address(Heap,(i + 1) * BlockSize);
		}
		void **head = static_cast<void **>(add_address(Heap,(Capacity-1) * BlockSize));
		*head = 0;
		AllocNum = 0;
	}

	void *Allocate(void)
	{
		AllocNum++;
		void *r = Next;
		Next = *static_cast<void **>(Next);
		return r;
	}
	void Deallocate(void *allocated)
	{
		AllocNum--;
		*static_cast<void **>(allocated) = Next;
		Next = allocated;
	}

	bool full(void) const {return !Next;}

	size_t blockSize(void) const {return BlockSize;}
	size_t capacity(void) const {return Capacity;}
	size_t allocNum(void) const {return AllocNum;}
private:
	void *Heap;
	void *Next;

	size_t BlockSize;
	size_t Capacity;
	size_t AllocNum;

	static void *add_address(void *address,ptrdiff_t offset)
		{return static_cast<char *>(address) + offset;}
private:
	BlockHeap(BlockHeap &);
	void operator=(BlockHeap &);
};


}//namespace Juna

#endif
