
#ifndef Juna_Direct2D_Blit_H
#define Juna_Direct2D_Blit_H

#include "Direct2D.h"

namespace Juna {
namespace D2D {

struct FreeBlitPoint
{
	float x,y;
	Color c;

	void Set(float X,float Y,Color argb) {x=X;y=Y;c=argb;}
};
struct LinePoint
{
	float x,y;
	float width;
	Color c;

	void Set(float X,float Y,float Width,Color argb) {x=X;y=Y;width=Width;c=argb;}
};

static const int MaxCircleVertexNum = 128;

class Blitter
{
public:
	static void Blit(I_ScreenSurface &dest,const FRect &destRect,
					 const I_SpriteSurface &src,const FRect &srcRect,Color argb);
	static void Blit(I_ScreenSurface &dest,const FRect &destRect,
					 const I_SpriteSurface &src,const FRect &srcRect,uint8 alpha = 255)
		{Blit(dest,destRect,src,srcRect,Color(alpha,255,255,255));}

	static void Blit(I_ScreenSurface &dest,float x,float y,
					 const I_SpriteSurface &src,const FRect &srcRect,
					 float angle,float x_scale,float y_scale,Color argb);
	static void Blit(I_ScreenSurface &dest,float x,float y,
					 const I_SpriteSurface &src,const FRect &iSrcRect,
					 float angle,float x_scale = 1,float y_scale = 1,uint8 alpha = 255)
		{Blit(dest,x,y,src,iSrcRect,angle,x_scale,y_scale,Color(alpha,255,255,255));}

	static void Fill(I_ScreenSurface &dest,float x,float y,float angle,float width,float height,Color argb);

	static void FreeBlit(I_ScreenSurface &dest,const FreeBlitPoint points[4],
						 const I_SpriteSurface &src,const FRect &srcRect);
	static void FreeFill(I_ScreenSurface &dest,const FreeBlitPoint points[4]);
	static void FreeBlit2(I_ScreenSurface &dest,const FreeBlitPoint points[4],
						  const I_SpriteSurface &src,const FRect &srcRect);
	static void FreeFill2(I_ScreenSurface &dest,const FreeBlitPoint points[4]);
	static void FreeBlitPerspective(I_ScreenSurface &dest,const FreeBlitPoint points[4],
									const I_SpriteSurface &src,const FRect &srcRect);
	static void FreeFillPerspective(I_ScreenSurface &dest,const FreeBlitPoint points[4]);

	static void LineBlit(I_ScreenSurface &dest,const LinePoint *points,int pointnum,
						 const I_SpriteSurface &src,const FRect &srcRect);
	static void LineFill(I_ScreenSurface &dest,const LinePoint *points,int pointnum);

	static void DrawLine(I_ScreenSurface &dest,const FPoint &point1,const FPoint &point2,
						 Color color1,Color color2);


	static void CircleFill(I_ScreenSurface &dest,const FPoint &centerPos,float radius,
						   Color centerColor,Color outsideColor,
						   int vertexnum = MaxCircleVertexNum);

	static void RingFill(I_ScreenSurface &dest,const FPoint &centerPos,
						 float insideRadius,float outsideRadius,
						 Color insideColor,Color outsideColor,
						 int vertexNum = MaxCircleVertexNum);
	static void RingBlit(I_ScreenSurface &dest,const FPoint &centerPos,
						 float insideRadius,float outsideRadius,
						 const I_SpriteSurface &src,const FRect &srcRect,
						 Color insideColor = Color(255,255,255,255),
						 Color outsideColor = Color(255,255,255,255),
						 int vertexNum = MaxCircleVertexNum);

private:
	static void PrivateBlit(I_ScreenSurface &dest,const FRect &destRect,
							IDirect3DTexture8 *texture,
							float left,float top,float right,float bottom,Color argb);

	static void PrivateBlit(I_ScreenSurface &dest,float x,float y,
							IDirect3DTexture8 *texture,
							float left,float top,float right,float bottom,
							float angle,float halfwidth,float halfheight,Color argb);
	static void PrivateFreeBlit(I_ScreenSurface &dest,const FreeBlitPoint points[4],
								IDirect3DTexture8 *texture,
								float left,float top,float right,float bottom);
	static void PrivateFreeBlit2(I_ScreenSurface &dest,const FreeBlitPoint points[4],
								 IDirect3DTexture8 *texture,
								 float left,float top,float right,float bottom);
	static void PrivateFreeBlitPerspective(I_ScreenSurface &dest,const FreeBlitPoint points[4],
										   IDirect3DTexture8 *texture,
										   float left,float top,float right,float bottom);

	static void PrivateLineBlit(I_ScreenSurface &dest,const LinePoint *points,int pointnum,
								IDirect3DTexture8 *texture,
								float left,float top,float right,float bottom);

	static void PrivateRingBlit(I_ScreenSurface &dest,const FPoint &centerPos,
								float insideRadius,float outsideRadius,
								IDirect3DTexture8 *texture,
								float left,float top,float right,float bottom,
								Color insideColor,Color outsideColor,int vertexNum);
};


//=============================================================================
//			I_ScreenSurface Blits inline
//=============================================================================
inline void Blitter::Blit(I_ScreenSurface &dest,const FRect &destRect,
	const I_SpriteSurface &src,const FRect &srcRect,Color argb)
{
	PrivateBlit(dest,destRect,src.GetD3DTexture(),
				srcRect.x1 / src.width(),srcRect.y1 / src.height(),
				srcRect.x2 / src.width(),srcRect.y2 / src.height(),argb);
}

inline void Blitter::Blit(I_ScreenSurface &dest,float x,float y,
	const I_SpriteSurface &src,const FRect &srcRect,
	float angle,float x_scale,float y_scale,Color argb)
{
	PrivateBlit(dest,x,y,src.GetD3DTexture(),
				srcRect.x1 / src.width(),srcRect.y1 / src.height(),
				srcRect.x2 / src.width(),srcRect.y2 / src.height(),
				angle,
				srcRect.width() * 0.5f * x_scale,srcRect.height() * 0.5f * y_scale,
				argb);
}

inline void Blitter::Fill(I_ScreenSurface &dest,float x,float y,float angle,float width,float height,Color argb)
	{PrivateBlit(dest,x,y,0,0,0,1,1,angle,width * 0.5f,height * 0.5f,argb);}


inline void Blitter::FreeBlit(I_ScreenSurface &dest,const FreeBlitPoint points[4],
	const I_SpriteSurface &src,const FRect &srcRect)
{
	PrivateFreeBlit(dest,points,src.GetD3DTexture(),
					srcRect.x1 / src.width(),srcRect.y1 / src.height(),
					srcRect.x2 / src.width(),srcRect.y2 / src.height());
}
inline void Blitter::FreeFill(I_ScreenSurface &dest,const FreeBlitPoint points[4])
	{PrivateFreeBlit(dest,points,0,0,0,1,1);}
inline void Blitter::FreeBlit2(I_ScreenSurface &dest,const FreeBlitPoint points[4],
	const I_SpriteSurface &src,const FRect &srcRect)
{
	PrivateFreeBlit2(dest,points,src.GetD3DTexture(),
					srcRect.x1 / src.width(),srcRect.y1 / src.height(),
					srcRect.x2 / src.width(),srcRect.y2 / src.height());
}
inline void Blitter::FreeFill2(I_ScreenSurface &dest,const FreeBlitPoint points[4])
	{PrivateFreeBlit2(dest,points,0,0,0,1,1);}
inline void Blitter::FreeBlitPerspective(I_ScreenSurface &dest,const FreeBlitPoint points[4],
	const I_SpriteSurface &src,const FRect &srcRect)
{
	PrivateFreeBlitPerspective(dest,points,src.GetD3DTexture(),
					srcRect.x1 / src.width(),srcRect.y1 / src.height(),
					srcRect.x2 / src.width(),srcRect.y2 / src.height());
}
inline void Blitter::FreeFillPerspective(I_ScreenSurface &dest,const FreeBlitPoint points[4])
	{PrivateFreeBlitPerspective(dest,points,0,0,0,1,1);}

inline void Blitter::LineBlit(I_ScreenSurface &dest,const LinePoint *points,
	int pointnum,const I_SpriteSurface &src,const FRect &srcRect)
{
	PrivateLineBlit(dest,points,pointnum,src.GetD3DTexture(),
					srcRect.x1 / src.width(),srcRect.y1 / src.height(),
					srcRect.x2 / src.width(),srcRect.y2 / src.height());
}

inline void Blitter::LineFill(I_ScreenSurface &dest,const LinePoint *points,int pointnum)
	{PrivateLineBlit(dest,points,pointnum,(IDirect3DTexture8 *)0,0,0,1,1);}

inline void Blitter::RingFill(I_ScreenSurface &dest,const FPoint &centerPos,
	float insideRadius,float outsideRadius,
	Color insideColor,Color outsideColor,int vertexNum)
{
	PrivateRingBlit(dest,centerPos,insideRadius,outsideRadius,
					0,0,0,1,1,
					insideColor,outsideColor,vertexNum);
}

inline void Blitter::RingBlit(I_ScreenSurface &dest,const FPoint &centerPos,
	float insideRadius,float outsideRadius,
	const I_SpriteSurface &src,const FRect &srcRect,
	Color insideColor,Color outsideColor,int vertexNum)
{
	PrivateRingBlit(dest,centerPos,insideRadius,outsideRadius,
					src.GetD3DTexture(),
					srcRect.x1 / src.width(),srcRect.y1 / src.height(),
					srcRect.x2 / src.width(),srcRect.y2 / src.height(),
					insideColor,outsideColor,vertexNum);
}


}//namespace D2D
}//namespace Juna

#endif
