
#ifndef Juna_DirectXGraphics2D_H
#define Juna_DirectXGraphics2D_H


#include "../int_def.h"
#include "../2D_Coordinate.h"
#include "../RGB.h"

//	O錾

struct IDirect3D8;
struct IDirect3DDevice8;
struct IDirect3DTexture8;
struct IDirect3DSurface8;
struct _D3DPRESENT_PARAMETERS_;

namespace Juna {
namespace D2D {

typedef D2::Rectangle<int>		Rect;
typedef D2::Rectangle<float>	FRect;
typedef D2::Point<int>			Point;
typedef D2::Point<float>		FPoint;
typedef D2::Size<int>			Size;
typedef D2::Size<float>			FSize;

using Juna::Color::ARGB;
typedef ARGB Color;

enum BlendType {
	BT_ALPHA,BT_ADD,	//
	BT_SRC,BT_ADDFULL,	//At@
	BT_HALFADD,BT_MULTI,BT_NEGA,	//ꍇ
	BT_ADDLIKE,BT_SUBLIKE,BT_LIGHT	//At@
};

enum SurfaceColorFormat {
	SCF_UNKNOWN = 0,
	SCF_ARGB8888,
	SCF_ARGB4444,
	SCF_RGB8888,
	SCF_RGB565,
	SCF_RGB555,
	SCF_RGB888,
};

class Manager;

//ManagerManageC^[tFCX
class I_Managed
{
protected:
	I_Managed(void) {next = this;}
	~I_Managed() {}

	bool isManaged(void) const {return next != this;}
private:
	friend class Manager;
	virtual void Terminalize(void) = 0;

	virtual void Release(void) = 0;
	virtual bool Restore(void) = 0;

	I_Managed *next;
private:
	I_Managed(I_Managed &);
	void operator=(I_Managed &);
};

//	
class I_Surface
{
protected:
	I_Surface(void) : Surface(0),Width(0),Height(0),Manager(0) {}
	~I_Surface(void) {}
public:
	bool isValid(void) const {return Surface != 0;}
	int width(void) const {return Width;}
	int height(void) const {return Height;}
	int depth(void) const;
	SurfaceColorFormat format(void) const;

	Rect rect(void) const {return Rect(0,0,Width,Height);}
	FRect frect(void) const {return FRect(0,0,float(Width),float(Height));}

	bool Copy(int x,int y,const I_Surface &srcSurf,const Rect *srcRect);

protected:
	IDirect3DSurface8 *Surface;
	int Width,Height;
	Manager *Manager;

	bool Initialize(D2D::Manager &manager,int width,int height,bool screen);
	void Terminalize(void);
	void Release(void);
	bool Restore(bool screen);
	void Swap(I_Surface &other);
public:
	IDirect3DSurface8 *GetD3DSurface(void) const {return Surface;}
	const D2D::Manager *GetManager(void) const {return Manager;}

private:
	I_Surface(I_Surface &);
	void operator=(I_Surface &);
};


//Blit摜C^[tFCX
class I_SpriteSurface : public I_Surface
{
protected:
	I_SpriteSurface(void) : Texture(0) {}
	~I_SpriteSurface(void) {}

	IDirect3DTexture8 *Texture;

	enum PoolType {VRAM,MANAGED,SYSMEM};
	bool Initialize(D2D::Manager &manager,int width,int height,int depth,
					PoolType pool,bool screen);
	void Terminalize(void);
	void Release(void);
	bool Restore(int depth,PoolType pool,bool screen);
	void Swap(I_SpriteSurface &other);
public:
	IDirect3DTexture8 *GetD3DTexture(void) const {return Texture;}

};

//Blit摜C^[tFCX
class I_ScreenSurface : public I_Surface
{
protected:
	I_ScreenSurface(void) {}
	~I_ScreenSurface() {}
public:

	void Clear(Color clearARGB = Color(255,0,0,0));
	void SetClipRect(const Rect *clipRect = 0);

	void BeginBlit(void);
	void EndBlit(void);

	void Blit(int x,int y,const I_SpriteSurface &src,const Rect &srcrect,
			  float x_scale = 1.0f,float y_scale = 1.0f,
			  Color color = Color(255,255,255,255));
	void Blit(float center_x,float center_y,
			  const I_SpriteSurface &src,const FRect &srcrect,
			  float angle = 0,float x_scale = 1.0f,float y_scale = 1.0f,
			  Color color = Color(255,255,255,255));

	void Fill(const Rect &rect,Color color);

	void SetBlendType(BlendType bt);
};



//BlitT[tBXi񃍃Xgj
class SpriteSurface : public I_SpriteSurface, private I_Managed
{
public:
	SpriteSurface(void) {}
	~SpriteSurface() {Terminalize();}

	bool Initialize(D2D::Manager &manager,int width,int height,int depth = 0);
	virtual void Terminalize(void);
	void Swap(SpriteSurface &other) {I_SpriteSurface::Swap(other);}

private:
	virtual void Release(void) {}
	virtual bool Restore(void) {return true;}
};


//BlitT[tBXiXgj
class ScreenSurface : public I_ScreenSurface, private I_Managed
{
public:
	ScreenSurface(void) {}
	~ScreenSurface() {Terminalize();}

	bool Initialize(D2D::Manager &manager,int width,int height);
	virtual void Terminalize(void);
	void Swap(ScreenSurface &other) {I_ScreenSurface::Swap(other);}

private:
	virtual void Release(void);
	virtual bool Restore(void);
};

class ImageSurface : public I_Surface, private I_Managed
{
public:
	ImageSurface(void) {}
	~ImageSurface() {Terminalize();}

	bool Initialize(D2D::Manager &manager,int width,int height);
	virtual void Terminalize(void);
	void Swap(ImageSurface &other) {I_Surface::Swap(other);}

	bool Initialize(I_Surface &src);

private:
	virtual void Release(void) {}
	virtual bool Restore(void) {return true;}
};


//Blit\͂A{IɂCopyT[tBXi񃍃Xgj
class BGSurface : public I_ScreenSurface, private I_Managed
{
public:
	BGSurface(void) {}
	~BGSurface() {Terminalize();}

	bool Initialize(D2D::Manager &manager,int width,int height);
	virtual void Terminalize(void);
	void Swap(BGSurface &other)
		{I_ScreenSurface::Swap(other);Backup.Swap(other.Backup);}

//	Screen=>Backup
	bool SaveImage(const Rect *rect = 0);
//	Backup=>Screen
	bool RestoreImage(const Rect *rect = 0);

	I_Surface &backupSurface(void) {return Backup;}
	const I_Surface &backupSurface(void) const {return Backup;}

private:
	ImageSurface Backup;

	virtual void Release(void);
	virtual bool Restore(void);
};


//ǗIuWFNg
class Manager
{
public:
	Manager(void) : Device(0),BackBuffer(),ManagedHeader(0) {}
	~Manager() {Terminalize();}

	bool Initialize(int width,int height,bool fullscreen = false,
					int depth = 0,int refreshrate = 0,
					bool use_double = false,void *hWindow = 0);
	void Terminalize(void);

	bool Reset(int width,int height,bool fullscreen = false,
			   int depth = 0,int refresh_rate = 0);
	bool ResetScreenMode(bool fullscreen,int depth = 0,int refresh_rate = 0)
		{return Reset(width(),height(),fullscreen,depth,refresh_rate);}

	I_ScreenSurface &backBuffer(void) {return BackBuffer;}
	bool Flip(void);

	int width(void) const {return BackBuffer.width();}
	int height(void) const {return BackBuffer.height();}
	int depth(void) const;
	SurfaceColorFormat format(void) const;

	bool isValid(void) const {return Device != 0;}
	bool isFullScreen(void) const;

	void SetLinearFilter(bool flag);

	void EnterManaged(I_Managed &managed);
	void ExitManaged(I_Managed &managed);

private:
	IDirect3DDevice8 *Device;

	class BackBufferSurface : public I_ScreenSurface {
		friend class Manager;
	} BackBuffer;

	I_Managed *ManagedHeader;

	IDirect3D8 *D3DObject;

//悢q͐^ʖځInewgȂ߂̂
	enum {D3DPRESENT_PARAMETERS_SIZE = 13*4};
//	D3DPRESENT_PARAMETERS D3D_PP;
	char D3D_PP_Buf[D3DPRESENT_PARAMETERS_SIZE];
	int WindowFormat;
public:
	IDirect3DDevice8 *GetD3DDevice(void) const {return Device;}
	void ResetDeviceState(void) const;
	const _D3DPRESENT_PARAMETERS_ *GetD3D_PP(void) const
		{return reinterpret_cast<const _D3DPRESENT_PARAMETERS_ *>(D3D_PP_Buf);}

private:
	friend class BackBufferSurface;
	void ReleaseManaged(void);
	int RestoreManaged(void);
private:
	Manager(Manager &);
	void operator=(Manager &);
};


}//namespace D2D
}//namespace Juna



#endif
