
#include "../D2D_Blit.h"
#include <D3d8.h>
#include <math.h>


namespace {

static const float Pi = 3.1415926535897932384626433832795f;

struct BlitVertex
{
	float x,y,z,rhw;
	DWORD color;
    float tu, tv;

	void Set(float x,float y,DWORD color,float tu,float tv)
	{
		this->x=x;this->y=y;z=0;rhw=1;
		this->color=color;this->tu=tu;this->tv=tv;
	}
};
static const DWORD D3DFVF_BLIT_VERTEX = (D3DFVF_XYZRHW|D3DFVF_DIFFUSE|D3DFVF_TEX1);
static BlitVertex BV[4] =
{
	{0,0,0,1,D3DCOLOR_ARGB(255,255,255,255),0,0},
	{1,0,0,1,D3DCOLOR_ARGB(255,255,255,255),1,0},
	{1,1,0,1,D3DCOLOR_ARGB(255,255,255,255),1,1},
	{0,1,0,1,D3DCOLOR_ARGB(255,255,255,255),0,1}
};

inline void SetBV(float x1,float y1,float tu1,float tv1,DWORD color1,
				  float x2,float y2,float tu2,float tv2,DWORD color2,
				  float x3,float y3,float tu3,float tv3,DWORD color3,
				  float x4,float y4,float tu4,float tv4,DWORD color4)
{
	BV[0].x = x1;BV[0].y = y1;BV[0].color = color1;BV[0].tu = tu1;BV[0].tv = tv1;
	BV[1].x = x2;BV[1].y = y2;BV[1].color = color2;BV[1].tu = tu2;BV[1].tv = tv2;
	BV[2].x = x3;BV[2].y = y3;BV[2].color = color3;BV[2].tu = tu3;BV[2].tv = tv3;
	BV[3].x = x4;BV[3].y = y4;BV[3].color = color4;BV[3].tu = tu4;BV[3].tv = tv4;
}
inline void SetBV(float x1,float y1,float tu1,float tv1,
				  float x2,float y2,float tu2,float tv2,
				  float x3,float y3,float tu3,float tv3,
				  float x4,float y4,float tu4,float tv4,DWORD color)
{
	SetBV(x1,y1,tu1,tv1,color,x2,y2,tu2,tv2,color,
		  x3,y3,tu3,tv3,color,x4,y4,tu4,tv4,color);
}

inline void SetBV(float x1,float y1,float tu1,float tv1,
				  float x2,float y2,float tu2,float tv2,
				  float x3,float y3,float tu3,float tv3,
				  float x4,float y4,float tu4,float tv4)
{
	SetBV(x1,y1,tu1,tv1,x2,y2,tu2,tv2,x3,y3,tu3,tv3,x4,y4,tu4,tv4,
		  D3DCOLOR_ARGB(255,255,255,255));
}

inline D3DCOLOR c2d3c(Juna::D2D::Color c) {return D3DCOLOR_ARGB(c.a,c.r,c.g,c.b);}

}//	anonymous namespace

namespace Juna {
namespace D2D {


void Blitter::PrivateBlit(I_ScreenSurface &dest,
					const FRect &destRect,IDirect3DTexture8 *texture,
					float left,float top,float right,float bottom,Color argb)
{
	IDirect3DDevice8 &device = *dest.GetManager()->GetD3DDevice();
	HRESULT hr;
	hr = device.SetTexture(0,texture);
	hr = device.SetVertexShader(D3DFVF_BLIT_VERTEX);

	SetBV(destRect.x1 - 0.5f,destRect.y1 - 0.5f,left,top,
		  destRect.x2 - 0.5f,destRect.y1 - 0.5f,right,top,
		  destRect.x2 - 0.5f,destRect.y2 - 0.5f,right,bottom,
		  destRect.x1 - 0.5f,destRect.y2 - 0.5f,left,bottom,
		  c2d3c(argb));

	hr = device.DrawPrimitiveUP(D3DPT_TRIANGLEFAN,2,BV,sizeof(BlitVertex));
}


//̂܂܉]`
void Blitter::PrivateBlit(I_ScreenSurface &dest,
				float x,float y,IDirect3DTexture8 *texture,
				float left,float top,float right,float bottom,
				float angle,float halfwidth,float halfheight,Color argb)
{
	IDirect3DDevice8 &device = *dest.GetManager()->GetD3DDevice();
	HRESULT hr;
	hr = device.SetTexture(0,texture);
	hr = device.SetVertexShader(D3DFVF_BLIT_VERTEX);

	const float cosa = ::cosf(angle);
	const float sina = ::sinf(angle);
	const float xcos = cosa * halfwidth;
	const float xsin = sina * halfwidth;
	const float ycos = cosa * halfheight;
	const float ysin = sina * halfheight;

	const float fx = x - 0.5f;
	const float fy = y - 0.5f;

	SetBV(fx - xcos + ysin,fy - xsin - ycos,left,top,
		  fx + xcos + ysin,fy + xsin - ycos,right,top,
		  fx + xcos - ysin,fy + xsin + ycos,right,bottom,
		  fx - xcos - ysin,fy - xsin + ycos,left,bottom,
		  c2d3c(argb));

	hr = device.DrawPrimitiveUP(D3DPT_TRIANGLEFAN,2,BV,sizeof(BlitVertex));
}



void Blitter::PrivateFreeBlit(I_ScreenSurface &dest,
			const FreeBlitPoint points[4],
			IDirect3DTexture8 *texture,
			float left,float top,float right,float bottom)
{
	IDirect3DDevice8 &device = *dest.GetManager()->GetD3DDevice();
	HRESULT hr;
	hr = device.SetTexture(0,texture);
	hr = device.SetVertexShader(D3DFVF_BLIT_VERTEX);

	BlitVertex bv[4] = {
		{points[0].x - 0.5f,points[0].y - 0.5f,0,1,c2d3c(points[0].c),left,top},
		{points[1].x - 0.5f,points[1].y - 0.5f,0,1,c2d3c(points[1].c),right,top},
		{points[2].x - 0.5f,points[2].y - 0.5f,0,1,c2d3c(points[2].c),right,bottom},
		{points[3].x - 0.5f,points[3].y - 0.5f,0,1,c2d3c(points[3].c),left,bottom},
	};
	hr = device.DrawPrimitiveUP(D3DPT_TRIANGLEFAN,2,bv,sizeof(BlitVertex));
}
void Blitter::PrivateFreeBlit2(I_ScreenSurface &dest,
			const FreeBlitPoint points[4],
			IDirect3DTexture8 *texture,
			float left,float top,float right,float bottom)
{
	IDirect3DDevice8 &device = *dest.GetManager()->GetD3DDevice();
	HRESULT hr;
	hr = device.SetTexture(0,texture);
	hr = device.SetVertexShader(D3DFVF_BLIT_VERTEX);
/*
	FPoint lcp,tcp,rcp,bcp;
	lcp.Set((points[3].x + points[0].x) / 2,(points[3].y + points[0].y) / 2);
	tcp.Set((points[0].x + points[1].x) / 2,(points[0].y + points[1].y) / 2);
	rcp.Set((points[1].x + points[2].x) / 2,(points[1].y + points[2].y) / 2);
	bcp.Set((points[2].x + points[3].x) / 2,(points[2].y + points[3].y) / 2);
	Color lc,tc,rc,bc;
	lc = Ave(points[3].c,points[0].c);
	tc = Ave(points[0].c,points[1].c);
	rc = Ave(points[1].c,points[2].c);
	bc = Ave(points[2].c,points[3].c);

	{
		BlitVertex bv[5] = {
			{points[0].x - 0.5f,points[0].y - 0.5f,0,1,points[0].c.argb,left,top},
			{tcp.x - 0.5f,tcp.y - 0.5f,0,1,tc.argb,(left + right) / 2,top},
			{lcp.x - 0.5f,lcp.y - 0.5f,0,1,lc.argb,left,(top + bottom) / 2},
			{bcp.x - 0.5f,bcp.y - 0.5f,0,1,bc.argb,(left + right) / 2,bottom},
			{points[3].x - 0.5f,points[3].y - 0.5f,0,1,points[3].c.argb,left,bottom},
		};
		hr = device.DrawPrimitiveUP(D3DPT_TRIANGLESTRIP,3,bv,sizeof(BlitVertex));
	}
	{
		BlitVertex bv[5] = {
			{points[1].x - 0.5f,points[1].y - 0.5f,0,1,points[1].c.argb,right,top},
			{tcp.x - 0.5f,tcp.y - 0.5f,0,1,tc.argb,(left + right) / 2,top},
			{rcp.x - 0.5f,rcp.y - 0.5f,0,1,rc.argb,right,(top + bottom) / 2},
			{bcp.x - 0.5f,bcp.y - 0.5f,0,1,bc.argb,(left + right) / 2,bottom},
			{points[2].x - 0.5f,points[2].y - 0.5f,0,1,points[2].c.argb,right,bottom},
		};
		hr = device.DrawPrimitiveUP(D3DPT_TRIANGLESTRIP,3,bv,sizeof(BlitVertex));
	}
/*/
	float cx = (points[0].x + points[1].x + points[2].x + points[3].x) / 4;
	float cy = (points[0].y + points[1].y + points[2].y + points[3].y) / 4;
	int a = (points[0].c.a + points[1].c.a + points[2].c.a + points[3].c.a) / 4;
	int r = (points[0].c.r + points[1].c.r + points[2].c.r + points[3].c.r) / 4;
	int g = (points[0].c.g + points[1].c.g + points[2].c.g + points[3].c.g) / 4;
	int b = (points[0].c.b + points[1].c.b + points[2].c.b + points[3].c.b) / 4;
	BlitVertex bv[6] = {
		{cx - 0.5f,cy - 0.5f,0,1,D3DCOLOR_ARGB(a,r,g,b),(left + right) / 2,(top + bottom) / 2},
		{points[0].x - 0.5f,points[0].y - 0.5f,0,1,c2d3c(points[0].c),left,top},
		{points[1].x - 0.5f,points[1].y - 0.5f,0,1,c2d3c(points[1].c),right,top},
		{points[2].x - 0.5f,points[2].y - 0.5f,0,1,c2d3c(points[2].c),right,bottom},
		{points[3].x - 0.5f,points[3].y - 0.5f,0,1,c2d3c(points[3].c),left,bottom},
		{points[0].x - 0.5f,points[0].y - 0.5f,0,1,c2d3c(points[0].c),left,top},
	};
	hr = device.DrawPrimitiveUP(D3DPT_TRIANGLEFAN,4,bv,sizeof(BlitVertex));

//*/
}

void Blitter::PrivateFreeBlitPerspective(I_ScreenSurface &dest,
				const FreeBlitPoint points[4],
				IDirect3DTexture8 *texture,
				float left,float top,float right,float bottom)
{
	IDirect3DDevice8 &device = *dest.GetManager()->GetD3DDevice();
	HRESULT hr;
	hr = device.SetTexture(0,texture);
	hr = device.SetVertexShader(D3DFVF_BLIT_VERTEX);

	struct local {
		static inline float len2(float dx,float dy) {return dx*dx + dy*dy;}
	};
	float lr,tb;
	lr = ::sqrtf(
			(local::len2(points[3].x - points[0].x,points[3].y - points[0].y)) /
			(local::len2(points[1].x - points[2].x,points[1].y - points[2].y))
			);
	tb = ::sqrtf(
			(local::len2(points[0].x - points[1].x,points[0].y - points[1].y)) /
			(local::len2(points[2].x - points[3].x,points[2].y - points[3].y))
			);
	BlitVertex bv[4] = {
		{points[0].x - 0.5f,points[0].y - 0.5f,0,lr * tb,c2d3c(points[0].c),left,top},
		{points[1].x - 0.5f,points[1].y - 0.5f,0,1 * tb,c2d3c(points[1].c),right,top},
		{points[2].x - 0.5f,points[2].y - 0.5f,0,1 * 1,c2d3c(points[2].c),right,bottom},
		{points[3].x - 0.5f,points[3].y - 0.5f,0,lr * 1,c2d3c(points[3].c),left,bottom},
	};
	hr = device.DrawPrimitiveUP(D3DPT_TRIANGLEFAN,2,bv,sizeof(BlitVertex));
}



namespace {
inline void SetPLVertex(BlitVertex *lbv,float x,float y,
						float angle,float width,
						DWORD color,float tu,float bottom,float top)
{
		float c = ::cosf(angle) * width;
		float s = ::sinf(angle) * width;

		lbv[0].x = x + c;
		lbv[0].y = y + s;
		lbv[1].x = x - c;
		lbv[1].y = y - s;
		lbv[0].color = lbv[1].color = color;

		lbv[0].tu = lbv[1].tu = tu;
		lbv[0].tv = bottom;lbv[1].tv = top;
}
class PolyLineWork
{
public:
	PolyLineWork(void) : LineBV(0),PointNum(0) {}
	~PolyLineWork() {delete[] LineBV;}

	BlitVertex *LineBV;
	int PointNum;
};

PolyLineWork PLWork;

}//namespace ""

void Blitter::PrivateLineBlit(I_ScreenSurface &dest,
						const LinePoint *iPoints,int pointnum,
						IDirect3DTexture8 *texture,
						float left,float top,float right,float bottom)
{
	if (PLWork.PointNum < pointnum)
	{
		delete[] PLWork.LineBV;
		PLWork.LineBV = new BlitVertex[pointnum * 2];
		PLWork.PointNum = pointnum;
		for (int i = 0;i < pointnum * 2;i++)
		{
			PLWork.LineBV[i].z = 0;
			PLWork.LineBV[i].rhw = 1;
		}
	}

	float pangle;
	{
		float w = iPoints[0].width / 2;
		pangle = float(::atan2(iPoints[1].y - iPoints[0].y,iPoints[1].x - iPoints[0].x)) + Pi/2;

		SetPLVertex(PLWork.LineBV + 0,iPoints[0].x,iPoints[0].y,
					pangle,w,
					c2d3c(iPoints[0].c),
					left,bottom,top);
	}
	for (int i = 1,j = 2;i < pointnum - 1;i++,j+=2)
	{
		float angle = pangle;
		pangle = float(::atan2(iPoints[i+1].y - iPoints[i].y,iPoints[i+1].x - iPoints[i].x)) + Pi/2;
		angle = (angle + pangle) / 2;

		SetPLVertex(PLWork.LineBV + j,iPoints[i].x,iPoints[i].y,
					angle,iPoints[i].width / 2,
					c2d3c(iPoints[i].c),
					left + ((right - left) * i / (pointnum - 1)),bottom,top);
	}
	{
		int i = pointnum-1;
		int j = i * 2;

		SetPLVertex(PLWork.LineBV + j,iPoints[i].x,iPoints[i].y,
					pangle,iPoints[i].width / 2,
					c2d3c(iPoints[i].c),
					right,bottom,top);
	}

	IDirect3DDevice8 &device = *dest.GetManager()->GetD3DDevice();
	HRESULT hr;
	hr = device.SetTexture(0,texture);
	hr = device.SetVertexShader(D3DFVF_BLIT_VERTEX);
	hr = device.DrawPrimitiveUP(D3DPT_TRIANGLESTRIP,(pointnum-1)*2,
								PLWork.LineBV,sizeof(BlitVertex));
}

void Blitter::DrawLine(I_ScreenSurface &dest,
							  const FPoint &point1,const FPoint &point2,
							  Color color1,Color color2)
{
	struct LineVertex
	{
		float x,y,z,rhw;
		DWORD color;
	} lv[2] =
	{
		{point1.x - 0.5f,point1.y - 0.5f,0,1,c2d3c(color1)},
		{point2.x - 0.5f,point2.y - 0.5f,0,1,c2d3c(color2)},
	};

	IDirect3DDevice8 &device = *dest.GetManager()->GetD3DDevice();
	HRESULT hr;
	hr = device.SetTexture(0,NULL);
	hr = device.SetVertexShader(D3DFVF_XYZRHW | D3DFVF_DIFFUSE);
	hr = device.DrawPrimitiveUP(D3DPT_LINELIST,2,lv,sizeof(LineVertex));
}


void Blitter::CircleFill(I_ScreenSurface &dest,
					const FPoint &centerPos,float radius,
					Color centerColor,Color outsideColor,
					int vertexNum)
{
	if (vertexNum > MaxCircleVertexNum)
		vertexNum = MaxCircleVertexNum;
	IDirect3DDevice8 &device = *dest.GetManager()->GetD3DDevice();
	HRESULT hr;
	hr = device.SetTexture(0,NULL);
	hr = device.SetVertexShader(D3DFVF_BLIT_VERTEX);

	BlitVertex bv[MaxCircleVertexNum + 2];
	const D3DCOLOR oc = c2d3c(outsideColor);
	bv[0].Set(centerPos.x,centerPos.y,c2d3c(centerColor),0,0);
	for (int i = 0;i <= vertexNum;i++)
	{
		float angle = Pi * 2 / vertexNum * i;
		bv[1 + i].Set(centerPos.x + ::cosf(angle)*radius,
					  centerPos.y + ::sinf(angle)*radius,
					  oc,0,0);
	}

	hr = device.DrawPrimitiveUP(D3DPT_TRIANGLEFAN,vertexNum,bv,sizeof(BlitVertex));
}

void Blitter::PrivateRingBlit(I_ScreenSurface &dest,
					const FPoint &centerPos,
					float insideRadius,float outsideRadius,
					IDirect3DTexture8 *texture,
					float left,float top,float right,float bottom,
					Color insideColor,Color outsideColor,int vertexNum)
{
	if (vertexNum > MaxCircleVertexNum)
		vertexNum = MaxCircleVertexNum;
	IDirect3DDevice8 &device = *dest.GetManager()->GetD3DDevice();
	HRESULT hr;
	hr = device.SetTexture(0,texture);
	hr = device.SetVertexShader(D3DFVF_BLIT_VERTEX);

	BlitVertex bv[MaxCircleVertexNum * 2 + 2];
	const D3DCOLOR ic = c2d3c(insideColor),oc = c2d3c(outsideColor);
	for (int i = 0;i < vertexNum;i++)
	{
		float angle = Pi * 2 / vertexNum * i;
		float tu  =left + (right - left) / vertexNum * i;
		bv[i*2].Set(centerPos.x + ::cosf(angle)*insideRadius,
					centerPos.y + ::sinf(angle)*insideRadius,
					ic,tu,bottom);

		bv[i*2+1].Set(centerPos.x + ::cosf(angle)*outsideRadius,
					  centerPos.y + ::sinf(angle)*outsideRadius,
					  oc,tu,top);
	}
	bv[vertexNum*2].Set(centerPos.x + 1*insideRadius,
						centerPos.y + 0*insideRadius,
						ic,right,bottom);
	bv[vertexNum*2+1].Set(centerPos.x + 1*outsideRadius,
						  centerPos.y + 0*outsideRadius,
						  oc,right,top);

	hr = device.DrawPrimitiveUP(D3DPT_TRIANGLESTRIP,vertexNum*2,bv,sizeof(BlitVertex));
}

}//namespace D2D
}//namespace Juna

