
#include "../D2D_Font.h"
#include <windows.h>
#include <d3d8.h>

namespace Juna {
namespace D2D {

FontWriter::FontWriter(void)
	: hFont(0)
{}

FontWriter::~FontWriter()
{
	if (hFont)
	{
		::DeleteObject(hFont);
	}
}

bool FontWriter::SetFont(const char *facename,int height,int width)
{
	HFONT hfont =
		::CreateFontA(height,width,0,0,0,
					  FALSE,FALSE,FALSE,
					  SHIFTJIS_CHARSET,OUT_DEFAULT_PRECIS,CLIP_DEFAULT_PRECIS,
					  DEFAULT_QUALITY,FIXED_PITCH | FF_DONTCARE,
					  facename);
	if (!hfont)
		return false;
	if (hFont)
		::DeleteObject(hFont);
	hFont = hfont;

	return true;
}

namespace {
//VtgJIS leading byte 
inline bool IsSJIS_LEAD(unsigned short c)
{
	return (bool)
		   (((unsigned char)(c) >= 0x81) & ((unsigned char)(c) <= 0x9F)) |
		   (((unsigned char)(c) >= 0xE0) & ((unsigned char)(c) <= 0xFC));
}
//VtgJIS trailing byte 
//قƂǎg񂯂
inline bool IsSJIS_TRAIL(unsigned short c)
{
	return (bool)(((unsigned char)(c) >= 0x40) &
				  ((unsigned char)(c) <= 0xFC) &
				  ((unsigned char)(c) != 0x7F));
}

void *Lock(I_SpriteSurface &surface,int &pitch)
{
	D3DLOCKED_RECT lr;
	if FAILED(surface.GetD3DTexture()->LockRect(0,&lr,NULL,D3DLOCK_NO_DIRTY_UPDATE))
		return 0;
	pitch = lr.Pitch;
	return lr.pBits;
}

void Unlock(I_SpriteSurface &surface,const RECT *rect)
{
	surface.GetD3DTexture()->UnlockRect(0);
	surface.GetD3DTexture()->AddDirtyRect(rect);
}

void DrawCharAlpha32(SpriteSurface &sprite,int dx,int dy,int width,int height,
					 const BYTE *buffer,int pitch,int quality)
{
	const BYTE *bp = buffer;
	int vpitch;
	DWORD *vram = static_cast<DWORD *>(Lock(sprite,vpitch));
	vpitch /= 4;
	vram += dy * vpitch;
	for (int i = 0;i < height;i++)
	{
		for (int j = 0;j < width;j++)
		{
			BYTE *argb = (BYTE *)&vram[dx + j];
			argb[3] = BYTE((bp[j] << (8 - (quality << 1))) - (bp[j] >> (quality << 1)));
		}
		bp += pitch;
		vram += vpitch;
	}
	RECT rect = {dx,dy,dx + width,dy + height};
	Unlock(sprite,&rect);
}


void DrawCharAlpha16(SpriteSurface &sprite,int dx,int dy,int width,int height,
					 const BYTE *buffer,int pitch,int quality)
{
	const BYTE *bp = buffer;
	int vpitch;
	WORD *vram = static_cast<WORD *>(Lock(sprite,vpitch));
	vpitch /= 2;
	vram += dy * vpitch;
	for (int i = 0;i < height;i++)
	{
		for (int j = 0;j < width;j++)
		{
			int a4bit = (bp[j] << (4 - (quality << 1))) - (bp[j] >> (quality << 1));
			vram[dx + j] = WORD((vram[dx + j] & 0x0FFF) | ((a4bit << 12) & 0xF000));
		}
		bp += pitch;
		vram += vpitch;
	}
	RECT rect = {dx,dy,dx + width,dy + height};
	Unlock(sprite,&rect);
}

class OutlineBuffer
{
public:
	OutlineBuffer(int DefCharSize)
	{
		BufSize = DefCharSize*DefCharSize;
		Buffer = new BYTE[BufSize];
		hDC = ::CreateCompatibleDC(NULL);
	}
	~OutlineBuffer()
	{
		if (!hDC)
			::DeleteDC(hDC);
		delete[] Buffer;
	}

	bool GetFontMetrics(HFONT hfont,TEXTMETRIC *tm)
	{
		HFONT oldfont = (HFONT)::SelectObject(hDC,hfont);
		int r = ::GetTextMetrics(hDC,tm);
		::SelectObject(hDC,oldfont);
		return r != 0;
	}

	DWORD GetCharBuf(unsigned short c,HFONT hfont,GLYPHMETRICS *gm,unsigned quality = 1);

	void DrawCharAlpha(SpriteSurface &surface,int dx,int dy,int width,int height,int pitch)
	{
		switch (surface.depth())
		{
		case 16:
			DrawCharAlpha16(surface,dx,dy,width,height,Buffer,pitch,Quality);
			break;
		case 32:
			DrawCharAlpha32(surface,dx,dy,width,height,Buffer,pitch,Quality);
			break;
		}
	}/*
	void DrawCharAlphaColor(DWORD **vram,int dx,int dy,int width,int height,int pitch,DWORD color)
	{
		int i,j;
		const BYTE *bp = Buffer;
		for (i = 0;i < height;i++)
		{
			for (j = 0;j < width;j++)
			{
				vram[i + dy][j + dx] = (((bp[j] << (8 - (Quality << 1))) - (bp[j] >> (Quality << 1))) << 24) | color;
			}
			bp += pitch;
		}
	}

	void DrawCharColor(DWORD **vram,int dx,int dy,int width,int height,int pitch,DWORD color)
	{
		int i,j;
		const BYTE *bp = Buffer;
		const BYTE *col = (BYTE *)&color;
		for (i = 0;i < height;i++)
		{
			for (j = 0;j < width;j++)
			{
				BYTE *dargb = (BYTE *)&vram[i + dy][j + dx];
				int rate = bp[j] << (8 - (Quality << 1));

				vram[i + dy][j + dx] =(
					((dargb[2] * (256 - rate) >> 8) << 16) |
					((dargb[1] * (256 - rate) >> 8) << 8) |
					(dargb[0] * (256 - rate) >> 8) )+(
					((col[2] * rate >> 8) << 16) |
					((col[1] * rate >> 8) << 8) |
					(col[0] * rate >> 8) );
			}
			bp += pitch;
		}
	}
*/
	BYTE *Buffer;
	DWORD BufSize;
	int Quality;

	HDC hDC;
};
OutlineBuffer OutlineBuf(32);

}//namespace ""


void FontWriter::Write(SpriteSurface &surface,int x,int y,const char *string)
{
	int quality = 2;
	if (quality < 1) quality = 1;
	if (quality > 2)
	{
		quality = 2 + (surface.depth() == 32);
	}

	HFONT hfont = static_cast<HFONT>(hFont);
	GLYPHMETRICS gm;
	TEXTMETRIC tm;
	DWORD size;

	OutlineBuf.GetFontMetrics(hfont,&tm);

	int dx,dy;
	dx = x;dy = y;
	const char *c;
	for (c = string;*c != '\0';c++)
	{
		unsigned short sc;
		if (*c == '\n')	//ꉞsɂΉĂ݂
		{
			dx = x;
			dy += tm.tmHeight;
			continue;
		}
		else if (IsSJIS_LEAD(*c))	//[fBOoCg
		{
			sc = ((unsigned char)c[0] << 8) | (unsigned char)c[1];
			c++;
		}
		else
			sc = *c;

		size = OutlineBuf.GetCharBuf(sc,hfont,&gm,quality);

		//͂ݏo
		if (dx + gm.gmptGlyphOrigin.x + gm.gmBlackBoxX > (unsigned)surface.width() ||
			dy + tm.tmAscent - gm.gmptGlyphOrigin.y + gm.gmBlackBoxY > (unsigned)surface.height())
			break;
		if (size == 0)	//`oȂ
		{
			dx += gm.gmCellIncX;
			continue;
		}


		OutlineBuf.DrawCharAlpha(surface,
						dx + gm.gmptGlyphOrigin.x,
						dy + tm.tmAscent - gm.gmptGlyphOrigin.y,
						gm.gmBlackBoxX,gm.gmBlackBoxY,
						(gm.gmBlackBoxX + 3) & ~3);

		dx += gm.gmCellIncX;
	}
}

namespace {
DWORD OutlineBuffer::GetCharBuf(unsigned short c,HFONT hfont,GLYPHMETRICS *gm,unsigned quality)
{
	MAT2 mat2 = {0};
	mat2.eM11.value = 1;
	mat2.eM22.value = 1;


	UINT uFormat;
	switch (quality)
	{
	case 0:
	case 1:
		uFormat = GGO_GRAY2_BITMAP;Quality = 1;break;
	case 2:
		uFormat = GGO_GRAY4_BITMAP;Quality = 2;break;
	case 3:default:
		uFormat = GGO_GRAY8_BITMAP;Quality = 3;break;
	}
	DWORD size;
	HFONT oldfont = (HFONT)::SelectObject(hDC,hfont);
	size = ::GetGlyphOutline(hDC,c,uFormat,gm,0,NULL,&mat2);
	if (size == GDI_ERROR)
	{
		gm->gmCellIncX = 0;
/*
		DWORD er = GetLastError();
		LPVOID lpMsgBuf;
		er = FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER | 
					  FORMAT_MESSAGE_FROM_SYSTEM | 
					  FORMAT_MESSAGE_IGNORE_INSERTS,
					  NULL,er,
					  MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), // ftHg
					  (LPTSTR) &lpMsgBuf,0,NULL);
		LocalFree(lpMsgBuf);
*/
		::SelectObject(hDC,oldfont);
		return 0;
	}
	if (OutlineBuf.BufSize < size)
	{
		delete[] Buffer;
		Buffer = new BYTE[size];
		BufSize = size;
	}
	::GetGlyphOutline(hDC,c,uFormat,gm,size,Buffer,&mat2);

	::SelectObject(hDC,oldfont);
	return size;
}
}//namespace ""

}//namespace D2D
}//namespace Juna
