
#include "../D2D_Picture.h"
#include "../../RGB.h"
#include <d3d8.h>

namespace Juna {
using namespace Color;
namespace D2D {

namespace {
struct PictColorAdapter : public Picture::Color
{
	typedef Picture::Color base;
	uint8 a(void) const {return base::a;}
	uint8 r(void) const {return base::r;}
	uint8 g(void) const {return base::g;}
	uint8 b(void) const {return base::b;}
	void a(uint8 v) {base::a = v;}
	void r(uint8 v) {base::r = v;}
	void g(uint8 v) {base::g = v;}
	void b(uint8 v) {base::b = v;}
};
template<class DestT>
void CopyRect(DestT *dest,Picture::I_Image &src,int dpitch,int width,int height)
{
	PictColorAdapter *s = static_cast<PictColorAdapter *>(src[0]);
	for (int i = 0;i < height;i++)
	{
		for (int j = 0;j < width;j++)
			RGB_Convert(dest[j],s[j]);
		dest = reinterpret_cast<DestT *>(reinterpret_cast<char *>(dest) + dpitch);
		s += src.distance();
	}
}
template<class SrcT>
void CopyRect(Picture::I_Image &dest,SrcT *src,int spitch,int width,int height)
{
	PictColorAdapter *d = static_cast<PictColorAdapter *>(dest[0]);
	for (int i = 0;i < height;i++)
	{
		for (int j = 0;j < width;j++)
			RGB_Convert(d[j],src[j]);
		src = reinterpret_cast<SrcT *>(reinterpret_cast<char *>(src) + spitch);
		d += dest.distance();
	}
}

void *Lock(I_Surface &surface,int &pitch)
{
	D3DLOCKED_RECT lr;
	if FAILED(surface.GetD3DSurface()->LockRect(&lr,NULL,0))
		return 0;

	pitch = lr.Pitch;
	return lr.pBits;
}

void Unlock(I_Surface &surface)
{
	surface.GetD3DSurface()->UnlockRect();
}

}//	anonymous namespace

bool LoadPictureImage(I_SpriteSurface &sprite,Juna::Picture::I_Image &image)
{
	if (LoadPictureImage(static_cast<I_Surface &>(sprite),image))
	{
		sprite.GetD3DTexture()->AddDirtyRect(0);
		return true;
	}
	return false;
}


bool LoadPictureImage(BGSurface &bg,Juna::Picture::I_Image &image)
{
	if (LoadPictureImage(bg.backupSurface(),image))
	{
		bg.RestoreImage();
		return true;
	}
	return true;
}


bool LoadPictureImage(I_Surface &surface,Juna::Picture::I_Image &image)
{
	if (!surface.isValid() || !image.isValid())
		return false;
	int width = min(surface.width(),image.width());
	int height = min(surface.height(),image.height());

	I_Surface *s = &surface;
	ImageSurface tmp;
	void *vram;
	int pitch;
	vram = Lock(surface,pitch);
	if (!vram)
	{
		if (!tmp.Initialize(surface))
			return false;
		vram = Lock(tmp,pitch);
		if (!vram)
			return false;
		s = &tmp;
	}
	switch (s->format())
	{
	case SCF_ARGB8888:
		CopyRect(static_cast<ARGB8888 *>(vram),image,pitch,width,height);
		break;
	case SCF_ARGB4444:
		CopyRect(static_cast<ARGB4444 *>(vram),image,pitch,width,height);
		break;
	case SCF_RGB8888:
		CopyRect(static_cast<RGB8888 *>(vram),image,pitch,width,height);
		break;
	case SCF_RGB555:
		CopyRect(static_cast<RGB555 *>(vram),image,pitch,width,height);
		break;
	case SCF_RGB565:
		CopyRect(static_cast<RGB565 *>(vram),image,pitch,width,height);
		break;
	case SCF_RGB888:
		CopyRect(static_cast<RGB888 *>(vram),image,pitch,width,height);
		break;
	default:
		Unlock(*s);
		return false;
	}

	Unlock(*s);
	if (s != &surface)
	{
		if (!surface.Copy(0,0,*s,0))
			return false;
	}

	return true;
}

bool SavePictureImage(Picture::Image &image,I_Surface &surface)
{
	if (!surface.isValid() || !image.isValid())
		return false;
	int width = surface.width();
	int height =surface.height();
	image.Initialize(width,height);

	I_Surface *s = &surface;
	ImageSurface tmp;
	void *vram;
	int pitch;
	vram = Lock(surface,pitch);
	if (!vram)
	{
		if (!tmp.Initialize(surface))
			return false;
		if (!tmp.Copy(0,0,surface,0))
			return false;
		vram = Lock(tmp,pitch);
		if (!vram)
			return false;
		s = &tmp;
	}

	switch (s->format())
	{
	case SCF_ARGB8888:
		CopyRect(image,static_cast<ARGB8888 *>(vram),pitch,width,height);
		break;
	case SCF_ARGB4444:
		CopyRect(image,static_cast<ARGB4444 *>(vram),pitch,width,height);
		break;
	case SCF_RGB8888:
		CopyRect(image,static_cast<RGB8888 *>(vram),pitch,width,height);
		break;
	case SCF_RGB555:
		CopyRect(image,static_cast<RGB555 *>(vram),pitch,width,height);
		break;
	case SCF_RGB565:
		CopyRect(image,static_cast<RGB565 *>(vram),pitch,width,height);
		break;
	case SCF_RGB888:
		CopyRect(image,static_cast<RGB888 *>(vram),pitch,width,height);
		break;
	default:
		Unlock(*s);
		return false;
	}

	Unlock(*s);

	return true;
}


}//namespace D2D
}//namespace Juna
