
#include "../Direct2D.h"
#include "../../RGB.h"

#include <d3d8.h>
#pragma comment(lib,"d3d8.lib")

#include <math.h>
#include <assert.h>


namespace Juna {
namespace D2D {
using namespace Color;

namespace {

void AdjustWindowRect(HWND hwnd,int width,int height)
{
	RECT wrect,crect;
	::GetWindowRect(hwnd,&wrect);
	::GetClientRect(hwnd,&crect);
	int ncwidth,ncheight;
	ncwidth = (wrect.right - wrect.left) - (crect.right - crect.left);
	ncheight = (wrect.bottom - wrect.top) - (crect.bottom - crect.top);

	::SetWindowPos(hwnd,NULL,0,0,
				   width + ncwidth,height + ncheight,
				   SWP_NOACTIVATE | SWP_NOZORDER | SWP_NOMOVE);

	::GetWindowRect(hwnd,&wrect);
	::GetClientRect(hwnd,&crect);
	int ncheight2 = (wrect.bottom - wrect.top) - (crect.bottom - crect.top);
	if (ncheight != ncheight2)
	{
		::SetWindowPos(hwnd,NULL,0,0,
					   width + ncwidth,height + ncheight2,
					   SWP_NOACTIVATE | SWP_NOZORDER | SWP_NOMOVE);
		::GetWindowRect(hwnd,&wrect);
	}
	RECT drect;
	::GetClientRect(::GetDesktopWindow(),&drect);
	int x = wrect.left - (wrect.right > drect.right ? wrect.right - drect.right : 0);
	int y = wrect.top - (wrect.bottom > drect.bottom ? wrect.bottom - drect.bottom : 0);
	x = x < 0 ? 0 : x;
	y = y < 0 ? 0 : y;
	::SetWindowPos(hwnd,NULL,x,y,0,0,
				   SWP_NOACTIVATE | SWP_NOZORDER | SWP_NOSIZE);
}

int TranslateFormatDepth(D3DFORMAT format)
{
	switch(format)
	{
	case D3DFMT_A8R8G8B8:case D3DFMT_X8R8G8B8:
		return 32;
	case D3DFMT_A4R4G4B4:
	case D3DFMT_R5G6B5:case D3DFMT_X1R5G5B5:
		return 16;
	case D3DFMT_R8G8B8:
		return 24;
	}
	return 0;
}
SurfaceColorFormat TranslateFormat(D3DFORMAT format)
{
	switch(format)
	{
	case D3DFMT_A8R8G8B8:
		return SCF_ARGB8888;
	case D3DFMT_X8R8G8B8:
		return SCF_RGB8888;
	case D3DFMT_A4R4G4B4:
		return SCF_ARGB4444;
	case D3DFMT_R5G6B5:
		return SCF_RGB565;
	case D3DFMT_X1R5G5B5:
		return SCF_RGB555;
	case D3DFMT_R8G8B8:
		return SCF_RGB888;
	}
	return SCF_UNKNOWN;
}

template<class DestT,class SrcT>
void CopyRect(DestT *dest,SrcT *src,int dpitch,int spitch,int width,int height)
{
	for (int i = 0;i < height;i++)
	{
		for (int j = 0;j < width;j++)
			RGB_Convert(dest[j],src[j]);
		dest = reinterpret_cast<DestT *>(reinterpret_cast<char *>(dest) + dpitch);
		src = reinterpret_cast<SrcT *>(reinterpret_cast<char *>(src) + spitch);
	}
}
bool CopySurface(IDirect3DSurface8 *dest,IDirect3DSurface8 *src)
{
	using namespace Juna::D2D;
	D3DSURFACE_DESC ddesc,sdesc;
	if FAILED(dest->GetDesc(&ddesc))
		return false;
	if FAILED(src->GetDesc(&sdesc))
		return false;

	D3DLOCKED_RECT drect,srect;
	if FAILED(dest->LockRect(&drect,NULL,0))
		return false;
	if FAILED(src->LockRect(&srect,NULL,D3DLOCK_READONLY)) {
		dest->UnlockRect();
		return false;
	}
	int width = min(ddesc.Width,sdesc.Width);
	int height = min(ddesc.Height,sdesc.Height);
	switch (ddesc.Format)
	{
	case D3DFMT_A8R8G8B8: case D3DFMT_X8R8G8B8:
		switch (sdesc.Format)
		{
		case D3DFMT_R5G6B5:
			CopyRect(static_cast<RGB8888 *>(drect.pBits),
					 static_cast<RGB565 *>(srect.pBits),
					 drect.Pitch,srect.Pitch,width,height);
			break;
		case D3DFMT_X1R5G5B5:
			CopyRect(static_cast<RGB8888 *>(drect.pBits),
					 static_cast<RGB555 *>(srect.pBits),
					 drect.Pitch,srect.Pitch,width,height);
			break;
		}
		break;
	case D3DFMT_R5G6B5:
		switch (sdesc.Format)
		{
		case D3DFMT_A8R8G8B8: case D3DFMT_X8R8G8B8:
			CopyRect(static_cast<RGB565 *>(drect.pBits),
					 static_cast<RGB8888 *>(srect.pBits),
					 drect.Pitch,srect.Pitch,width,height);
			break;
		case D3DFMT_X1R5G5B5:
			CopyRect(static_cast<RGB565 *>(drect.pBits),
					 static_cast<RGB555 *>(srect.pBits),
					 drect.Pitch,srect.Pitch,width,height);
			break;
		}
		break;
	case D3DFMT_X1R5G5B5:
		switch (sdesc.Format)
		{
		case D3DFMT_A8R8G8B8: case D3DFMT_X8R8G8B8:
			CopyRect(static_cast<RGB555 *>(drect.pBits),
					 static_cast<RGB8888 *>(srect.pBits),
					 drect.Pitch,srect.Pitch,width,height);
			break;
		case D3DFMT_R5G6B5:
			CopyRect(static_cast<RGB555 *>(drect.pBits),
					 static_cast<RGB565 *>(srect.pBits),
					 drect.Pitch,srect.Pitch,width,height);
			break;
		}
		break;
	}

	src->UnlockRect();
	dest->UnlockRect();
	return true;
}

const D3DMATRIX IdentityMatrix = {1,0,0,0, 0,1,0,0, 0,0,1,0, 0,0,0,1};

}//	anonymous namespace



bool Manager::Initialize(int width,int height,bool fullscreen,
						 int depth,int refresh_rate,bool use_double,
						 void *hWindow)
{
	assert(depth == 0 || depth == 15 || depth == 16 || depth == 32);
	Terminalize();

	UINT adapter = D3DADAPTER_DEFAULT;
	D3DDEVTYPE devtype = D3DDEVTYPE_HAL;
	DWORD bflag = D3DCREATE_HARDWARE_VERTEXPROCESSING | D3DCREATE_PUREDEVICE;
	{
		const TCHAR * const section = TEXT("Direct2D");
		TCHAR filename[MAX_PATH];
		::GetModuleFileName(NULL,filename,MAX_PATH);
		int len = ::lstrlen(filename);
		if (::lstrcmpi(filename + len - 4,TEXT(".exe")) == 0)
			::lstrcpy(filename + len - 3,TEXT("ini"));
		else
			::lstrcpy(filename,TEXT(""));

		adapter = ::GetPrivateProfileInt(section,TEXT("ADAPTER"),0,filename);
		if (::GetPrivateProfileInt(section,TEXT("USE_REFERENCE"),0,filename))
			devtype = D3DDEVTYPE_REF;
		if (::GetPrivateProfileInt(section,TEXT("FORCE_SOFTVERTEX"),0,filename))
			bflag = D3DCREATE_SOFTWARE_VERTEXPROCESSING;
	}

	HWND hwnd = (HWND)hWindow;
	{	//EnumThreadWindowsŒT
		struct local {	//Enum̂
			static BOOL CALLBACK ETWP(HWND hwnd,LPARAM lParam)
			{(*(HWND *)lParam) = hwnd;return FALSE;}
		};
		::EnumThreadWindows(GetCurrentThreadId(),local::ETWP,(LPARAM)&hwnd);
	}
	if (!hwnd)
		return false;
	AdjustWindowRect(hwnd,width,height);

	//RpC`FbN
#pragma warning ( disable : 4189 )
	{int static_test[D3DPRESENT_PARAMETERS_SIZE == sizeof(D3DPRESENT_PARAMETERS)] = {0};}
	D3DPRESENT_PARAMETERS &D3D_PP = *reinterpret_cast<D3DPRESENT_PARAMETERS*>(D3D_PP_Buf);

	IDirect3D8 *object;
	object = ::Direct3DCreate8(D3D_SDK_VERSION);
	if (object == NULL)
		goto ERR;


	D3D_PP.MultiSampleType = D3DMULTISAMPLE_NONE;
	D3D_PP.EnableAutoDepthStencil = FALSE;
	D3D_PP.Flags = 0;

	D3D_PP.SwapEffect = D3DSWAPEFFECT_DISCARD;
	D3D_PP.BackBufferWidth = width;
	D3D_PP.BackBufferHeight = height;
	D3D_PP.BackBufferCount = 1;
	D3D_PP.hDeviceWindow = hwnd;

	HRESULT hr;
	if (fullscreen)
	{
		D3D_PP.Windowed = FALSE;
		switch (depth)
		{
		case 16:
			hr = object->CheckDeviceType(adapter,devtype,
										 D3DFMT_R5G6B5,D3DFMT_R5G6B5,FALSE);
			if SUCCEEDED(hr)
			{
				D3D_PP.BackBufferFormat = D3DFMT_R5G6B5;
				break;
			}
			depth = 15;
		case 15:
			hr = object->CheckDeviceType(adapter,devtype,
										 D3DFMT_X1R5G5B5,D3DFMT_X1R5G5B5,FALSE);
			if SUCCEEDED(hr)
				D3D_PP.BackBufferFormat = D3DFMT_X1R5G5B5;
			else
				goto ERR;
			break;
		case 32:
			D3D_PP.BackBufferFormat = D3DFMT_X8R8G8B8;
			break;
		}
		D3D_PP.FullScreen_RefreshRateInHz = refresh_rate;
		D3D_PP.FullScreen_PresentationInterval = D3DPRESENT_INTERVAL_ONE;
	}
	else
	{
		depth = 0;
		D3D_PP.Windowed = TRUE;
		D3D_PP.FullScreen_RefreshRateInHz = 0;
		D3D_PP.FullScreen_PresentationInterval = D3DPRESENT_INTERVAL_DEFAULT;
	}
	{
		D3DDISPLAYMODE d3ddm;
		hr = object->GetAdapterDisplayMode(adapter,&d3ddm);
		if FAILED(hr)
			goto ERR;
		WindowFormat = d3ddm.Format;
	}
	if (depth == 0)
	{
		D3D_PP.BackBufferFormat = D3DFORMAT(WindowFormat);
		depth = TranslateFormat(D3D_PP.BackBufferFormat);
		if (depth == 0)
			goto ERR;
	}

	bflag |= use_double ? D3DCREATE_FPU_PRESERVE : 0;
	hr = object->CreateDevice(adapter,devtype,hwnd,bflag,&D3D_PP,&Device);
	if FAILED(hr)
	{
		bflag = (use_double ? D3DCREATE_FPU_PRESERVE : 0) | D3DCREATE_SOFTWARE_VERTEXPROCESSING;
		hr = object->CreateDevice(adapter,devtype,hwnd,bflag,&D3D_PP,&Device);
		if FAILED(hr)
		{
			Device = NULL;
			goto ERR;
		}
	}

	hr = Device->GetBackBuffer(0,D3DBACKBUFFER_TYPE_MONO,&BackBuffer.Surface);
	if FAILED(hr)
	{
		Device->Release();
		Device = NULL;
		goto ERR;
	}
	BackBuffer.Manager = this;
	BackBuffer.Width = width;
	BackBuffer.Height = height;

	ResetDeviceState();

	D3DObject = object;
	return true;

ERR:;
	if (object)
		object->Release();
	return false;
}

void Manager::Terminalize(void)
{
	if (ManagedHeader != 0)
	{
		I_Managed *next = ManagedHeader;
		do {
			I_Managed *m = next;
			next = m->next;
			m->Terminalize();
		} while (next != 0);
	}
	if (BackBuffer.Surface) {
		BackBuffer.Surface->Release();
		BackBuffer.Surface = 0;
	}
	if (Device) {
		Device->Release();
		Device = 0;
	}
	if (D3DObject) {
		D3DObject->Release();
		D3DObject = 0;
	}
	BackBuffer.Manager = 0;
	BackBuffer.Width = BackBuffer.Height = 0;
}

bool Manager::Reset(int width,int height,bool fullscreen,
					int depth,int refresh_rate)
{
	assert(depth == 0 || depth == 15 || depth == 16 || depth == 32);
	if (!Device)
		return false;

	HRESULT hr;

	D3DCAPS8 caps;
	hr = Device->GetDeviceCaps(&caps);
	if FAILED(hr)
		return false;
	UINT adapter = caps.AdapterOrdinal;
	D3DDEVTYPE devtype = caps.DeviceType;

	D3DPRESENT_PARAMETERS &D3D_PP = *reinterpret_cast<D3DPRESENT_PARAMETERS*>(D3D_PP_Buf);
	bool last_fullscreen = !D3D_PP.Windowed;

	D3D_PP.BackBufferWidth = width;
	D3D_PP.BackBufferHeight = height;

	if (fullscreen)
	{
		D3D_PP.Windowed = FALSE;
		switch (depth)
		{
		case 0:
			depth = this->depth();
			break;
		case 16:
			hr = D3DObject->CheckDeviceType(adapter,devtype,
											D3DFMT_R5G6B5,D3DFMT_R5G6B5,FALSE);
			if SUCCEEDED(hr)
			{
				D3D_PP.BackBufferFormat = D3DFMT_R5G6B5;
				break;
			}
			depth = 15;
		case 15:
			hr = D3DObject->CheckDeviceType(adapter,devtype,
											D3DFMT_X1R5G5B5,D3DFMT_X1R5G5B5,FALSE);
			if SUCCEEDED(hr)
				D3D_PP.BackBufferFormat = D3DFMT_X1R5G5B5;
			else
				return false;
			break;
		case 32:
			D3D_PP.BackBufferFormat = D3DFMT_X8R8G8B8;
			break;
		}
		if (refresh_rate)
			D3D_PP.FullScreen_RefreshRateInHz = refresh_rate;
	}
	else
	{
		D3D_PP.BackBufferFormat = D3DFORMAT(WindowFormat);
		depth = TranslateFormat(D3D_PP.BackBufferFormat);
		if (depth == 0)
			return false;
		D3D_PP.Windowed = TRUE;
		D3D_PP.FullScreen_RefreshRateInHz = 0;
	}
	ReleaseManaged();
	hr = Device->Reset(&D3D_PP);
	if FAILED(hr)
		return false;

	BackBuffer.Width = width;
	BackBuffer.Height = height;
	ResetDeviceState();

	if (RestoreManaged())
		return false;


	if (!fullscreen)
	{
		if (last_fullscreen)
		{
			HWND hwnd = D3D_PP.hDeviceWindow;
			HWND hDesktop = ::GetDesktopWindow();
	//	悭񂪁ANC̈悪
	//		::ShowWindow(hwnd,SW_SHOWMINIMIZED);
	//		::ShowWindow(hwnd,SW_RESTORE);
//			::SetWindowPos(hwnd,0,0,0,0,0,
//				SWP_NOACTIVATE | SWP_NOMOVE | SWP_NOREDRAW | SWP_NOSIZE | SWP_NOZORDER | SWP_DRAWFRAME);
			::SetWindowPos(hwnd,HWND_NOTOPMOST,0,0,0,0,
				SWP_NOACTIVATE | SWP_NOMOVE | SWP_NOREDRAW | SWP_NOSIZE | SWP_DRAWFRAME);

 			AdjustWindowRect(hwnd,width,height);

			RECT dtrect,rect;
			::GetWindowRect(hDesktop,&dtrect);
			::GetWindowRect(hwnd,&rect);
			::SetWindowPos(hwnd,HWND_TOP,
				(dtrect.right - (rect.right - rect.left)) / 2,
				(dtrect.bottom - (rect.bottom - rect.top)) / 2,
				0,0,
				SWP_NOSIZE);
		}
		else
		 	AdjustWindowRect(D3D_PP.hDeviceWindow,width,height);
	}
	return true;
}


void Manager::ResetDeviceState(void) const
{
	Device->SetRenderState(D3DRS_ZENABLE,FALSE);
	Device->SetRenderState(D3DRS_LIGHTING,FALSE);
	Device->SetRenderState(D3DRS_CULLMODE,D3DCULL_NONE);

	Device->SetTextureStageState(0,D3DTSS_ALPHAOP,D3DTOP_MODULATE);

	Device->SetRenderState(D3DRS_ALPHABLENDENABLE,TRUE);
	{	//
		Device->SetRenderState(D3DRS_SRCBLEND,D3DBLEND_SRCALPHA);
		Device->SetRenderState(D3DRS_DESTBLEND,D3DBLEND_INVSRCALPHA);
	}
}
void Manager::SetLinearFilter(bool flag)
{
	if (flag)
	{
		Device->SetTextureStageState(0,D3DTSS_MAGFILTER,D3DTEXF_LINEAR);
		Device->SetTextureStageState(0,D3DTSS_MINFILTER,D3DTEXF_LINEAR);
	}
	else
	{
		Device->SetTextureStageState(0,D3DTSS_MAGFILTER,D3DTEXF_POINT);
		Device->SetTextureStageState(0,D3DTSS_MINFILTER,D3DTEXF_POINT);
	}
}

int Manager::depth(void) const
{
	const D3DPRESENT_PARAMETERS &D3D_PP =
		*static_cast<const D3DPRESENT_PARAMETERS*>(GetD3D_PP());
	return TranslateFormatDepth(D3D_PP.BackBufferFormat);
}
SurfaceColorFormat Manager::format(void) const
{
	const D3DPRESENT_PARAMETERS &D3D_PP =
		*static_cast<const D3DPRESENT_PARAMETERS*>(GetD3D_PP());
	return TranslateFormat(D3D_PP.BackBufferFormat);
}

bool Manager::isFullScreen(void) const
{
	const D3DPRESENT_PARAMETERS &D3D_PP =
		*static_cast<const D3DPRESENT_PARAMETERS*>(GetD3D_PP());
	return D3D_PP.Windowed == FALSE;
}


void Manager::EnterManaged(I_Managed &managed)
{
	assert(managed.next == &managed);
	managed.next = ManagedHeader;
	ManagedHeader = &managed;
}

void Manager::ExitManaged(I_Managed &managed)
{
	if (ManagedHeader != &managed)
	{
		I_Managed *m;
		for (m = ManagedHeader;m->next != &managed;m = m->next)
			;
		m->next = managed.next;
	}
	else
		ManagedHeader = managed.next;
	managed.next = &managed;
}

void Manager::ReleaseManaged(void)
{
	BackBuffer.Surface->Release();
	BackBuffer.Surface = 0;
	I_Managed *m;
	for (m = ManagedHeader;m != 0;m = m->next)
		m->Release();
}


int Manager::RestoreManaged(void)
{
	HRESULT hr;
	hr = Device->GetBackBuffer(0,D3DBACKBUFFER_TYPE_MONO,&BackBuffer.Surface);
	int fail = FAILED(hr);
	I_Managed *m;
	for (m = ManagedHeader;m != 0;m = m->next)
		fail += !m->Restore();
	return fail;
}

bool Manager::Flip(void)
{
	HRESULT hr = Device->Present(NULL,NULL,NULL,NULL);
	if (hr == D3DERR_DEVICELOST)
	{
		hr = Device->TestCooperativeLevel();
		if (hr == D3DERR_DEVICENOTRESET)
		{
			ReleaseManaged();
			D3DPRESENT_PARAMETERS *D3D_PP =
				reinterpret_cast<D3DPRESENT_PARAMETERS*>(D3D_PP_Buf);
			hr = Device->Reset(D3D_PP);
			if FAILED(hr)
				return false;
			if (RestoreManaged())
				return false;
			ResetDeviceState();
		}
	}
	return true;
}




bool I_Surface::Initialize(D2D::Manager &manager,int width,int height,bool screen)
{
	Terminalize();
	Manager = &manager;
	Width = width;
	Height = height;
	if (!Restore(screen)) {
		Terminalize();
		return false;
	}
	return true;
}
void I_Surface::Terminalize(void)
{
	Manager = 0;
	Release();
	Width = Height = 0;
}
void I_Surface::Release(void)
{
	if (Surface)
		Surface->Release();
	Surface = 0;
}
bool I_Surface::Restore(bool screen)
{
	Release();
	IDirect3DDevice8 * const device = Manager->GetD3DDevice();
	if (!device)
		return false;

	const D3DPRESENT_PARAMETERS &D3D_PP = *Manager->GetD3D_PP();
	HRESULT hr;
	if (screen)
		hr = device->CreateRenderTarget(Width,Height,D3D_PP.BackBufferFormat,
										D3DMULTISAMPLE_NONE,FALSE,&Surface);
	else
		hr = device->CreateImageSurface(Width,Height,D3D_PP.BackBufferFormat,&Surface);
	if FAILED(hr) {
		Surface = 0;
		return false;
	}
	return true;
}

int I_Surface::depth(void) const
{
	if (!Surface)
		return 0;
	D3DSURFACE_DESC desc;
	if FAILED(Surface->GetDesc(&desc))
		return 0;
	return TranslateFormatDepth(desc.Format);
}
SurfaceColorFormat I_Surface::format(void) const
{
	if (!Surface)
		return SCF_UNKNOWN;
	D3DSURFACE_DESC desc;
	if FAILED(Surface->GetDesc(&desc))
		return SCF_UNKNOWN;
	return TranslateFormat(desc.Format);
}

bool I_Surface::Copy(int x,int y,
					 const I_Surface &src,const Rect *rect)
{
	if (!Surface)
		return false;
	IDirect3DDevice8 *device = Manager->GetD3DDevice();
	HRESULT hr;
	const POINT pt = {x,y};
	if (rect) {
		const RECT r = {rect->x1,rect->y1,rect->x2,rect->y2};
		hr = device->CopyRects(src.Surface,&r,1,Surface,&pt);
	}
	else
	{
		const RECT r = {0,0,src.width(),src.height()};
		hr = device->CopyRects(src.Surface,&r,1,Surface,&pt);
	}
	return SUCCEEDED(hr);
}

void I_Surface::Swap(I_Surface &other)
{
	{IDirect3DSurface8 *tmp = Surface;Surface = other.Surface;other.Surface = tmp;}
	{int tmp = Width;Width = other.Width;other.Width = tmp;}
	{int tmp = Height;Height = other.Height;other.Height = tmp;}
	{D2D::Manager *tmp = Manager;Manager = other.Manager;other.Manager = tmp;}
}


bool I_SpriteSurface::Initialize(D2D::Manager &manager,
						  int width,int height,int depth,PoolType pool,bool screen)
{
	Terminalize();
	{//̗ݏɍ킹
		int n;
		for (n = 2;n < width;n <<= 1)
			;
		width = n;
		for (n = 2;n < height;n <<= 1)
			;
		height = n;
	}
	Manager = &manager;
	Width = width;
	Height = height;
	if (!Restore(depth,pool,screen)) {
		Terminalize();
		return false;
	}
	return true;
}
void I_SpriteSurface::Terminalize(void)
{
	Release();
	I_Surface::Terminalize();
}
void I_SpriteSurface::Release(void)
{
	if (Texture)
		Texture->Release();
	Texture = 0;
	I_Surface::Release();
}
bool I_SpriteSurface::Restore(int depth,PoolType pool,bool screen)
{
	Release();
	IDirect3DDevice8 * const device = Manager->GetD3DDevice();
	if (!device)
		return false;
	if (depth == 0)
		depth = Manager->depth();
	D3DFORMAT form;
	switch (depth)
	{
	case 16: form = D3DFMT_A4R4G4B4;break;
	case 32: form = D3DFMT_A8R8G8B8;break;
	default: return false;
	}
	D3DPOOL p;
	switch (pool)
	{
	case VRAM:		p = D3DPOOL_DEFAULT;	break;
	case MANAGED:	p = D3DPOOL_MANAGED;	break;
	case SYSMEM:	p = D3DPOOL_SYSTEMMEM;	break;
	default: return false;
	}
	DWORD usage = 0;
	if (screen)
		usage = D3DUSAGE_RENDERTARGET;

	if FAILED(device->CreateTexture(Width,Height,1,usage,form,p,&Texture)) {
		Texture = 0;
		return false;
	}
	if FAILED(Texture->GetSurfaceLevel(0,&Surface)) {
		Texture->Release();
		Texture = 0;
		return false;
	}
	return true;
}

void I_SpriteSurface::Swap(I_SpriteSurface &other)
{
	{IDirect3DTexture8 *tmp = Texture;Texture = other.Texture;other.Texture = tmp;}
	{int tmp = Width;Width = other.Width;other.Width = tmp;}
	{int tmp = Height;Height = other.Height;other.Height = tmp;}
	{D2D::Manager *tmp = Manager;Manager = other.Manager;other.Manager = tmp;}
}



void I_ScreenSurface::Clear(Color color)
{
	IDirect3DDevice8 &device = *Manager->GetD3DDevice();
	HRESULT hr;
	hr = device.SetRenderTarget(Surface,NULL);
	D3DCOLOR c = D3DCOLOR_ARGB(color.a,color.r,color.g,color.b);
	hr = device.Clear(0,NULL,D3DCLEAR_TARGET,c,1.0f,0);
}

void I_ScreenSurface::SetClipRect(const Rect *rect)
{
	struct D3DVIEWPORT8_Set : public D3DVIEWPORT8
	{
		void Set(DWORD x,DWORD y,DWORD width,DWORD height)
			{X = x;Y = y;Width = width;Height = height;MinZ = 0.0;MaxZ = 1.0;}
	};
	HRESULT hr;
	D3DVIEWPORT8_Set d3dvp;
	if (rect)
		d3dvp.Set(rect->x1,rect->y1,rect->width(),rect->height());
	else
		d3dvp.Set(0,0,Width,Height);
	hr = Manager->GetD3DDevice()->SetViewport(&d3dvp);
}

void I_ScreenSurface::BeginBlit(void)
{
	IDirect3DDevice8 &device = *Manager->GetD3DDevice();
	HRESULT hr;
	hr = device.SetRenderTarget(Surface,NULL);
	hr = device.BeginScene();
}
void I_ScreenSurface::EndBlit(void)
{
	HRESULT hr;
	hr = Manager->GetD3DDevice()->EndScene();
}
void I_ScreenSurface::SetBlendType(BlendType bt)
{
	IDirect3DDevice8 &device = *Manager->GetD3DDevice();
	switch (bt)
	{
	case BT_ALPHA:	//ʂ̔
		device.SetRenderState(D3DRS_SRCBLEND,D3DBLEND_SRCALPHA);
		device.SetRenderState(D3DRS_DESTBLEND,D3DBLEND_INVSRCALPHA);
		break;
	case BT_ADD:	//Z
		device.SetRenderState(D3DRS_SRCBLEND,D3DBLEND_SRCALPHA);
		device.SetRenderState(D3DRS_DESTBLEND,D3DBLEND_ONE);
		break;
	case BT_SRC:	//̂܂܃Rs[
		device.SetRenderState(D3DRS_SRCBLEND,D3DBLEND_ONE);
		device.SetRenderState(D3DRS_DESTBLEND,D3DBLEND_ZERO);
		break;
	case BT_ADDFULL:	//At@Z
		device.SetRenderState(D3DRS_SRCBLEND,D3DBLEND_ONE);
		device.SetRenderState(D3DRS_DESTBLEND,D3DBLEND_ONE);
		break;

	case BT_HALFADD:	//Z
		device.SetRenderState(D3DRS_SRCBLEND,D3DBLEND_ONE);
		device.SetRenderState(D3DRS_DESTBLEND,D3DBLEND_INVSRCALPHA);
		break;
	case BT_MULTI:	//Z
		device.SetRenderState(D3DRS_SRCBLEND,D3DBLEND_DESTCOLOR);
		device.SetRenderState(D3DRS_DESTBLEND,D3DBLEND_INVSRCALPHA);
		break;
	case BT_NEGA:	//]
		device.SetRenderState(D3DRS_SRCBLEND,D3DBLEND_INVDESTCOLOR);
		device.SetRenderState(D3DRS_DESTBLEND,D3DBLEND_INVSRCALPHA);
		break;

	case BT_ADDLIKE:	//Zۂ
		device.SetRenderState(D3DRS_SRCBLEND,D3DBLEND_INVDESTCOLOR);
		device.SetRenderState(D3DRS_DESTBLEND,D3DBLEND_ONE);
		break;
	case BT_SUBLIKE:	//Zۂ
		device.SetRenderState(D3DRS_SRCBLEND,D3DBLEND_ZERO);
		device.SetRenderState(D3DRS_DESTBLEND,D3DBLEND_INVSRCCOLOR);
		break;
	case BT_LIGHT:	//Cg
		device.SetRenderState(D3DRS_SRCBLEND,D3DBLEND_DESTCOLOR);
		device.SetRenderState(D3DRS_DESTBLEND,D3DBLEND_ONE);
		break;
//	case BT_NEGA?:	//]ۂH
//		device.SetRenderState(D3DRS_SRCBLEND,D3DBLEND_INVDESTCOLOR);
//		device.SetRenderState(D3DRS_DESTBLEND,D3DBLEND_INVSRCCOLOR);
		break;
	}
}

void I_ScreenSurface::Blit(int x,int y,
						   const I_SpriteSurface &src,const Rect &srcrect,
						   float x_scale,float y_scale,Color color)
{
	IDirect3DDevice8 &device = *Manager->GetD3DDevice();
	HRESULT hr;
	hr = device.SetTexture(0,src.GetD3DTexture());
	hr = device.SetVertexShader(D3DFVF_XYZRHW|D3DFVF_DIFFUSE|D3DFVF_TEX1);

	struct vertex {float x,y,z,rhw;DWORD color;float tu, tv;};
	const float dleft = x - 0.5f,dtop = y - 0.5f;
	const float dright = x + srcrect.width() * x_scale - 0.5f;
	const float dbottom = y + srcrect.height() * y_scale - 0.5f;
	const float sleft = float(srcrect.x1) / src.width();
	const float stop = float(srcrect.y1) / src.height();
	const float sright = float(srcrect.x2) / src.width();
	const float sbottom = float(srcrect.y2) / src.height();
	const D3DCOLOR c = D3DCOLOR_ARGB(color.a,color.r,color.g,color.b);
	vertex v[4] = {	{dleft,dtop,0,1,c,sleft,stop},
					{dright,dtop,0,1,c,sright,stop},
					{dright,dbottom,0,1,c,sright,sbottom},
					{dleft,dbottom,0,1,c,sleft,sbottom}};

	hr = device.DrawPrimitiveUP(D3DPT_TRIANGLEFAN,2,v,sizeof(v[0]));
}
void I_ScreenSurface::Blit(float center_x,float center_y,
	const I_SpriteSurface &src,const FRect &srcrect,
	float angle,float x_scale,float y_scale,Color color)
{
	IDirect3DDevice8 &device = *Manager->GetD3DDevice();
	HRESULT hr;
	hr = device.SetTexture(0,src.GetD3DTexture());
	hr = device.SetVertexShader(D3DFVF_XYZRHW|D3DFVF_DIFFUSE|D3DFVF_TEX1);

	const float cx = center_x - 0.5f;
	const float cy = center_y - 0.5f;

	const float left = srcrect.x1 / src.width();
	const float top = srcrect.y1 / src.height();
	const float right = srcrect.x2 / src.width();
	const float bottom = srcrect.y2 / src.height();

	const float halfwidth = srcrect.width() * 0.5f * x_scale;
	const float halfheight = srcrect.height() * 0.5f * y_scale;

	const float cosa = ::cos(angle);
	const float sina = ::sin(angle);
	const float xcos = cosa * halfwidth;
	const float xsin = sina * halfwidth;
	const float ycos = cosa * halfheight;
	const float ysin = sina * halfheight;

	struct vertex {float x,y,z,rhw;DWORD color;float tu, tv;};
	const D3DCOLOR c = D3DCOLOR_ARGB(color.a,color.r,color.g,color.b);
	vertex v[4] = {	{cx - xcos + ysin,cy - xsin - ycos,0,1,c,left,top},
					{cx + xcos + ysin,cy + xsin - ycos,0,1,c,right,top},
					{cx + xcos - ysin,cy + xsin + ycos,0,1,c,right,bottom},
					{cx - xcos - ysin,cy - xsin + ycos,0,1,c,left,bottom}};

	hr = device.DrawPrimitiveUP(D3DPT_TRIANGLEFAN,2,v,sizeof(v[0]));
}
void I_ScreenSurface::Fill(const Rect &rect,Color color)
{
	IDirect3DDevice8 &device = *Manager->GetD3DDevice();
	struct vertex {float x,y,z,rhw;DWORD color;};
	const float left = rect.x1 - 0.5f;
	const float top = rect.y1 - 0.5f;
	const float right = rect.x2 - 0.5f;
	const float bottom = rect.y2 - 0.5f;
	const D3DCOLOR c = D3DCOLOR_ARGB(color.a,color.r,color.g,color.b);
	vertex v[4] = {	{left,top,0,1,c},
					{right,top,0,1,c},
					{right,bottom,0,1,c},
					{left,bottom,0,1,c}};
	HRESULT hr;
	hr = device.SetTexture(0,0);
	hr = device.SetVertexShader(D3DFVF_XYZRHW|D3DFVF_DIFFUSE);

	hr = device.DrawPrimitiveUP(D3DPT_TRIANGLEFAN,2,v,sizeof(v[0]));
}





bool SpriteSurface::Initialize(D2D::Manager &manager,int width,int height,int depth)
{
	if (!I_SpriteSurface::Initialize(manager,width,height,depth,MANAGED,false))
		return false;
	manager.EnterManaged(*this);
	return true;
}
void SpriteSurface::Terminalize(void)
{
	if (isManaged())
		Manager->ExitManaged(*this);
	I_SpriteSurface::Terminalize();
}


bool ScreenSurface::Initialize(D2D::Manager &manager,int width,int height)
{
	if (!I_ScreenSurface::Initialize(manager,width,height,true))
		return false;
	manager.EnterManaged(*this);
	return true;
}
void ScreenSurface::Terminalize(void)
{
	if (isManaged())
		Manager->ExitManaged(*this);
	I_ScreenSurface::Terminalize();
}

void ScreenSurface::Release(void) {I_ScreenSurface::Release();}
bool ScreenSurface::Restore(void) {return I_ScreenSurface::Restore(true);}


bool ImageSurface::Initialize(D2D::Manager &manager,int width,int height)
{
	Terminalize();
	if (!I_Surface::Initialize(manager,width,height,false))
		return false;
	manager.EnterManaged(*this);
	return true;
}
void ImageSurface::Terminalize(void)
{
	if (isManaged())
		Manager->ExitManaged(*this);
	I_Surface::Terminalize();
}
bool ImageSurface::Initialize(I_Surface &src)
{
//	ʓ|Ȃ̂łƂ肠Z
	D2D::Manager *manager = static_cast<ImageSurface &>(src).Manager;
	Terminalize();
	if (!src.GetD3DSurface())
		return false;
	if (!I_Surface::Initialize(*manager,src.width(),src.height(),false))
		return false;
	if FAILED(manager->GetD3DDevice()->CopyRects(src.GetD3DSurface(),0,0,Surface,NULL))
	{
		I_Surface::Terminalize();
		return false;
	}
	manager->EnterManaged(*this);
	return true;
}


bool BGSurface::Initialize(D2D::Manager &manager,int width,int height)
{
	Terminalize();
	if (!I_ScreenSurface::Initialize(manager,width,height,true))
		return false;
	if (!Backup.Initialize(*this))
	{
		I_ScreenSurface::Terminalize();
		return false;
	}
	manager.EnterManaged(*this);
	return true;
}

void BGSurface::Terminalize(void)
{
	if (isManaged())
		Manager->ExitManaged(*this);
	Backup.Terminalize();
	I_ScreenSurface::Terminalize();
}

bool BGSurface::SaveImage(const Rect *rect)
{
	if (rect)
		return Backup.Copy(rect->x1,rect->y1,*this,rect);
	else
		return Backup.Copy(0,0,*this,0);
}
bool BGSurface::RestoreImage(const Rect *rect)
{
	if (rect)
		return Copy(rect->x1,rect->y1,Backup,rect);
	else
		return Copy(0,0,Backup,0);
}
void BGSurface::Release(void) {I_ScreenSurface::Release();}
bool BGSurface::Restore(void)
{
	if (!I_ScreenSurface::Restore(true))
		return false;
	SurfaceColorFormat f = format(),bf = Backup.format();
	if (f == SCF_UNKNOWN || bf == SCF_UNKNOWN)
		return false;
	IDirect3DDevice8 * const device = Manager->GetD3DDevice();
	if (f != bf)	//	tH[}bgႤƂȂ
	{
		ImageSurface backup;
		if (!backup.Initialize(*Manager,Width,Height))
			return false;
//	OŕϊȂRs[
		if (!CopySurface(backup.GetD3DSurface(),Backup.GetD3DSurface()))
			return false;
		Backup.Swap(backup);
	}
	return RestoreImage();
}



}//namespace D2D
}//namespace Juna

