
#include "../InputFile.h"


//#define USE_WINDOWS_API


#ifndef USE_WINDOWS_API

#include <stdio.h>

namespace Juna {
namespace IO {

namespace {
long GetFileSize(FILE *file)
{
	if (::fseek(file,0,SEEK_END))
		return -1;
	return ::ftell(file);
}
}//anonymous namespace

bool InputFile::Open(const char *filename)
{
	Close();
	FILE *file = ::fopen(filename,"rb");
	if (!file)
		return false;
	unsigned long size = GetFileSize(file);
	if (size == 0xFFFFFFFF) {
		::fclose(file);
		return false;
	}
	FileHandle = file;
	Offset = 0;Size = size;TellPosition = 0;
	return true;
}
bool InputFile::Open(const char *filename,unsigned long offset,unsigned long size)
{
	Close();
	FILE *file = ::fopen(filename,"rb");
	if (!file)
		return false;
	unsigned long fsize = GetFileSize(file);
	if (fsize == 0xFFFFFFFF || 
		offset + size > fsize || ::fseek(file,offset,SEEK_SET))
	{
		::fclose(file);
		return false;
	}
	FileHandle = file;
	Offset = offset;Size = size;TellPosition = 0;
	return true;
}

void InputFile::Close(void)
{
	if (FileHandle)
		::fclose(static_cast<FILE *>(FileHandle));
	FileHandle = 0;
	Offset = Size = TellPosition = 0;
}

unsigned long InputFile::Read(void *dest,unsigned long size)
{
	if (TellPosition + size > Size) size = Size - TellPosition;
	unsigned long r = ::fread(dest,1,size,static_cast<FILE *>(FileHandle));
	if (r < size && ferror(static_cast<FILE *>(FileHandle)))
	{
		::clearerr(static_cast<FILE *>(FileHandle));
		return 0;
	}
	TellPosition += r;
	return r;
}

bool InputFile::Seek(unsigned long position)
{
	if (position > Size)
		position = Size;
	if (::fseek(static_cast<FILE *>(FileHandle),position,SEEK_SET))
		return false;
	TellPosition = position;
	return true;
}

}//namespace IO
}//namespace Juna



	//ifndef USE_WINDOWS_API
#else
	//ifdef USE_WINDOWS_API

#include <windows.h>

namespace Juna {
namespace IO {

bool InputFile::Open(const char *filename)
{
	Close();
	HANDLE hfile = INVALID_HANDLE_VALUE;
	hfile = ::CreateFile(filename,GENERIC_READ,FILE_SHARE_READ,NULL,
						 OPEN_EXISTING,FILE_ATTRIBUTE_NORMAL,NULL);
	if(hfile == INVALID_HANDLE_VALUE)
		return false;
	unsigned long size = ::GetFileSize(hfile,NULL);
	if (size == 0xFFFFFFFF) {
		::CloseHandle(hfile);
		return false;
	}
	FileHandle = hfile;
	Offset = 0;Size = size;TellPosition = 0;
	return true;
}
bool InputFile::Open(const char *filename,unsigned long offset,unsigned long size)
{
	Close();
	HANDLE hfile = INVALID_HANDLE_VALUE;
	hfile = ::CreateFile(filename,GENERIC_READ,FILE_SHARE_READ,NULL,
						 OPEN_EXISTING,FILE_ATTRIBUTE_NORMAL,NULL);
	if(hfile == INVALID_HANDLE_VALUE)
		return false;
	unsigned long fsize = ::GetFileSize(hfile,NULL);
	if (fsize == 0xFFFFFFFF || offset + size > fsize ||
		::SetFilePointer(hfile,offset,NULL,FILE_BEGIN))
	{
		::CloseHandle(hfile);
		return false;
	}
	FileHandle = hfile;
	Offset = offset;Size = size;TellPosition = 0;
	return true;
}
void InputFile::Close(void)
{
	if (FileHandle)
		::CloseHandle(FileHandle);
	FileHandle = 0;
	Offset = Size = TellPosition = 0;
}

unsigned long InputFile::Read(void *dest,unsigned long size)
{
	if (TellPosition + size > Size) size = Size - TellPosition;
	DWORD r;
	if (!::ReadFile(FileHandle,dest,size,&r,NULL))
		return 0;
	TellPosition += r;
	return r;
}

bool InputFile::Seek(unsigned long position)
{
	if (position > Size)
		position = Size;
	if (::SetFilePointer(FileHandle,position,NULL,FILE_BEGIN) == 0xFFFFFFFF)
		return false;
	TellPosition = position;
	return true;
}

}//namespace IO
}//namespace Juna

	//ifdef USE_WINDOWS_API
#endif

namespace Juna {
namespace IO {

bool InputFile::ReduceRange(unsigned long front,unsigned long back)
{
	bool r = true;
	if (front > Size) front = Size;
	Offset += front;
	Size -= front;
	if (TellPosition < front)
		r = false;
	TellPosition -= front;

	if (back > Size) back = Size;
	Size -= back;
	if (TellPosition > Size)
		r = false;
	return r;
}

}//namespace IO
}//namespace Juna

