
#define _CRT_SECURE_NO_DEPRECATE

#include "../BitmapFile.h"
#include "../../int_def.h"
#include <stdio.h>
#include <string.h>
//#include <windows.h>


namespace Juna {
namespace Picture {

namespace {

static const uint32 BI_RGB = 0;

#pragma pack(push,1)
struct BITMAPINFOHEADER
{
	uint32	size;
	int32	width;
	int32	height;
	uint16	planes;
	uint16	bitCount;
	uint32	compression;
	uint32	sizeImage;
	int32	xPelsPerMeter;
	int32	yPelsPerMeter;
	uint32	clrUsed;
	uint32	clrImportant;
};
struct BITMAPFILEHEADER
{
	uint16	type;
	uint32	size;
	uint16	reserved1;
	uint16	reserved2;
	uint32	offBits;
};
struct RGBQUAD
{
	uint8	blue;
	uint8	green;
	uint8	red;
	uint8	reserved;
};
#pragma pack(pop)

template<typename T> inline T max(T a,T b) {return a < b ? b : a;}
template<typename T> inline T min(T a,T b) {return a < b ? a : b;}
}//	anonymous namespace

bool GetBitmapInfo(BITMAPINFOHEADER &bmpInfo,	//ރobt@
				   const char *filename)		//BMPt@C
{
	FILE *file = ::fopen(filename,"rb");
	if (file == NULL)
		return false;

	BITMAPFILEHEADER bmpHead;
	if (::fread(&bmpHead,sizeof(bmpHead),1,file) < 1)
	{
		::fclose(file);
		return false;
	}
	if (::fread(&bmpInfo,sizeof(bmpInfo),1,file) < 1)
	{
		::fclose(file);
		return false;
	}
	::fclose(file);
	return true;
}


static bool Load8bitBmp(I_Image &image,
				 const BITMAPINFOHEADER &bmpinfo,
				 FILE *file);
static bool Load24bitBmp(I_Image &image,
				  const BITMAPINFOHEADER &bmpinfo,
				  FILE *hfile);
static bool Load32bitBmp(I_Image &image,
				  const BITMAPINFOHEADER &bmpinfo,
				  FILE *hfile);

bool LoadBitmapFile(I_Image &rTarget,const char *iFileName)
{
	BITMAPFILEHEADER bmpHead;
	BITMAPINFOHEADER bmpInfo;

	FILE *file = ::fopen(iFileName,"rb");
	if (file == NULL)
		return false;

	if (::fread(&bmpHead,sizeof(bmpHead),1,file) < 1)
		goto ERR;
	if (::fread(&bmpInfo,sizeof(bmpInfo),1,file) < 1)
		goto ERR;

//rbg}bvBPP`FbNēǂݍ
	switch (bmpInfo.bitCount)
	{
	case 8:
		if (!Load8bitBmp(rTarget,bmpInfo,file))
			goto ERR;
		break;
	case 24:
		if (!Load24bitBmp(rTarget,bmpInfo,file))
			goto ERR;
		break;
	case 32:
		if (!Load32bitBmp(rTarget,bmpInfo,file))
			goto ERR;
		break;
	default:
		goto ERR;
	}

	::fclose(file);
	return true;

ERR:;
	::fclose(file);
	return false;
}


bool SaveBitmapFile(const char *filename,const I_Image &image,bool alpha,bool overwrite)
{
	unsigned int width,height,bpp;
	width = image.width();
	height = image.height();
	if (alpha)
		bpp = 32;
	else
		bpp = 24;

	BITMAPFILEHEADER bmpHead = {0};
	BITMAPINFOHEADER bmpInfo = {0};

	bmpHead.type = ('M' << 8) + 'B';
	bmpHead.size = sizeof(bmpHead);

	bmpHead.offBits = sizeof(bmpHead) + sizeof(bmpInfo);

	bmpInfo.size = sizeof(bmpInfo);
	bmpInfo.width = width;
	bmpInfo.height = height;
	bmpInfo.planes = 1;
	bmpInfo.bitCount = uint16(bpp);
	bmpInfo.compression = BI_RGB;
	bmpInfo.sizeImage = width * height * bpp / 8;

	if (!overwrite)
	{
		FILE *file = ::fopen(filename,"rb");
		::fclose(file);
		if (file)
			return false;
	}
	FILE *file = ::fopen(filename,"wb");
	if (file == NULL)
		return false;


	if (::fwrite(&bmpHead,sizeof(bmpHead),1,file) < 1)
		goto ERR;
	if (::fwrite(&bmpInfo,sizeof(bmpInfo),1,file) < 1)
		goto ERR;

	if (bpp == 32)
	{	//At@ 32bitrbg}bv
		for (int i = height - 1;i >= 0;i--)
		{
			if (::fwrite(image[i],4,width,file) < width)
				goto ERR;
		}
	}
	else
	{	//At@ 24bitrbg}bv
		int bwidth = width * bpp / 8;
		int pitch = (bwidth + 3) & ~3;
		for (int i = height - 1;i >= 0;i--)
		{
			for (unsigned int j = 0;j < width;j++)
			{
				//ꂾł܂s񂾂
				if (::fwrite(&image[i][j],1,3,file) < 3)
					goto ERR;
			}
			//[ȃTCY猊
			::fseek(file,pitch - bwidth,SEEK_CUR);
//			WriteFile(hfile,"\0\0",pitch - bwidth,&write,NULL);
		}
	}

	::fclose(file);

	return true;
ERR:;
	::fclose(file);

	return false;
}

static bool Load8bitBmp(I_Image &target,
						const BITMAPINFOHEADER &bmpinfo,
						FILE *file)
{
	RGBQUAD pal[256];
	if (::fread(pal,sizeof(RGBQUAD),256,file) < 256)
		return false;

	uint8 *image;
	image = new uint8[bmpinfo.width];
	if (image == 0)
		return false;

	int width,height;
	width = (bmpinfo.width < target.width()) ? bmpinfo.width : target.width();
	height = (bmpinfo.height < target.height()) ? bmpinfo.height : target.height();

	int readwidth = (bmpinfo.width + 3) & ~3;	//4byteAC
//Ȃꍇ̕Ƃ΂
	::fseek(file,(bmpinfo.height - height) * readwidth,SEEK_CUR);
	ARGB *dest = target[height - 1];
	for (int i = height - 1;i >= 0;i--)
	{
		if (::fread(image,sizeof(uint8),bmpinfo.width,file) < size_t(bmpinfo.width))
		{
			delete[] image;
			return false;
		}
//]vȕXLbv
		::fseek(file,readwidth - bmpinfo.width,SEEK_CUR);

		for (int j = 0;j < width;j++)
		{
			ARGB &pxl = dest[j];
			pxl.a = 0xFF;
			pxl.r = pal[image[j]].red;
			pxl.g = pal[image[j]].green;
			pxl.b = pal[image[j]].blue;
		}
		dest -= target.distance();
	}
	delete[] image;
	return true;
}

static bool Load24bitBmp(I_Image &target,
						 const BITMAPINFOHEADER &bmpinfo,
						 FILE *file)
{
	uint8 *image;
	unsigned int bmpwidth = bmpinfo.width * 3;
	image = static_cast<uint8 *>(new uint8[bmpwidth]);
	if (image == NULL)
		return false;

	int width,height;
	width = (bmpinfo.width < target.width()) ? bmpinfo.width : target.width();
	height = (bmpinfo.height < target.height()) ? bmpinfo.height : target.height();

	unsigned int readwidth = (bmpwidth + 3) & ~3;	//4byteAC
//Ȃꍇ̕Ƃ΂
	for (int i = height - 1;i >= 0;i--)
	{
		if (::fread(image,sizeof(uint8),bmpwidth,file) < bmpwidth)
		{
			delete[] image;
			return false;
		}
//]vȕXLbv
		::fseek(file,readwidth - bmpwidth,SEEK_CUR);
		for (int j = 0,j2 = 0;j < width;j++,j2 += 3)
		{
			ARGB &pxl = target[i][j];
			pxl.a = 0xFF;
			pxl.r = image[j2 + 2];
			pxl.g = image[j2 + 1];
			pxl.b = image[j2 + 0];
		}
	}
	delete[] image;
	return true;
}

static bool Load32bitBmp(I_Image &target,
						 const BITMAPINFOHEADER &bmpinfo,
						 FILE *file)
{
	unsigned width,height;
	width = (bmpinfo.width < target.width()) ? bmpinfo.width : target.width();
	height = (bmpinfo.height < target.height()) ? bmpinfo.height : target.height();
//4ByteACA蓾Ȃy
	for (int i = height - 1;i >= 0;i--)
	{
		if (::fread(target[i],sizeof(uint32),width,file) < width)
			return false;
		::fseek(file,(bmpinfo.width - width) * sizeof(uint32),SEEK_CUR);
	}
	return true;
}



bool CreateARGBPicture(Image &dest,const I_Image &black,const I_Image &white)
{
	if (black.width() != white.width() ||
		black.height() != white.height())
		return false;

	dest.Initialize(black.width(),black.height());

	int i,j;
	for (i = 0;i < dest.height();i++)
		::memcpy(dest[i],black[i],dest.width() * sizeof(ARGB));
	for (i = 0;i < dest.height();i++)
	{
		for (j = 0;j < dest.width();j++)
		{
//			BYTE *dargb = (BYTE *)&dest.Vram[i][j];
			ARGB w = white[i][j],b = black[i][j];
			ARGB sub(w.a - b.a,w.r - b.r,w.g - b.g,w.b - b.b);
			if ((sub.r | sub.g | sub.b) == 0)
			{
				dest[i][j] = black[i][j];
				dest[i][j].a = 0xFF;
			}
			else if ((sub.r & sub.g & sub.b) == 255)
			{
				dest[i][j].Set(0,0,0,0);
			}
			else
			{
//vZ
				uint8 maxs = max(max(sub.r,sub.g),sub.b);
				uint8 rate = 255 - maxs;
				ARGB bpix = black[i][j];
				dest[i][j].Set(rate,bpix.r * 255 / rate,bpix.g * 255 / rate,bpix.b * 255 / rate);
			}
		}
	}
	return true;
}

bool CreateARGBBitmapFile(const char *dest,const char *black,const char *white)
{
	Image dimage;
	Image bimage;
	Image wimage;

	BITMAPINFOHEADER binfo,winfo;
	if (!GetBitmapInfo(binfo,black))
		return false;
	if (!GetBitmapInfo(winfo,white))
		return false;

	if (binfo.width != winfo.width ||
		binfo.height != winfo.height)
		return false;
	bimage.Initialize(binfo.width,binfo.height);
	wimage.Initialize(winfo.width,winfo.height);
	if (LoadBitmapFile(bimage,black) && LoadBitmapFile(wimage,white))
	{
		if (!CreateARGBPicture(dimage,bimage,wimage))
			return false;
		if (!SaveBitmapFile(dest,dimage,true,false))
			return false;
		return true;
	}
	return false;
}

//*/

bool InitLoadBitmapFile(Image &oTarget,const char *iFileName)
{
	BITMAPINFOHEADER bmpinfo;
	if (!GetBitmapInfo(bmpinfo,iFileName))
		return false;
	oTarget.Initialize(bmpinfo.width,bmpinfo.height);
	if (!LoadBitmapFile(oTarget,iFileName))
	{
		oTarget.Terminalize();
		return false;
	}
	return true;
}



}//namespace Picture
}//namespace Juna
