
#ifndef Juna_RGB_Color_H
#define Juna_RGB_Color_H

#include "int_def.h"

namespace Juna {
namespace Color {

struct RGB
{
	uint8 b,g,r;
	RGB(uint8 r,uint8 g,uint8 b) : b(b),g(g),r(r) {}
	RGB(void) {}
};

struct ARGB
{
	uint8 b,g,r,a;
	ARGB(uint8 a,uint8 r,uint8 g,uint8 b) : b(b),g(g),r(r),a(a) {}
	ARGB(void) {}
};


class ARGB8888
{
public:
	uint8 a(void) const {return argb.a;}
	uint8 r(void) const {return argb.r;}
	uint8 g(void) const {return argb.g;}
	uint8 b(void) const {return argb.b;}
	void a(uint8 v) {argb.a = v;}
	void r(uint8 v) {argb.r = v;}
	void g(uint8 v) {argb.g = v;}
	void b(uint8 v) {argb.b = v;}

	ARGB8888(uint8 alpha,uint8 red,uint8 green,uint8 blue) {a(alpha);r(red);g(green);b(blue);}
private:
	ARGB argb;
};
class ARGB4444
{
public:
	uint8 a(void) const {int e = (argb >> 12) & 0xF;return uint8((e << 4) | e);}
	uint8 r(void) const {int e = (argb >> 8)  & 0xF;return uint8((e << 4) | e);}
	uint8 g(void) const {int e = (argb >> 4)  & 0xF;return uint8((e << 4) | e);}
	uint8 b(void) const {int e = (argb)		  & 0xF;return uint8((e << 4) | e);}
	void a(uint8 v) {argb = (argb & 0x0FFF) | (((v >> 4) & 0xF) << 12);}
	void r(uint8 v) {argb = (argb & 0xF0FF) | (((v >> 4) & 0xF) << 8);}
	void g(uint8 v) {argb = (argb & 0xFF0F) | (((v >> 4) & 0xF) << 4);}
	void b(uint8 v) {argb = (argb & 0xFFF0) | (((v >> 4) & 0xF));}

	ARGB4444(uint8 alpha,uint8 red,uint8 green,uint8 blue) {a(alpha);r(red);g(green);b(blue);}
private:
	uint16 argb;
};

class RGB8888
{
public:
	uint8 a(void) const {return 255;}
	uint8 r(void) const {return argb.r;}
	uint8 g(void) const {return argb.g;}
	uint8 b(void) const {return argb.b;}
	void a(uint8) {}
	void r(uint8 v) {argb.r = v;}
	void g(uint8 v) {argb.g = v;}
	void b(uint8 v) {argb.b = v;}

	RGB8888(uint8 alpha,uint8 red,uint8 green,uint8 blue) {a(alpha);r(red);g(green);b(blue);}
private:
	ARGB argb;
};
class RGB888
{
public:
	uint8 a(void) const {return 255;}
	uint8 r(void) const {return rgb.r;}
	uint8 g(void) const {return rgb.g;}
	uint8 b(void) const {return rgb.b;}
	void a(uint8) {}
	void r(uint8 v) {rgb.r = v;}
	void g(uint8 v) {rgb.g = v;}
	void b(uint8 v) {rgb.b = v;}

	RGB888(uint8 alpha,uint8 red,uint8 green,uint8 blue) {a(alpha);r(red);g(green);b(blue);}
private:
	RGB rgb;
};
class RGB565
{
public:
	uint8 a(void) const {return 255;}
	uint8 r(void) const {int e = (rgb >> 11) & 0x1F;return uint8((e << 3) | (e >> 2));}
	uint8 g(void) const {int e = (rgb >> 5)  & 0x3F;return uint8((e << 2) | (e >> 4));}
	uint8 b(void) const {int e = (rgb)		 & 0x1F;return uint8((e << 3) | (e >> 2));}
	void a(uint8) {}
	void r(uint8 v) {rgb = (rgb & 0x07FF) | (((v >> 3) & 0x1F) << 11);}
	void g(uint8 v) {rgb = (rgb & 0xF81F) | (((v >> 2) & 0x3F) << 6);}
	void b(uint8 v) {rgb = (rgb & 0xFFE0) | (((v >> 3) & 0x1F));}

	RGB565(uint8 alpha,uint8 red,uint8 green,uint8 blue) {a(alpha);r(red);g(green);b(blue);}
private:
	uint16 rgb;
};
class RGB555
{
public:
	uint8 a(void) const {return 255;}
	uint8 r(void) const {int e = (rgb >> 10) & 0x1F;return uint8((e << 3) | (e >> 2));}
	uint8 g(void) const {int e = (rgb >> 5)	 & 0x1F;return uint8((e << 3) | (e >> 2));}
	uint8 b(void) const {int e = (rgb)		 & 0x1F;return uint8((e << 3) | (e >> 2));}
	void a(uint8) {}
	void r(uint8 v) {rgb = (rgb & 0x83FF) | (((v >> 3) & 0x1F) << 10);}
	void g(uint8 v) {rgb = (rgb & 0xFC1F) | (((v >> 3) & 0x1F) << 5);}
	void b(uint8 v) {rgb = (rgb & 0xFFE0) | (((v >> 3) & 0x1F));}

	RGB555(uint8 alpha,uint8 red,uint8 green,uint8 blue) {a(alpha);r(red);g(green);b(blue);}
private:
	uint16 rgb;
};
typedef RGB555 RGB1555;


template<class DestRGB_T,class SrcRGB_T>
void RGB_Convert(DestRGB_T &dest,const SrcRGB_T &src)
	{dest.a(src.a());dest.r(src.r());dest.g(src.g());dest.b(src.b());}

template<class DestRGB_T,class SrcRGB_T>
DestRGB_T RGB_Convert(SrcRGB_T src)
	{return DestRGB_T(src.a(),src.r(),src.g(),src.b());}


}// namespace Color
}//	namespace Juna

#endif
