
#ifndef JunaSound_DirectSoundStreamPlayer_H
#define JunaSound_DirectSoundStreamPlayer_H

#include "WindowsPlayer.h"

struct IDirectSoundBuffer;

namespace Juna {
namespace Sound {

//ŎgpDirectSoundObjectbp[NX
namespace DirectSound {
class DirectSoundObject;
}

class DirectSoundStreamPlayer : public I_DirectSoundStreamPlayer
{
	friend class DirectSoundPlayDriver;
public:
	DirectSoundStreamPlayer(void);
	~DirectSoundStreamPlayer();

//tH[JXI^Initialize
//focus < 0  : O[otH[JX(DSBCAPS_STICKYFOCUS)
//focus == 0 : [JtH[JX
//focus > 0  : O[otH[JX(DSBCAPS_GLOBALFOCUS)
	bool Initialize(int iDevice,const Format &iFormat,
					long iBufsize = 0,int iBufnum = 3,int iFocus = 1);
	bool Initialize(int iDevice,I_InputStream &ioStream,
					long iBufsize = 0,int iBufnum = 3,int iFocus = 1);

	void Terminalize(void);

	bool isInitialized(void) const {return lpDSBuffer != 0;}

	virtual bool Start(void);
	virtual bool Stop(void);
	virtual bool isPlaying(void) const;

	virtual bool SetVolume(unsigned long iLeft,unsigned long iRight)
	{return SetVolumef(static_cast<float>(iLeft)/0x10000,static_cast<float>(iRight)/0x10000);}
	virtual bool GetVolume(unsigned long &oLeft,unsigned long &oRight) const
	{
		float fleft,fright;
		bool r = GetVolumef(fleft,fright);
		oLeft = static_cast<unsigned long>(fleft * 0x10000);
		oRight = static_cast<unsigned long>(fright * 0x10000);
		return r;
	}

	virtual Format GetFormat(void) const {return PlayFormat;}

	virtual const I_PlayDriver *GetDriver(void);

	virtual bool SetSupplier(I_InputStream *setStream);

	virtual long GetPlayTime(void);

	virtual void GetBufferSize(unsigned long &oBufsize,int &oBufnum) const
	{
		oBufsize = BufferSize;
		oBufnum = BufferNum;
	}

	virtual void SetPlayEndCallback(I_PlayEndCallback *setCallback);

	virtual int LoadStream(void);


	virtual bool SetVolumef(float iLeft,float iRight);
	virtual bool GetVolumef(float &oLeft,float &oRight) const;
	virtual bool SetVolume_dB(long iLeft,long iRight);
	virtual bool GetVolume_dB(long &oLeft,long &oRight) const;

private:
	bool CreateBuffer(const Format &iFormat,long bufsize,int bufnum,
					  int focus);

	int LoadData_NoCheck(void);
	int LoadData_OverCheck(void);
	int FillSilent(void);

private:
	IDirectSoundBuffer *lpDSBuffer;	//Đobt@{́B
	DirectSound::DirectSoundObject *Master;		//obt@DirectSoundObject
	I_InputStream *Stream;		//ĐXg[

	unsigned long BufferSize;	//obt@̃TCY
	int BufferNum;				//obt@̐
	volatile int WriteBufferPos;	//ɏރobt@̈ʒu

	Format PlayFormat;

	enum THREAD_STATE{TS_EXIT = 0,TS_STANDBY = 1,TS_PLAY = 2,TS_PLAYEND = 3,TS_PAUSE = 4};
	volatile THREAD_STATE ThreadState;
	HANDLE hNotifyEvent;	//ĐʒuʒmCxg
	HANDLE hThread;			//Đʒmmƃf[^[hpXbh̃nh
	CRITICAL_SECTION CritSect;

	volatile long TotalLoadBufferNum;
	volatile int LoadEndPos;			//TEhf[^ǂݍݏIʒu
	volatile DWORD LoadEndSize;			//ŏIobt@TCY

	I_PlayEndCallback *PlayEndCallback;


	static unsigned __stdcall PlayThread(void *lpDSPlayer);

private://
	DirectSoundStreamPlayer(DirectSoundStreamPlayer &);
	void operator=(const DirectSoundStreamPlayer &);
};

class DirectSoundPoolPlayer : public I_DirectSoundPoolPlayer
{
	friend class DirectSoundPlayDriver;
public:
	DirectSoundPoolPlayer(void) : lpDSBuffer(0),Master(0),LoopFlag(false) {}
	virtual ~DirectSoundPoolPlayer() {Terminalize();}

//tH[JXI^Initialize
//focus < 0  : O[otH[JX(DSBCAPS_STICKYFOCUS)
//focus == 0 : [JtH[JX
//focus > 0  : O[otH[JX(DSBCAPS_GLOBALFOCUS)
	bool Initialize(int device,const Format &iFormat,long bufsize,int focus = 1);
	bool Initialize(int device,I_SeekableInputStream &ioStream,int focus = 1);
	bool Initialize(I_PoolPlayer &rDuplicateSrc);

	void Terminalize(void);

	bool isInitialized(void) const {return lpDSBuffer != 0;}


	virtual bool Start(void);
	virtual bool Stop(void);
	virtual bool isPlaying(void) const;

	virtual Format GetFormat(void) const {return PlayFormat;}
	virtual const I_PlayDriver *GetDriver(void);

	virtual bool SetVolume(unsigned long iLeft,unsigned long iRight)
	{
		return SetVolumef(static_cast<float>(iLeft)/0x10000,
						  static_cast<float>(iRight)/0x10000);
	}
	virtual bool GetVolume(unsigned long &oLeft,unsigned long &oRight) const
	{
		float fleft,fright;
		bool r = GetVolumef(fleft,fright);
		oLeft = static_cast<unsigned long>(fleft * 0x10000);
		oRight = static_cast<unsigned long>(fright * 0x10000);
		return r;
	}

	virtual void SetLoopFlag(bool loop);
	virtual bool GetLoopFlag(void) const {return LoopFlag;}

	virtual long GetPlayPosition(void);
	virtual bool SetPlayPosition(long iSamplePos);

	virtual unsigned long GetBufferSize(void) const;

	virtual bool LoadStream(I_InputStream &ioStream);


	virtual bool SetVolumef(float iLeft,float iRight);
	virtual bool GetVolumef(float &oLeft,float &oRight) const;
	virtual bool SetVolume_dB(long iLeft,long iRight);
	virtual bool GetVolume_dB(long &oLeft,long &oRight) const;

private:
	bool CreateBuffer(const Format &iFormat,long bufsize,
					  int focus);

private:
	IDirectSoundBuffer *lpDSBuffer;	//Đobt@{́B
	DirectSound::DirectSoundObject *Master;		//obt@DirectSoundObject

	Format PlayFormat;
	bool LoopFlag;

private://
	DirectSoundPoolPlayer(DirectSoundPoolPlayer &);
	void operator=(const DirectSoundPoolPlayer &);
};

}//namespace Sound
}//namespace Juna

#endif
