
#include "../../DirectSoundDriver.h"
#include "../../DirectSoundPlayer.h"

#include "../ExtensibleFormat.h"

#include "DirectSoundPrivate.h"

namespace Juna {
namespace Sound {
using namespace DirectSound;

/*=============================================================================

=============================================================================*/


DirectSoundPlayDriver::DirectSoundPlayDriver(void)
	: list(NULL),wlist(NULL),IndexNum(-1),hDSdll(NULL) {}

DirectSoundPlayDriver::~DirectSoundPlayDriver() {DeleteDeviceList();}

I_StreamPlayer *DirectSoundPlayDriver::CreateStreamPlayer
	(int iDevice,const Format &iFormat,long iBufsize,int iBufnum)
{return CreateDSStreamPlayer(iDevice,iFormat,iBufsize,iBufnum);}

I_StreamPlayer *DirectSoundPlayDriver::CreateStreamPlayer
	(int iDevice,I_InputStream &setStream,long iBufsize,int iBufnum)
{return CreateDSStreamPlayer(iDevice,setStream,iBufsize,iBufnum);}

I_PoolPlayer *DirectSoundPlayDriver::CreatePoolPlayer
	(int iDevice,const Format &iFormat,long iBufSize)
{return CreateDSPoolPlayer(iDevice,iFormat,iBufSize);}

I_PoolPlayer *DirectSoundPlayDriver::CreatePoolPlayer
	(int iDevice,I_SeekableInputStream &iStream)
{return CreateDSPoolPlayer(iDevice,iStream);}

I_PoolPlayer *DirectSoundPlayDriver::CreatePoolPlayer
	(I_PoolPlayer &iDuplicateSrc)
{return CreateDSPoolPlayer(iDuplicateSrc);}

void DirectSoundPlayDriver::DestroyPlayer(I_StreamPlayer *player)
{
	if (!player || player->GetDriver() != this)
		return;
	delete static_cast<DirectSoundStreamPlayer *>(player);
}
void DirectSoundPlayDriver::DestroyPlayer(I_PoolPlayer *player)
{
	if (!player || player->GetDriver() != this)
		return;
	delete static_cast<DirectSoundPoolPlayer *>(player);
}

int DirectSoundPlayDriver::GetDeviceNum(void)
{
	if (IndexNum < 0) InitDeviceNum();
	return IndexNum;
}

const PlayDeviceData *DirectSoundPlayDriver::GetDeviceData(int index)
{return GetDSDeviceData(index);}

const PlayDeviceDataW *DirectSoundPlayDriver::GetDeviceDataW(int index)
{return GetDSDeviceDataW(index);}


I_DirectSoundStreamPlayer *DirectSoundPlayDriver::CreateDSStreamPlayer
	(int iDevice,const Format &iFormat,long iBufsize,int iBufnum,int iFocus)
{
	DirectSoundStreamPlayer *p = new DirectSoundStreamPlayer();
	if (p && !p->Initialize(iDevice,iFormat,iBufsize,iBufnum,iFocus))
	{
		delete p;
		return 0;
	}
	return p;
}
I_DirectSoundStreamPlayer *DirectSoundPlayDriver::CreateDSStreamPlayer
	(int iDevice,I_InputStream &setStream,long iBufsize,int iBufnum,int iFocus)
{
	DirectSoundStreamPlayer *p = new DirectSoundStreamPlayer();
	if (p && !p->Initialize(iDevice,setStream,iBufsize,iBufnum,iFocus))
	{
		delete p;
		return 0;
	}
	return p;
}
I_DirectSoundPoolPlayer *DirectSoundPlayDriver::CreateDSPoolPlayer
	(int iDevice,const Format &iFormat,long iBufSize,int iFocus)
{
	DirectSoundPoolPlayer *p = new DirectSoundPoolPlayer();
	if (p && !p->Initialize(iDevice,iFormat,iBufSize,iFocus))
	{
		delete p;
		return 0;
	}
	return p;
}
I_DirectSoundPoolPlayer *DirectSoundPlayDriver::CreateDSPoolPlayer
	(int iDevice,I_SeekableInputStream &iStream,int iFocus)
{
	DirectSoundPoolPlayer *p = new DirectSoundPoolPlayer();
	if (p && !p->Initialize(iDevice,iStream,iFocus))
	{
		delete p;
		return 0;
	}
	return p;
}
I_DirectSoundPoolPlayer *DirectSoundPlayDriver::CreateDSPoolPlayer
	(I_PoolPlayer &iDuplicateSrc)
{
	DirectSoundPoolPlayer *p = new DirectSoundPoolPlayer();
	if (p && !p->Initialize(iDuplicateSrc))
	{
		delete p;
		return 0;
	}
	return p;
}

const DSoundDeviceData *DirectSoundPlayDriver::GetDSDeviceData(int index)
{
	if (list == NULL) UpdateDeviceList();
	if (index < 0 || index >= IndexNum)
		return NULL;
	return &list[index];
}

const DSoundDeviceDataW *DirectSoundPlayDriver::GetDSDeviceDataW(int index)
{
	if (wlist == NULL) UpdateDeviceListW();
	if (index < 0 || index >= IndexNum)
		return NULL;
	return &wlist[index];
}

const GUID *DirectSoundPlayDriver::GetDSDeviceGUID(int index)
{
	if (wlist)
	{
		const DSoundDeviceDataW *dd = GetDSDeviceDataW(index);
		if (dd == NULL)
			return NULL;
		return &dd->guid;
	}
	else
	{
		const DSoundDeviceData *dd = GetDSDeviceData(index);
		if (dd == NULL)
			return NULL;
		return &dd->guid;
	}
}


void DirectSoundPlayDriver::UpdateDeviceList(void)
{
	DeleteDeviceList();
	InitDeviceNum();
	if (IndexNum <= 0)
		return;

	//pm
	unsigned long needsize = 0;
	struct getsize
	{
		static BOOL CALLBACK DSEnumCallback(LPGUID,LPCSTR lpcstrDesc,LPCSTR lpcstrModule,LPVOID lpContext)
		{
			unsigned long &size = *static_cast<unsigned long *>(lpContext);
			size += ::lstrlenA(lpcstrDesc) + ::lstrlenA(lpcstrModule) + 2;
			return TRUE;
		}
	};
	const typeDSEnumA DSEnumProc = GetEnumProcA();
	if (DSEnumProc == NULL) return;
	DSEnumProc(getsize::DSEnumCallback,&needsize);
	list = static_cast<DSoundDeviceData *>(operator new(sizeof(DSoundDeviceData) * IndexNum + needsize));

	struct local
	{
		struct Data {	//[Jm͌ȂHdȂ
			DSoundDeviceData *list;
			int max,i;
			char *buf;
		};
		static BOOL CALLBACK DSEnumCallback(LPGUID lpGuid,LPCSTR lpcstrDescription,LPCSTR lpcstrModule,LPVOID lpContext)
		{
			Data *data = static_cast<Data *>(lpContext);
			if (data->i >= data->max)
				return FALSE;
			if (lpGuid)
				data->list[data->i].guid = *lpGuid;
			else
				data->list[data->i].guid = GUID_NULL;

			data->list[data->i].name = data->buf;
			lstrcpyA(data->buf,lpcstrDescription);
			data->buf += lstrlenA(data->buf) + 1;

			data->list[data->i].Module = data->buf;
			lstrcpyA(data->buf,lpcstrModule);
			data->buf += lstrlenA(data->buf) + 1;

			data->i++;
			return TRUE;
		}
	};//struct local

	local::Data data = {list,IndexNum,0,reinterpret_cast<char *>(list + IndexNum)};
	DSEnumProc(local::DSEnumCallback,&data);

	return;
}

void DirectSoundPlayDriver::UpdateDeviceListW(void)
{
	DeleteDeviceList();
	InitDeviceNum();
	if (IndexNum <= 0)
		return;

	//pm
	unsigned long needsize = 0;
	struct getsize
	{
		static BOOL CALLBACK DSEnumCallback(LPGUID,LPCWSTR lpcstrDesc,LPCWSTR lpcstrModule,LPVOID lpContext)
		{
			unsigned long &size = *static_cast<unsigned long *>(lpContext);
			//UTF-16ƌ߂ŗǂ񂾂ȁH
			size += (lstrlenW(lpcstrDesc) + lstrlenW(lpcstrModule) + 2) * sizeof(unsigned short);
			return TRUE;
		}
	};
	const typeDSEnumW DSEnumProc = GetEnumProcW();
	if (DSEnumProc == NULL) return;
	DSEnumProc(getsize::DSEnumCallback,&needsize);
	wlist = static_cast<DSoundDeviceDataW *>(operator new(sizeof(DSoundDeviceDataW) * IndexNum + needsize));

	struct local
	{
		struct Data {	//[Jm͌ȂHdȂ
			DSoundDeviceDataW *wlist;
			int max,i;
			LPWSTR buf;
		};
		static BOOL CALLBACK DSEnumCallback(LPGUID lpGuid,LPCWSTR lpcstrDescription,LPCWSTR lpcstrModule,LPVOID lpContext)
		{
			Data *data = static_cast<Data *>(lpContext);
			if (data->i >= data->max)
				return FALSE;
			if (lpGuid)
				data->wlist[data->i].guid = *lpGuid;
			else
				data->wlist[data->i].guid = GUID_NULL;


			data->wlist[data->i].name = data->buf;
			lstrcpyW(data->buf,lpcstrDescription);
			data->buf += lstrlenW(data->buf) + 1;

			data->wlist[data->i].Module = data->buf;
			lstrcpyW(data->buf,lpcstrModule);
			data->buf += lstrlenW(data->buf) + 1;

			data->i++;
			return TRUE;
		}
	};//struct local

	local::Data data = {wlist,IndexNum,0,reinterpret_cast<LPWSTR>(wlist + IndexNum)};
	DSEnumProc(local::DSEnumCallback,&data);

	return;
}

void DirectSoundPlayDriver::DeleteDeviceList(void)
{
	if (list) operator delete(list);
	list = NULL;
	if (wlist) operator delete(wlist);
	wlist = NULL;
	if (hDSdll)
		FreeLibrary(hDSdll);
	hDSdll = NULL;
	IndexNum = -1;
}

bool DirectSoundPlayDriver::TestOpenDevice(int index,const Format &format)
{
	if ((format.bit == 8 && format.quantization == QT_SIGNED) ||
		(format.bit != 8 && format.quantization == QT_UNSIGNED))
		return false;
//DirectSoundObject̊l
	const GUID *guid = DirectSoundPlayDriver::GetInstance().GetDSDeviceGUID(index);
	if (guid == NULL)
		return false;
	DirectSoundObject *dso = DirectSoundObjectManager::GetInstance().GetDSObject(*guid);
	if (dso == NULL)
	{
		dso = DirectSoundObjectManager::GetInstance().InitializeDSObject(*guid,format);
		if (dso == NULL)
			return false;
	}

	DWORD dsbdflag = DSBCAPS_GETCURRENTPOSITION2 | DSBCAPS_CTRLVOLUME | DSBCAPS_CTRLPAN | DSBCAPS_CTRLPOSITIONNOTIFY;
	dsbdflag |= DSBCAPS_LOCSOFTWARE;
	dsbdflag |= DSBCAPS_GLOBALFOCUS;

	WAVEFORMATEX wfx;
	SetWaveFormatEx(wfx,format);

	DSBUFFERDESC dsbd = {sizeof(dsbd),dsbdflag,format.sample_size() * 1,0,&wfx};

	LPDIRECTSOUNDBUFFER lpBuffer;
	if (format.bit > 16 || format.channel > 2 ||
		FAILED(dso->GetDSObject()->CreateSoundBuffer(&dsbd,&lpBuffer,NULL)))
	{	//ExtensibleFormat
		const int extsize = sizeof(WAVEFORMATEX) + 22;
		BYTE buf[extsize];
		WAVEFORMATEX *wfext = reinterpret_cast<WAVEFORMATEX *>(buf);
		memcpy(wfext,&wfx,sizeof(WAVEFORMATEX));
		wfext->cbSize = extsize - sizeof(WAVEFORMATEX);
		if (!MakeExtensibleFormat(wfext))
		{
			dso->Release();
			return false;
		}
		dsbd.lpwfxFormat = wfext;
		if FAILED(dso->GetDSObject()->CreateSoundBuffer(&dsbd,&lpBuffer,NULL))
		{
			dso->Release();
			return false;
		}
	}
	lpBuffer->Release();
	dso->Release();
	return true;
}

const DirectSoundPlayDriver::typeDSCreate DirectSoundPlayDriver::GetDSCreateProc(void)
{
	if (LoadDll() == NULL)
		return NULL;
	return  (typeDSCreate)GetProcAddress(hDSdll,"DirectSoundCreate");
}



HINSTANCE DirectSoundPlayDriver::LoadDll(void)
{
	if (hDSdll == NULL) hDSdll= LoadLibraryA("dsound");
	return hDSdll;
}
void DirectSoundPlayDriver::FreeDll(void)
{
	if (hDSdll) FreeLibrary(hDSdll);
	hDSdll = NULL;
}

const DirectSoundPlayDriver::typeDSEnumA DirectSoundPlayDriver::GetEnumProcA(void)
{
	if (LoadDll() == NULL) return NULL;
	const typeDSEnumA DSEnumProc = (typeDSEnumA)GetProcAddress(hDSdll,"DirectSoundEnumerateA");
	if (DSEnumProc == NULL) FreeDll();
	return DSEnumProc;
}
const DirectSoundPlayDriver::typeDSEnumW DirectSoundPlayDriver::GetEnumProcW(void)
{
	if (LoadDll() == NULL) return NULL;
	const typeDSEnumW DSEnumProc = (typeDSEnumW)GetProcAddress(hDSdll,"DirectSoundEnumerateW");
	if (DSEnumProc == NULL) FreeDll();
	return DSEnumProc;
}

void DirectSoundPlayDriver::InitDeviceNum(void)
{
	struct local {	//ȂƂȂ[J֐
		static BOOL CALLBACK DSEnumCallback(LPGUID,LPCSTR,LPCSTR,LPVOID lpContext)
		{(*static_cast<int *>(lpContext))++;return TRUE;}
	};

	IndexNum = -1;
	const typeDSEnumA DSEnumProc = GetEnumProcA();
	if (DSEnumProc == NULL)
		return;
	IndexNum = 0;
	DSEnumProc(local::DSEnumCallback,&IndexNum);
	return;
}



DirectSoundPlayDriver &DirectSoundPlayDriver::GetInstance(void)
{
	static DirectSoundPlayDriver Driver;
	return Driver;
}

I_DirectSoundPlayDriver *GetDirectSoundPlayDriver(void)
{
	return &DirectSoundPlayDriver::GetInstance();
}



}//namespace Sound
}//namespace Juna
