
#include "../../DirectSoundPlayer.h"
#include "../../DirectSoundDriver.h"

#include "../ExtensibleFormat.h"

#include "DirectSoundPrivate.h"

#include <math.h>
#include <process.h>

#include "EmptySupplier.h"

namespace Juna {
namespace Sound {
using namespace DirectSound;


namespace {
class Synchronizer
{
public:
	Synchronizer(CRITICAL_SECTION &locker) : Locker(locker)
	{::EnterCriticalSection(&Locker);}
	~Synchronizer(void)
	{::LeaveCriticalSection(&Locker);}

//UnlockԂŃfXgN^Ă΂Ȃ悤
	void Unlock(void) {::LeaveCriticalSection(&Locker);}
	void Lock(void) {::EnterCriticalSection(&Locker);}
private:
	CRITICAL_SECTION &Locker;
};

class NullCallbackClass : public I_StreamPlayer::I_PlayEndCallback
{
public: virtual void PlayEndCallback(void) {}
};

NullCallbackClass NullCallback; 

}//anonymous namespace


/*=============================================================================

=============================================================================*/

DirectSoundStreamPlayer::DirectSoundStreamPlayer(void)
  : lpDSBuffer(NULL),Master(NULL),
	Stream(&EmptySupplier::GetInstance()),
	BufferSize(0),BufferNum(0),
	hNotifyEvent(NULL),hThread(NULL),ThreadState(TS_EXIT),
	PlayFormat(),
	TotalLoadBufferNum(0),WriteBufferPos(-1),
	LoadEndPos(-1),LoadEndSize(0),
	PlayEndCallback(&NullCallback)
{
	::InitializeCriticalSection(&CritSect);
}

DirectSoundStreamPlayer::~DirectSoundStreamPlayer()
{
	Terminalize();
	::DeleteCriticalSection(&CritSect);
}

//Xbh֐NX֐ɂӖ́ApravateoɃANZX邽߂ɂ̂B
//mfriend̓JbRC
unsigned __stdcall DirectSoundStreamPlayer::PlayThread(void *lpDSPlayer)
{
	DirectSoundStreamPlayer &this_ = *static_cast<DirectSoundStreamPlayer *>(lpDSPlayer);

	for (;;)
	{
		::WaitForSingleObject(this_.hNotifyEvent,INFINITE);

		Synchronizer sync(this_.CritSect);
		switch (this_.ThreadState)
		{
		case TS_EXIT:
			return 0;

		case TS_PLAY:
			if (this_.LoadEndPos >= 0)	//SẴf[^ǂݍݏII҂
			{
				DWORD play;
				this_.lpDSBuffer->GetCurrentPosition(&play,NULL);

				DWORD lep = (DWORD)this_.LoadEndPos;
				lep = (lep + this_.BufferNum - 1) % this_.BufferNum;
//ŏIĐobt@ɓBA|[Oő҂ďI点
				if (play / this_.BufferSize == lep)
				{
					lep = lep * this_.BufferSize + this_.LoadEndSize;
					this_.ThreadState = TS_PLAYEND;
					do {
						DWORD play2;
						this_.lpDSBuffer->GetCurrentPosition(&play2,NULL);
//IʒuzAA擪܂Ŋ߂ĂAłĎ~߂
						if ((play2 > lep) || (lep - play2 > (DWORD)this_.BufferSize))
						{
							this_.lpDSBuffer->Stop();
							this_.lpDSBuffer->SetCurrentPosition(0);
							this_.WriteBufferPos = -1;
							this_.ThreadState = TS_STANDBY;
							this_.PlayEndCallback->PlayEndCallback();
							break;
						}
						sync.Unlock();
						::Sleep(1);
						sync.Lock();
					} while (this_.ThreadState == TS_PLAYEND);
				}
				else
				{
	//ÔߖŖ߂BłyȂ疳ׂ
					this_.FillSilent();
				}
			}
			else 
			{
				if (this_.LoadData_OverCheck() == 0)
					this_.LoadData_OverCheck();
			}
			break;//case TS_PLAY:
		}
	}
//BR[h
//	return 0;
}

const I_PlayDriver *DirectSoundStreamPlayer::GetDriver(void)
{return &DirectSoundPlayDriver::GetInstance();}

bool DirectSoundStreamPlayer::Start(void)
{
	Synchronizer sync(CritSect);
	if (ThreadState == TS_PLAYEND)
		return true;
	if (WriteBufferPos < 0)
	{
		if (LoadEndPos >= 0)
			return true;
		WriteBufferPos = 0;
		if (LoadData_NoCheck() == -1)
		{
			WriteBufferPos = -1;
			return false;
		}
	}
	ThreadState = TS_PLAY;
	return SUCCEEDED(lpDSBuffer->Play(0,0,DSBPLAY_LOOPING));
}

bool DirectSoundStreamPlayer::Stop(void)
{
	Synchronizer sync(CritSect);
	if (ThreadState == TS_PLAY || ThreadState == TS_PLAYEND)
	{
		ThreadState = TS_PAUSE;
		lpDSBuffer->Stop();
	}
	return true;
}

bool DirectSoundStreamPlayer::isPlaying(void) const
{
//	Synchronizer sync(CritSect);
	DWORD st;
	if SUCCEEDED(lpDSBuffer->GetStatus(&st))
		return (st & DSBSTATUS_PLAYING);
	return false;
//	return 	(ThreadState & TS_PLAY) != 0;
}


long DirectSoundStreamPlayer::GetPlayTime(void)
{
	Synchronizer sync(CritSect);
	int samplesize = PlayFormat.sample_size();
//ByteSampleɕϊ邽߃tH[}bg߂
	DWORD bufsamp = BufferSize / samplesize;	//obt@ЂƂsample
	long tp = bufsamp * TotalLoadBufferNum;		//ǂݍ񂾃f[^̗ʂ߂

	if (WriteBufferPos < 0)	//~܂Ă
		return tp + LoadEndSize / samplesize;	//ǂݍݗʂ

//obt@̍Đʒu߂
	DWORD pcs;
	if FAILED(lpDSBuffer->GetCurrentPosition(&pcs,NULL))
		return -1;
	pcs /= samplesize;	//obt@̐擪sample
	long playtime = tp;	//Ƃ肠f[^ǂݍ݈ʒuĂ

	if (LoadEndPos > 0)	//ǂݍ݂IĂ
	{
		const unsigned long les = LoadEndSize / samplesize;
		playtime += les;
		int e = LoadEndPos;
		if (pcs > e * bufsamp)
			playtime -= bufsamp * BufferNum;
		e = (e + BufferNum - 1) % BufferNum;
		if (pcs / bufsamp == (unsigned)e && pcs % bufsamp > les)
			pcs = e * bufsamp + les;
		playtime += pcs - e * bufsamp;
		const long total = tp + les;
		if (playtime > total)
			playtime = total;
		return playtime;
	}

	DWORD wbps = WriteBufferPos * bufsamp;
	if (pcs > wbps)	//ĐʒuƓǂݍ݈ʒu߂
		playtime -= bufsamp * BufferNum;
	playtime += pcs - wbps;

	return playtime;
}

void DirectSoundStreamPlayer::SetPlayEndCallback(I_PlayEndCallback *setCallback)
{
	if (setCallback)
		PlayEndCallback = setCallback;
	else
		PlayEndCallback = &NullCallback;
}


//_ŉʐݒ[1.0 - 0.00001]
bool DirectSoundStreamPlayer::SetVolumef(float left,float right)
{
//	Synchronizer sync(CritSect);

//0.0log100Ԃ₪邩玩ŒׂĂ݂B
//Ȃ݂SetVolumeoRȂ΂Ƀ}CiXɂ͂ȂȂ璲ׂȂB
	if (max(left,right) <= 0.00001)
		return SUCCEEDED(lpDSBuffer->SetVolume(-10000));

//ΐnfVxɕϊAłĂ̂ȁH
	LONG dleft = static_cast<LONG>(100.0 * 20.0 * log10(left));
	LONG dright = static_cast<LONG>(100.0 * 20.0 * log10(right));
	if FAILED(lpDSBuffer->SetVolume(min(max(dleft,dright),0)))
		return false;
	if (PlayFormat.channel > 1 && FAILED(lpDSBuffer->SetPan(dright - dleft)))
	{}
	return true;
}
bool DirectSoundStreamPlayer::GetVolumef(float &left,float &right) const
{
//	Synchronizer sync(CritSect);
	LONG dleft,dright;
	if FAILED(lpDSBuffer->GetVolume(&dleft))
		return false;
	LONG pan;
	if (PlayFormat.channel == 1 || FAILED(lpDSBuffer->GetPan(&pan)))
	{
//fVxwaveOutnŎgĂΐnɕϊ	ŗǂ̂H悭
		left = right =
			static_cast<float>(pow(10.0,(static_cast<double>(dleft) / 100.0) / 20.0));
		return true;
	}
	dright = dleft;
	if (pan > 0)
		dleft -= pan;
	else
		dright += pan;
	left = static_cast<float>(pow(10.0,(static_cast<double>(dleft) / 100.0) / 20.0));
	right = static_cast<float>(pow(10.0,(static_cast<double>(dright) / 100.0) / 20.0));
	return true;
}


bool DirectSoundStreamPlayer::SetVolume_dB(long left,long right)
{
//	Synchronizer sync(CritSect);
	if FAILED(lpDSBuffer->SetVolume(max(max(left,right),DSBVOLUME_MIN)))
		return false;
	if (PlayFormat.channel > 1 && FAILED(lpDSBuffer->SetPan(right - left)))
	{}
	return true;
}
bool DirectSoundStreamPlayer::GetVolume_dB(long &left,long &right) const
{
//	Synchronizer sync(CritSect);
	if FAILED(lpDSBuffer->GetVolume(&left))
		return false;
	right = left;
	LONG pan;
	if (PlayFormat.channel == 1 || FAILED(lpDSBuffer->GetPan(&pan)))
		return true;

	if (pan > 0)
		left -= pan;
	else
		right += pan;
	return true;
}


bool DirectSoundStreamPlayer::Initialize
	(int device,const Format &format,long bufsize,int bufnum,int focus)
{
	if (lpDSBuffer)
	{
		SetSupplier(0);
		lpDSBuffer->Release();
		lpDSBuffer = NULL;
	}

	if (Master) {
		Master->Release();
		Master = NULL;
	}
	{	//DirectSoundObject̊l
		const GUID *guid = DirectSoundPlayDriver::GetInstance().GetDSDeviceGUID(device);
		if (guid == NULL)
			goto ERR;

		DirectSoundObject *dso = DirectSoundObjectManager::GetInstance().GetDSObject(*guid);
		if (dso == NULL)
		{
			dso = DirectSoundObjectManager::GetInstance().InitializeDSObject(*guid,format);
			if (dso == NULL)
				goto ERR;
		}
		Master = dso;
	}

	if (hNotifyEvent == NULL)
	{
		hNotifyEvent = ::CreateEvent(NULL,FALSE,FALSE,NULL);
		if (hNotifyEvent == NULL)
			goto ERR;
	}

	if (!CreateBuffer(format,bufsize,bufnum,focus))
		goto ERR;

	if (hThread == NULL)
	{
		ThreadState = TS_STANDBY;
		unsigned threadID;
		hThread = (HANDLE)::_beginthreadex(NULL,0,PlayThread,this,0,&threadID);
		if (hThread == NULL)
			goto ERR;
		::SetThreadPriority(hThread,THREAD_PRIORITY_ABOVE_NORMAL);
	}

	WriteBufferPos = -1;
	LoadEndPos = -1;
	LoadEndSize = 0;

	PlayFormat = format;
	return true;

ERR:;
	Terminalize();
	return false;
}

bool DirectSoundStreamPlayer::CreateBuffer
	(const Format &format,long bufsize,int bufnum,int focus)
{
	if ((format.bit == 8 && format.quantization == QT_SIGNED) ||
		(format.bit != 8 && format.quantization == QT_UNSIGNED))
		return false;

	if (bufsize == 0)	//ftHgł0w肳ƎIɌ
	{
		bufsize = format.frequency * format.sample_size() * 1;
	}
//TvTCYɍ킹
	if (bufsize % format.sample_size())
		bufsize -= bufsize % format.sample_size();
	BufferSize = bufsize;
	BufferNum = bufnum;

	DWORD dsbdflag = DSBCAPS_GETCURRENTPOSITION2 | DSBCAPS_CTRLVOLUME | DSBCAPS_CTRLPAN | DSBCAPS_CTRLPOSITIONNOTIFY;
	dsbdflag |= DSBCAPS_LOCSOFTWARE;
//tH[JXݒ
	if (focus < 0)
		dsbdflag |= DSBCAPS_STICKYFOCUS;
	else if (focus > 0)
		dsbdflag |= DSBCAPS_GLOBALFOCUS;

	WAVEFORMATEX wfx;
	SetWaveFormatEx(wfx,format);

	DSBUFFERDESC dsbd = {sizeof(dsbd),dsbdflag,bufsize * bufnum,0,&wfx};

	LPDIRECTSOUNDBUFFER lpBuffer = NULL;
	if (format.bit > 16 || format.channel > 2 ||
		FAILED(Master->GetDSObject()->CreateSoundBuffer(&dsbd,&lpBuffer,NULL)))
	{	//ExtensibleFormat
		const int extsize = sizeof(WAVEFORMATEX) + 22;
		BYTE buf[extsize];
		WAVEFORMATEX *wfext = reinterpret_cast<WAVEFORMATEX *>(buf);
		::memcpy(wfext,&wfx,sizeof(WAVEFORMATEX));
		wfext->cbSize = extsize - sizeof(WAVEFORMATEX);
		if (!MakeExtensibleFormat(wfext))
			return false;
		dsbd.lpwfxFormat = wfext;
		if FAILED(Master->GetDSObject()->CreateSoundBuffer(&dsbd,&lpBuffer,NULL))
			return false;
	}
	lpDSBuffer = lpBuffer;

//NotifyEventݒ
	LPDIRECTSOUNDNOTIFY lpdsNotify = NULL;
	DSBPOSITIONNOTIFY *dsbpn = new DSBPOSITIONNOTIFY[bufnum];
	for (int i = 0;i < bufnum;i++)
	{
		dsbpn[i].dwOffset = bufsize * i;
		dsbpn[i].hEventNotify = hNotifyEvent;
	}
	if FAILED(lpBuffer->QueryInterface(IID_IDirectSoundNotify,(VOID **)&lpdsNotify))
		goto ERR;
	if FAILED(lpdsNotify->SetNotificationPositions(bufnum,dsbpn))
		goto ERR;

	lpdsNotify->Release();	//ݒ肳ł΂̃C^[tFCX͂
	delete[] dsbpn;

	return true;

ERR:;
	if (lpdsNotify)
		lpdsNotify->Release();
	if (dsbpn != NULL)
		delete[] dsbpn;
	return false;
}

void DirectSoundStreamPlayer::Terminalize(void)
{
	if (lpDSBuffer)
		SetSupplier(0);
	if (hThread)
	{
//ʂɌJԂKvȂǁBقƂ͈莞ԌTerminateThreadŋIƂ
		do
		{
			ThreadState = TS_EXIT;		//XbhItOZbg
			::SetEvent(hNotifyEvent);	//蓮őҋ@Ԃ
//Î҂AIĂȂx
		} while (WaitForSingleObject(hThread,10) == WAIT_TIMEOUT);
		::CloseHandle(hThread);
		hThread = NULL;
	}
	if (hNotifyEvent) {
		::CloseHandle(hNotifyEvent);
		hNotifyEvent = NULL;
	}
	if (lpDSBuffer) {
		lpDSBuffer->Release();
		lpDSBuffer = NULL;
	}
	if (Master) {
		Master->Release();
		Master = NULL;
	}
	TotalLoadBufferNum = 0;
	PlayFormat.reset();
	BufferSize = BufferNum = 0;
}

bool DirectSoundStreamPlayer::SetSupplier(I_InputStream *setStream)
{
	Stop();
	Synchronizer sync(CritSect);

	ThreadState = TS_STANDBY;
	lpDSBuffer->SetCurrentPosition(0);
	WriteBufferPos = -1;
	TotalLoadBufferNum = 0;
	LoadEndPos = -1;
	LoadEndSize = 0;

	if (setStream == Stream)
		return true;
	if (setStream)
	{
		if (PlayFormat != setStream->GetFormat())
			return false;
	}
	else
		setStream = &EmptySupplier::GetInstance();

	setStream->AddRef();
	Stream->Release();
	Stream = setStream;
	return true;
}

bool DirectSoundStreamPlayer::Initialize
	(int device,I_InputStream &stream,long bufsize,int bufnum,int focus)
{
	if (!Initialize(device,stream.GetFormat(),bufsize,bufnum,focus))
		return false;
	if (!SetSupplier(&stream))
	{
		Terminalize();
		return false;
	}
	return true;
}

int DirectSoundStreamPlayer::LoadStream(void)
{
	if (LoadEndPos >= 0)
		return 2;

	if (WriteBufferPos < 0)
	{
		WriteBufferPos = 0;
		int r = LoadData_NoCheck();
		if (r < 0)
			WriteBufferPos = -1;
		return r;
	}
	return LoadData_OverCheck();
}

int DirectSoundStreamPlayer::LoadData_OverCheck(void)
{
	{
		DWORD play,write;
		lpDSBuffer->GetCurrentPosition(&play,&write);
//͈͓蔻
//		long center = WriteBufferPos * BufferSize + (BufferSize >> 1);
//		if (abs(center - play) < (BufferSize >> 1) ||
//			abs(center - write) < (BufferSize >> 1))
//			return 1;

//		play = ((play - 1 + BufferSize) / BufferSize - 1 + BufferNum) % BufferNum;
//		write = ((write - 1 + BufferSize) / BufferSize - 1 + BufferNum) % BufferNum;
//		if (play == (DWORD)WriteBufferPos || write == (DWORD)WriteBufferPos)
//			return 1;	//߂Ȃ

		play /= BufferSize;
		write /= BufferSize;
		if (play == (DWORD)WriteBufferPos || write == (DWORD)WriteBufferPos)
			return 1;	//߂Ȃ

	}
	return LoadData_NoCheck();
}

int DirectSoundStreamPlayer::LoadData_NoCheck(void)
{
	BYTE *lock;
	DWORD size;
	if FAILED(lpDSBuffer->Lock(BufferSize * WriteBufferPos,BufferSize,(void **)&lock,&size,NULL,NULL,0))
	{
		return -1;
	}
	if (size < BufferSize)	//N肦Ȃ͂
	{
		lpDSBuffer->Unlock(lock,0,NULL,0);
		return -1;
	}

	unsigned long readsize = Stream->Read(lock,BufferSize);

	if (readsize < BufferSize)	//IHȂĂ
	{	//Ȃ͖
		int fill = 128 & -(PlayFormat.bit == 8);
		::memset(lock + readsize,fill,BufferSize - readsize);
	}
	lpDSBuffer->Unlock(lock,BufferSize,NULL,0);
	WriteBufferPos = (WriteBufferPos + 1) % BufferNum;
	if (Stream->isEnd())	//ŏIf[^
	{
		LoadEndSize = readsize;
		LoadEndPos = WriteBufferPos;
		return 2;
	}
	TotalLoadBufferNum++;
	return 0;
}



int DirectSoundStreamPlayer::FillSilent(void)
{
	DWORD play;

	lpDSBuffer->GetCurrentPosition(&play,NULL);
	play /= BufferSize;

	if (play == (DWORD)WriteBufferPos)
		return 1;	//f[^^

	BYTE *lock;
	DWORD size;
	if (FAILED(lpDSBuffer->Lock(BufferSize * WriteBufferPos,BufferSize,(void **)&lock,&size,NULL,NULL,0)))
		return -1;
	if (size < (DWORD)BufferSize)	//N肦ȂƎv
	{
		lpDSBuffer->Unlock(lock,0,NULL,0);
		return -1;
	}

	WriteBufferPos = (WriteBufferPos + 1) % BufferNum;
	int fill = 128 & -(PlayFormat.bit == 8);
	::memset(lock,fill,BufferSize);
	lpDSBuffer->Unlock(lock,BufferSize,NULL,0);
	return 0;
}


}//namespace Sound
}//namespace Juna
