
#include "../../DirectSoundDriver.h"
#include "DirectSoundPrivate.h"

#include "../ExtensibleFormat.h"

namespace Juna {
namespace Sound {
namespace DirectSound {


bool DirectSoundObject::ResetFormat(const Format &format)
{
	if (lpPrimaryBuffer == NULL)
		return false;
//vC}BTEhobt@̃tH[}bgݒ

	if ((format.bit == 8 && format.quantization == QT_SIGNED) ||
		(format.bit != 8 && format.quantization == QT_UNSIGNED))
		return false;
	WAVEFORMATEX wfx;
	SetWaveFormatEx(wfx,format);

	if FAILED(lpPrimaryBuffer->SetFormat(&wfx))
	{
		const int extsize = sizeof(WAVEFORMATEX) + 22;
		BYTE buf[extsize];
		WAVEFORMATEX *wfext = reinterpret_cast<WAVEFORMATEX *>(buf);
		memcpy(wfext,&wfx,sizeof(WAVEFORMATEX));
		wfext->cbSize = extsize - sizeof(WAVEFORMATEX);
		if (!MakeExtensibleFormat(wfext))
			return false;
		if FAILED(lpPrimaryBuffer->SetFormat(wfext))
			return false;
	}

	return true;
}

bool DirectSoundObject::Initialize(REFGUID guid,const Format &format)
{
	if (lpDSObject)	//łɏς݂Ȃ
		return false;

	const DirectSoundPlayDriver::typeDSCreate DSCreate =
				DirectSoundPlayDriver::GetInstance().GetDSCreateProc();
	if (DSCreate == NULL)
		return false;
	const GUID *lpguid = IsEqualGUID(guid,GUID_NULL) ? NULL : &guid;
	if (FAILED(DSCreate(lpguid,&lpDSObject,NULL))) {
		lpDSObject = NULL;
		return false;
	}
	DeviceGUID = guid;
//
	{	//SetCoopelativeLevelׂ̈̃EChEnhTI
		HWND hwnd = NULL;
		HWND hsearch = NULL;
		{	//EnumThreadWindowsŒT
			struct local {	//Enum̂
				static BOOL CALLBACK ETWP(HWND hwnd,LPARAM lParam)
				{(*(HWND *)lParam) = hwnd;return FALSE;}
			};
			HWND hsearch = NULL;
			EnumThreadWindows(GetCurrentThreadId(),local::ETWP,(LPARAM)&hsearch);
			hwnd = hsearch;
		}
		if (hwnd == NULL)
		{	//EnumWindows + GetWindowThreadProcessIdŒT
			struct local
			{
				static BOOL CALLBACK EWP(HWND hwnd,LPARAM lParam)
				{
					DWORD wpid;
					GetWindowThreadProcessId(hwnd,&wpid);
					if (wpid == GetCurrentProcessId()) {
						(*(HWND *)lParam) = hwnd;
						return FALSE;
					}
					return TRUE;
				}
			};
			EnumWindows(local::EWP,(LPARAM)&hsearch);
			hwnd = hsearch;
			if (hwnd == NULL)	//HȂdȂ̂ōŌ̎i
				hwnd = GetDesktopWindow();
		}
		if FAILED(lpDSObject->SetCooperativeLevel(hwnd,DSSCL_PRIORITY)) {
			Release();
			return false;
		}
	}

//vC}BTEhobt@̍쐬
	DSBUFFERDESC dsbd = {sizeof(dsbd),DSBCAPS_PRIMARYBUFFER,0,0,NULL};
#if DIRECTSOUND_VERSION >= 0x0700
	dsbd.guid3DAlgorithm = DS3DALG_DEFAULT;
#endif
	if FAILED(lpDSObject->CreateSoundBuffer(&dsbd,&lpPrimaryBuffer,NULL))
	{
		lpPrimaryBuffer = NULL;
		Release();
		return false;
	}
	if (!ResetFormat(format)) {
		Release();
		return false;
	}

	return true;
}

int DirectSoundObject::Release(void)
{
	ULONG r = 0;
	if (lpDSObject)
	{
		if (lpPrimaryBuffer)
			lpPrimaryBuffer->Release();
		r = lpDSObject->Release();
		if (r == 0)
		{
			lpDSObject = NULL;
			lpPrimaryBuffer = NULL;
			DeviceGUID = GUID_NULL;
		}
	}
	return r;
}

int DirectSoundObject::AddRef(void)
{
	lpPrimaryBuffer->AddRef();
	return lpDSObject->AddRef();
}



DirectSoundObjectManager::DirectSoundObjectManager(void)
  : ListNum(DirectSoundPlayDriver::GetInstance().GetDeviceNum()),
	List(ListNum > 0 ? new DirectSoundObject[ListNum] : NULL)
{
}

DirectSoundObjectManager::~DirectSoundObjectManager()
{
	delete[] List;
}


DirectSoundObject *DirectSoundObjectManager::GetDSObject(REFGUID guid)
{
	for (int i = 0;i < ListNum;i++)
	{
		if (List[i].lpDSObject == NULL)
			continue;
		if (IsEqualGUID(List[i].DeviceGUID,guid)) {
			List[i].AddRef();
			return List + i;
		}
	}
	return NULL;
}
DirectSoundObject *DirectSoundObjectManager::InitializeDSObject(REFGUID guid,const Format &format)
{
	for (int i = 0;i < ListNum;i++)
	{
		if (List[i].lpDSObject == NULL)
		{
			if (List[i].Initialize(guid,format))
				return List + i;
			return NULL;
		}
	}
	return NULL;
}

DirectSoundObjectManager &DirectSoundObjectManager::GetInstance(void)
{
	static DirectSoundObjectManager Maker;
	return Maker;
}


}//namespace DirectSound
}//namespace Sound
}//namespace Juna

