
#include "MachineGun.h"

#include "Entity.h"
#include "Visual.h"
#include "Attack.h"
#include "Generator.h"
#include "Allocator.h"
#include "VisualizeParam.h"

#include <new>
#include "MersenneTwister.h"

class MachineGunShot : public I_Entity, public I_Visual, public I_Attacker
{
private:
	~MachineGunShot() {}
public:
	MachineGunShot(Generator &generator,
				   const Point &position,RadianAngle dir,float speed);

	virtual bool Update(void);
	virtual void Break(void);
	virtual void Visualize(const VisualizeParam &param) const;
	virtual void Attack(const AttackParam &param);

	const Point &position(void) const {return Position;}
	RadianAngle direction(void) const {return Direction;}
	float range(void) const {return Range;}

private:
	Point Position;
	RadianAngle Direction;
	Vector Velocity;
	float Range;

	uint32 Power;
	int32 LifeCounter;
	Generator &Gen;
};

MachineGunShot::MachineGunShot(
	Generator &generator,
   const Point &pos,RadianAngle dir,float speed)
  : I_Entity(generator.Entity),
	I_Visual(generator.Visual,VL_Top),
	I_Attacker(generator.Attacker),
	Position(pos), Direction(dir), Velocity(dir.cos() * speed,dir.sin() * speed),
	Range(2), Power(10), LifeCounter(24),
	Gen(generator)
{
}
void MachineGunShot::Break(void)
{
	I_Allocator &a = Gen.Allocator;
	this->~MachineGunShot();
	a.Deallocate(this);
}

bool MachineGunShot::Update(void)
{
	LifeCounter--;
	Position += Velocity;
	return LifeCounter > 0;
}
void MachineGunShot::Attack(const AttackParam &param)
{
	struct Circle : public Juna::D2::Intersection::I_Circle
	{
		Circle(const MachineGunShot &shot) : shot(shot) {}
		virtual Data GetCircleData(void) const
		{
			return Data(shot.position(),shot.range());
		}
		const MachineGunShot &shot;
	};
	Circle c(*this);

	typedef AttackParam::list list;
	const list &enemy = param.Enemy;
	for (list::const_iterator it = enemy.begin();it != enemy.end();++it)
	{
		if (it->attackedShape().Test(c))
		{
			Damage(*it,Power,Vector(0,0));
			LifeCounter = 0;
		}
	}
}
void MachineGunShot::Visualize(const VisualizeParam &param) const
{
	param.Screen.Blit(Position.x,Position.y,
		param.Sprite,Rect(96,64,Size(16,16)),Direction.radian());
}




MachineGun::MachineGun(void)
  : Gen(0), BulletCount(0), Reload(false)
{
}

MachineGun::~MachineGun(void)
{
}

void MachineGun::BootUp(Generator &generator)
{
	Gen = &generator;
	BulletCount = 0;
	Reload = false;
}
void MachineGun::Active(bool trigger,const Point &pos,RadianAngle dir)
{
	if (BulletCount > 128)
		Reload = true;
	if (Reload)
	{
		BulletCount -= 2;
		if (BulletCount < 0)
		{
			Reload = false;
			BulletCount = 0;
		}
	}
	else
	{
		if (trigger)
		{
			static MT::Generator rand;
			I_Allocator &a = Gen->Allocator;
			assert(a.blockSize() >= sizeof(MachineGunShot));
			if (!a.full())
			{
				float r = (rand.GenerateReal1() - 0.5) * (Pi / 8);
				new(a.Allocate()) MachineGunShot(*Gen,pos,dir + r,16);
			}
			BulletCount++;
		}
	}
}
bool MachineGun::ShutDown(void)
{
	return true;
}

void MachineGun::Break(void)
{
	Gen = 0;
}
