
#ifndef Mersenne_Twister_Random_Generator_H
#define Mersenne_Twister_Random_Generator_H

namespace MT {

class Generator
{
public:
	Generator(unsigned long seed = 1) : mti(N+1) {Initialize(seed);}
	Generator(unsigned long init_key[], int key_length)
		: mti(N+1) {Initialize(init_key,key_length);}

//	[0,0xffffffff]
	unsigned long GenerateInt32(void);
//	[0,0x7fffffff]
	long GenerateInt31(void) {return GenerateInt32()>>1;}
//	[0,1]
	double GenerateReal1(void) {return GenerateInt32()*(1.0/4294967295.0);}
//	[0,1)
	double GenerateReal2(void) {return GenerateInt32()*(1.0/4294967296.0);}
//	(0,1)
	double GenerateReal3(void) {return (((double)GenerateInt32()) + 0.5)*(1.0/4294967296.0);}
//	[0,1)  (53bit)
	double GenerateRes53(void) 
	{ 
	    unsigned long a=GenerateInt32()>>5, b=GenerateInt32()>>6; 
	    return(a*67108864.0+b)*(1.0/9007199254740992.0); 
	} 


	void Initialize(unsigned long seed);
	void Initialize(unsigned long init_key[], int key_length);
private:
	enum {N = 624};
	unsigned long mt[N];
	int mti;
};



void init_genrand(unsigned long s);
void init_by_array(unsigned long init_key[], int key_length);

unsigned long genrand_int32(void);
long genrand_int31(void);
double genrand_real1(void);
double genrand_real2(void);
double genrand_real3(void);
double genrand_res53(void);


}//namespace MT

#endif
