
#ifndef Game_Observe_H
#define Game_Observe_H

#include "Juna/intrusive_list.hpp"
#include "TypeDef.h"

class I_Observed;
class I_Observer;

class ObserveParam;

class I_ObservedCreator;
class I_ObserverCreator;

class ObserveManager;

typedef Juna::intrusive_list<I_Observed> ObservedList;
typedef Juna::intrusive_list<I_Observer> ObserverList;


class I_Observed : public Juna::intrusive_list_node
{
public:
	enum ObservedType {
		OT_UNKNOWN,
	};
	virtual const Point &position(void) const = 0;
	virtual RadianAngle direction(void) const = 0;
	virtual const Vector &velocity(void) const = 0;

//	virtual const I_Shape &shape(void) const = 0;

	virtual ObservedType type(void) const {return OT_UNKNOWN;}
protected:
	I_Observed(I_ObservedCreator &creater);
	~I_Observed() {unlink();}
private:
	using Juna::intrusive_list_node::unlink;
};

class I_Observer : public Juna::intrusive_list_node
{
private: friend class ObserveManager;
	virtual void Observe(const ObserveParam &param) = 0;

protected:
	I_Observer(I_ObserverCreator &creator);
	~I_Observer() {unlink();}
private:
	using Juna::intrusive_list_node::unlink;
};

class ObserveParam
{
public:
	typedef Juna::intrusive_list<const I_Observed> list;

	const list &Enemy;
	const list &Friend;

	ObserveParam(const list &enemy,const list &friend_)
		: Enemy(enemy),Friend(friend_) {}
private:
	void operator=(ObserveParam &);
};


class I_ObservedCreator
{
private: friend class I_Observed;
	void Create(I_Observed &observed) {Creature.push_back(observed);}
private: friend class ObserveManager;
	Juna::intrusive_list<I_Observed> Creature;
};
class I_ObserverCreator
{
private: friend class I_Observer;
	void Create(I_Observer &observer) {Creature.push_back(observer);}
private: friend class ObserveManager;
	Juna::intrusive_list<I_Observer> Creature;
};

inline I_Observed::I_Observed(I_ObservedCreator &creator) {creator.Create(*this);}
inline I_Observer::I_Observer(I_ObserverCreator &creator) {creator.Create(*this);}


class ObserveManager
{
public:
	void ObservePhase(const ObserveManager &enemy)
	{
		ObserveParam op(enemy.Observed.const_element_list(),Observed.const_element_list());
		for (ObserverList::iterator it = Observer.begin();it != Observer.end();++it)
			it->Observe(op);
	}
	void RealizeCreature(void)
	{
		Observed.splice(Observed.end(),dCreator.Creature);
		Observer.splice(Observer.end(),rCreator.Creature);
	}

private:
	ObservedList Observed;
	ObserverList Observer;

public:
	I_ObservedCreator dCreator;
	I_ObserverCreator rCreator;
};



#endif
