
#ifndef Game_Obstruct_H
#define Game_Obstruct_H

#include "Juna/intrusive_list.hpp"
#include "TypeDef.h"


class I_Obstructed;
class I_Obstructor;

class ObstructParam;

class I_ObstructedCreator;
class I_ObstructorCreator;

class ObstructManager;

typedef Juna::intrusive_list<I_Obstructed> ObstructedList;
typedef Juna::intrusive_list<I_Obstructor> ObstructorList;


class I_Obstructed : public Juna::intrusive_list_node
{
public:
	struct Data
	{
		Point position;
		Vector velocity;
		float radius;

		Data(const Point &p,const Vector &v,float r) : position(p),velocity(v),radius(r) {}
	};
	virtual Data obstructedData(void) const = 0;

private: friend class I_Obstructor;
	virtual void Pressured(const Vector &vector,const I_Obstructor &obstructor) = 0;

protected:
	I_Obstructed(I_ObstructedCreator &creator);
	~I_Obstructed() {unlink();}
private:
	using Juna::intrusive_list_node::unlink;
};

class I_Obstructor : public Juna::intrusive_list_node
{
private: friend class ObstructManager;
	virtual void Obstruct(const ObstructParam &param) = 0;

protected:
	void Pressure(I_Obstructed &target,const Vector &vec) {target.Pressured(vec,*this);}

protected:
	I_Obstructor(I_ObstructorCreator &creator);
	~I_Obstructor() {unlink();}
private:
	using Juna::intrusive_list_node::unlink;
};

class ObstructParam
{
public:
	typedef ObstructedList list;

	const list &Enemy;
	const list &Friend;

	ObstructParam(const list &enemy,const list &friend_)
		: Enemy(enemy),Friend(friend_) {}
private:
	void operator=(ObstructParam &);
};


class I_ObstructedCreator
{
private: friend class I_Obstructed;
	void Create(I_Obstructed &obstructed) {Creature.push_back(obstructed);}
private: friend class ObstructManager;
	ObstructedList Creature;
};
class I_ObstructorCreator
{
private: friend class I_Obstructor;
	void Create(I_Obstructor &obstructor) {Creature.push_back(obstructor);}
private: friend class ObstructManager;
	ObstructorList Creature;
};

inline I_Obstructed::I_Obstructed(I_ObstructedCreator &creator) {creator.Create(*this);}
inline I_Obstructor::I_Obstructor(I_ObstructorCreator &creator) {creator.Create(*this);}



class ObstructManager
{
public:
	void ObstructPhase(const ObstructManager &enemy)
	{
		ObstructParam op(enemy.Obstructed,Obstructed);
		for (ObstructorList::iterator it = Obstructor.begin();it != Obstructor.end();++it)
			it->Obstruct(op);
	}
	void RealizeCreature(void)
	{
		Obstructed.splice(Obstructed.end(),dCreator.Creature);
		Obstructor.splice(Obstructor.end(),rCreator.Creature);
	}
	const ObstructedList &obstructed(void) const {return Obstructed;}
	const ObstructorList &obstructor(void) const {return Obstructor;}

private:
	ObstructedList Obstructed;
	ObstructorList Obstructor;
public:
	I_ObstructedCreator dCreator;
	I_ObstructorCreator rCreator;
};

#endif
