
#include "Player.h"
#include "VisualizeParam.h"
#include "Sounder.h"

#include "Shot.h"
#include "XLaser.h"
#include "NWayShooter.h"

#include "GameDefine.h"

#include <math.h>

void PlayerController::UnitControl(I_Unit &controlled,const ObserveParam &)
{
	int x = right - left,y = down - up;
	if (x | y)
	{
		RadianAngle table[3][3] =
		{
			{Pi + Pi/4,Pi + Pi/2,Pi2 - Pi/4},
			{Pi,0,0},
			{Pi - Pi/4,Pi / 2,Pi / 4}
		};
		controlled.Move(table[y + 1][x + 1]);
//			controlled.Turn(table[y + 1][x + 1]);
	}
	for (int i = 0;i < 8;i++)
		if (act[i])
			controlled.Act(i);
}

Player::Player(I_UnitController &controller,
			   ::Generator &generator,
			   const Point &position,RadianAngle angle)
  : I_Entity(generator.Entity), I_Visual(generator.Visual,VL_Lower),
	I_Observed(generator.Observed), I_Observer(generator.Observer),
	I_Obstructed(generator.Obstructed),
	I_Attacked(generator.Attacked), I_Attacker(generator.Attacker),
	Position(position), Direction(angle), Velocity(0,0), AngleV(0),
	Pressure(0,0), Radius(4),
	ShotInterval(0),
	Controller(controller),
	Generator(generator)
{
	MachineGun.BootUp(generator);
}

Player *Player::Create(I_UnitController &controller,::Generator &generator,const Point &position,RadianAngle angle)
{
	return new Player(controller,generator,position,angle);
}

void Player::Break(void)
{
	delete this;
}


bool Player::Update(void)
{
	Position += Velocity;
	Position += Pressure;
	Pressure = Vector(0,0);
	Direction += AngleV;
	Direction.SetNormal();
	return true;
}
void Player::Visualize(const VisualizeParam &param) const
{
	param.Screen.Blit(Position.x,Position.y,
		param.Sprite,Rect(0,0,Size(64,64)),Direction.radian());
}

void Player::Observe(const ObserveParam &param)
{
	Velocity.Set(0,0);
	AngleV = 0;
	MG_Trigger = false;
	if (ShotInterval)
		ShotInterval--;
	Controller.UnitControl(*this,param);
	Point shot_pos = Position + Vector(Direction.cos() * 24,Direction.sin() * 24);
	MachineGun.Active(MG_Trigger,shot_pos,Direction);

}
void Player::Pressured(const Vector &vector,const I_Obstructor &obstructor)
{
	Pressure += vector;
}

void Player::Damaged(uint32 power,const Vector &vector,const I_Attacker &attacker)
{
	Velocity += vector;
	Sounder::instance.PlaySE2();
}

void Player::Attack(const AttackParam &param)
{
}

void Player::Move(RadianAngle direction)
{
	const float Speed = 4.f;
	Velocity = Vector(direction.cos() * Speed,direction.sin() * Speed);
}
void Player::Turn(RadianAngle direction)
{
	AngleV = direction - Direction;
}
void Player::Act(unsigned int command)
{
	struct Shooter : I_Shooter
	{
		virtual void Shoot(::Generator &generator,const Point &pos,RadianAngle dir)
		{
			Shot::Create(generator,pos,dir.radian(),Vector(dir.cos() * Speed,dir.sin() * Speed));
		}
		virtual float SetSpeed(float speed) {return Speed = speed;}
		float Speed;
	};
	Shooter shooter;
	shooter.Speed = 8;
	Point shot_pos = Position + Vector(Direction.cos() * 24,Direction.sin() * 24);
	switch (command)
	{
	case 0:

		if (!ShotInterval)
		{
//			PiercingShot::Create(Generator,shot_pos,Direction.radian(),
//						 Vector(Direction.cos() * 8,Direction.sin() * 8));
			OddParallelShooter::Shoot(Generator,shot_pos,Direction,
								 shooter,2,16);
			ShotInterval = 4;
		}
//		XLaser::Create(Generator,shot_pos,Direction.radian(),16);
		break;
	case 1:
		LaserShot::Create(Generator,shot_pos,Direction.radian());
		break;
	case 2:
		MG_Trigger = true;
//		LaserGun.Shoot(Generator,shot_pos,Direction.radian(),12);
		break;
	case 3:
		if (!ShotInterval)
		{
			OddWayShooter::Shoot(Generator,shot_pos,Direction,
								 shooter,2,Pi/8);
			ShotInterval = 4;
		}
		break;
	}

}
