

#ifndef Game_Player_H
#define Game_Player_H


#include "Entity.h"
#include "Observe.h"
#include "Obstruct.h"
#include "Attack.h"
#include "Visual.h"
#include "Unit.h"

#include "Generator.h"

#include "XLaser.h"
#include "MachineGun.h"

class PlayerController : public I_UnitController
{
public:
	virtual void UnitControl(I_Unit &controlled,const ObserveParam &);

	bool up,down,left,right;
	bool act[8];
};


class Player : public I_Entity, public I_Visual,
			   public I_Observed, public I_Observer,
			   public I_Obstructed,
			   public I_Attacked, public I_Attacker,
			   public I_Unit,
			   public Juna::D2::Intersection::I_Circle
{
	Player(I_UnitController &controller,Generator &generator,const Point &pos,RadianAngle dir);
	virtual ~Player() {}
public:
	static Player *Create(I_UnitController &controller,Generator &generator,const Point &pos,RadianAngle dir);

	virtual bool Update(void);
	virtual void Break(void);
	virtual void Visualize(const VisualizeParam &param) const;

	virtual const Point &position(void) const {return Position;}
	virtual RadianAngle direction(void) const {return Direction;}
	virtual const Vector &velocity(void) const {return Velocity;}

	virtual void Observe(const ObserveParam &param);

	virtual const I_Circle &obstructedShape(void) const {return *this;}

	virtual I_Obstructed::Data obstructedData(void) const
	{
		return I_Obstructed::Data(Position,Velocity,Radius);
	}
	virtual void Pressured(const Vector &vector,const I_Obstructor &obstructor);

	virtual const I_Shape &attackedShape(void) const {return *this;}

	virtual void Damaged(uint32 power,const Vector &vector,
						 const I_Attacker &attacker);

	virtual void Attack(const AttackParam &param);


	virtual void Move(RadianAngle direction);
	virtual void Turn(RadianAngle direction);
	virtual void Act(unsigned int command);


	virtual I_Circle::Data GetCircleData(void) const
	{
		return I_Circle::Data(Position,Radius);
	}


private:
	Point Position;
	RadianAngle Direction;
	Vector Velocity;
	RadianAngle AngleV;
	Vector Pressure;

	float Radius;

	int ShotInterval;
	XLaserGun LaserGun;
	MachineGun MachineGun;
	bool MG_Trigger;

	I_UnitController &Controller;
	Generator &Generator;
};

#endif
