
#include <math.h>

#include "ShootingScene.h"

#include "VisualizeParam.h"

#include "Player.h"
#include "Shot.h"
#include "Enemy.h"

#include "Wall.h"

#include "GameDefine.h"

#include "MersenneTwister.h"

#include "BulletMLObject.h"


bool ShootingScene::Initialize(
	Juna::D2D::I_ScreenSurface &screen,
	Juna::D2D::I_SpriteSurface &sprite,
	Juna::D2D::I_SpriteSurface &ascii,
	Juna::D2D::I_ScreenSurface &bg,
	const Juna::Input::I_GameController &input)
{
	if (!screen.isValid() || !sprite.isValid() || !ascii.isValid())
		return false;

	Screen = &screen;
	Sprite = &sprite;
	Ascii = &ascii;
	BG = &bg;
	Input = &input;

	Player::Create(PCInput,ForceA.Generator,Point(320,400),Pi2 / 4 * 3);
	Wall::Create(ForceB.Generator);
	return true;
}
void ShootingScene::Terminalize(void)
{
	ClearEntities();
}

void DrawASCIIString(Juna::D2D::I_ScreenSurface &screen,Juna::D2D::I_SpriteSurface &sprite,
					 int x,int y,const char *string);
void DrawNumber(Juna::D2D::I_ScreenSurface &screen,Juna::D2D::I_SpriteSurface &sprite,
				int x,int y,unsigned num,unsigned figure = 4,bool fillzero = false);

I_Scene *ShootingScene::Action(void)
{
	PCInput.left = Input->left();
	PCInput.right = Input->right();
	PCInput.up = Input->up();
	PCInput.down = Input->down();
	for (int i = 0;i < 8;i++)
		PCInput.act[i] = Input->buttonState(i);

	static float angle;
//	Enemy::Create(ForceB.Generator,120,Point(::cos(angle) * 320 + 320,120),
//				  angle,Vector(::cos(angle),::sin(angle)));
	static int frame = -1;
	static MT::Generator rand_gen;
	frame++;
	if ((frame & 0x3F) == 0 && ForceB.Attack.attacked().size() < 400)
	{
		for (int i = 0;i < 1;i++)
		{
			float angle = rand_gen.GenerateReal1() * Pi2;
			Enemy::Create(ForceB.Generator,120,Point(StgViewLeft + rand_gen.GenerateReal1() * StgViewWidth,1),
				  angle,Vector(::cos(angle),::sin(angle)));
		}
	}


	if ((frame & 0xFF) == 0)
	{
		BulletMLManager &bml = BulletMLManager::GetInstance();
		if (!bml.GetParsers().empty() && bml.objectCount() < 3)
		{
			static int i;
			bml.Create(bml.GetParsers()[i],ForceB.Generator,
					   Point(StgViewLeft + rand_gen.GenerateReal1() * StgViewWidth,120),Pi/2);
			i = (i + 1) % bml.GetParsers().size();
		}
	}
//*/

	ForceA.Observe.ObservePhase(ForceB.Observe);
	ForceB.Observe.ObservePhase(ForceA.Observe);
	RealizeCreature();

	ForceA.Attack.AttackPhase(ForceB.Attack);
	ForceB.Attack.AttackPhase(ForceA.Attack);

	ForceA.Obstruct.ObstructPhase(ForceB.Obstruct);
	ForceB.Obstruct.ObstructPhase(ForceA.Obstruct);

	Screen->Copy(StgViewLeft,StgViewTop,*BG,0);
	Visuals.VisualizePhase(VisualizeParam(*Screen,*Sprite));

	Entities.UpdatePhase();

	RealizeCreature();


	DrawNumber(*Screen,*Ascii,StgViewRight + 16, 0,ForceB.Attack.attacked().size());
	DrawNumber(*Screen,*Ascii,StgViewRight + 16,16,ForceB.Attack.attacker().size());
	DrawNumber(*Screen,*Ascii,StgViewRight + 16,48,ForceA.Attack.attacker().size());

	return this;
}


void DrawASCIIString(Juna::D2D::I_ScreenSurface &screen,Juna::D2D::I_SpriteSurface &sprite,
					 int x,int y,const char *string)
{
	using namespace Juna;
	for (const char *c = string;*c != '\0';c++)
	{
		D2D::Rect crect((*c & 15) * 16,((*c - ' ') / 16) * 16,D2D::Size(16,16));
		screen.Blit(x + (c - string) * 16,y,sprite,crect);
	}
}

void DrawNumber(Juna::D2D::I_ScreenSurface &screen,Juna::D2D::I_SpriteSurface &sprite,
				int x,int y,unsigned num,unsigned figure,bool fillzero)
{
	using namespace Juna;
	const int div_table[9] = {1,10,100,1000,10000,100000,1000000,10000000,100000000};
	int div = div_table[figure - 1];

	unsigned i = 0;
	if (!fillzero)
	{
		for (;i < figure - 1;i++)
		{
			if ((num / div) % 10 != 0)
				break;
			div /= 10;
		}
	}
	for (;i < figure;i++)
	{
		int n = (num / div) % 10;
		div /= 10;
		D2D::Rect crect(n * 16,16,D2D::Size(16,16));
		screen.Blit(x + i * 16,y,sprite,crect);
	}
}
