
#ifndef Game_Visual_Object_H
#define Game_Visual_Object_H

#include "Juna/intrusive_list.hpp"

class I_Visual;

class I_VisualCreator;

class VisualManager;

typedef Juna::intrusive_list<I_Visual> VisualList;

enum VisualLayer
{
	VL_Bottom,
	VL_Lower,
	VL_Upper,
	VL_Top,
};

class VisualizeParam;

class I_Visual : public Juna::intrusive_list_node
{
private: friend class VisualManager;
	virtual void Visualize(const VisualizeParam &param) const = 0;

protected:
	I_Visual(I_VisualCreator &creater,VisualLayer z);
	~I_Visual() {unlink();}
};

class I_VisualCreator
{
private: friend class I_Visual;
	virtual void Create(I_Visual &visual,VisualLayer z) = 0;
};

inline I_Visual::I_Visual(I_VisualCreator &creator,VisualLayer z) {creator.Create(*this,z);}


class VisualManager : public I_VisualCreator
{
public:
	void VisualizePhase(const VisualizeParam &param)
	{
		for (VisualList::iterator it = Bottom.begin();it != Bottom.end();++it)
			it->Visualize(param);
		for (VisualList::iterator it = Lower.begin();it != Lower.end();++it)
			it->Visualize(param);
		for (VisualList::iterator it = Upper.begin();it != Upper.end();++it)
			it->Visualize(param);
		for (VisualList::iterator it = Top.begin();it != Top.end();++it)
			it->Visualize(param);
	}

	VisualManager(void) {}
private: friend class I_VisualCreateor;
	VisualList Bottom;
	VisualList Lower;
	VisualList Upper;
	VisualList Top;
private:
	void Create(I_Visual &visual,VisualLayer z)
	{
		switch (z)
		{
		case VL_Bottom:
			Bottom.push_back(visual); break;
		case VL_Lower:
			Lower.push_back(visual); break;
		case VL_Upper:
			Upper.push_back(visual); break;
		case VL_Top:
			Top.push_back(visual); break;
		default:
			break;
		}
		return;
	}
};


#endif
