
#include "XLaser.h"

#include "Allocator.h"
#include "VisualizeParam.h"
#include "GameDefine.h"

#include <new>

XLaser::XLaser(Generator &generator,
			   const Point &pos,RadianAngle dir,float speed)
  : I_Entity(generator.Entity),
	I_Visual(generator.Visual,VL_Top),
	I_Observer(generator.Observer),
	I_Attacker(generator.Attacker),
	Position(pos), Direction(dir), Velocity(Vector(dir.cos() * speed,dir.sin() * speed)),
	Speed(speed), Range(4), Power(32), LifeCounter(60)
{
}

XLaser *XLaserGun::Shoot(Generator &generator, const Point &pos, RadianAngle dir, float speed)
{
	I_Allocator &a = generator.Allocator;
	assert(a.blockSize() >= sizeof(XLaser) + sizeof(I_Allocator *));
	if (a.full())
		return 0;
	I_Allocator **holder = static_cast<I_Allocator **>(a.Allocate());
	*holder = &a;
	XLaser *nl = new(holder + 1) XLaser(generator,pos,dir,speed);
	if (last.get())
	{
		XLaser *ll = last.get();
		last.unlink();
		nl->prev.link(*ll);
	}
	last.link(*nl);
	return nl;
}

void XLaser::Break(void)
{
	I_Allocator **holder = reinterpret_cast<I_Allocator **>(this) - 1;
	this->~XLaser();
	(*holder)->Deallocate(holder);
}

bool XLaser::Update(void)
{
	LifeCounter--;
	Velocity.Set(Direction.cos() * Speed,Direction.sin() * Speed);
	Position += Velocity;
	return LifeCounter > 0;
}

void XLaser::Observe(const ObserveParam &param)
{
	typedef ObserveParam::list list;
	const list &enemy = param.Enemy;
	if (!enemy.empty())
	{
		list::const_iterator target = enemy.begin();
		float distance = (target->position() - Position).Power2();
		list::const_iterator it = target;
		for (++it;it != enemy.end();++it)
		{
			float d = (it->position() - Position).Power2();
			if (d < distance)
			{
				target = it;
				distance = d;
			}
		}
//		if (distance < 256 * 256)
		{
			Vector v = target->position() - Position;
			Direction.SetAtan(v.y,v.x);
		}
	}
	if (prev.get())
	{
		XLaser &p = *prev.get();
		Vector v = p.position() - position();
		RadianAngle dir = ::atan2(v.y,v.x);
		dir -= Direction;
		dir.SetNormal2();
		Direction += dir / 2;
		Direction.SetNormal();
	}

	
}

void XLaser::Attack(const AttackParam &param)
{
	struct Circle : public Juna::D2::Intersection::I_Circle
	{
		Circle(const XLaser &laser) : laser(laser) {}
		virtual Data GetCircleData(void) const
		{
			const Point &pos = laser.position();
			return Data(pos,laser.range());
		}
		const XLaser &laser;
	};
	Circle c(*this);

	typedef AttackParam::list list;
	const list &enemy = param.Enemy;
	for (list::const_iterator it = enemy.begin();it != enemy.end();++it)
	{
		if (it->attackedShape().Test(c))
		{
			Damage(*it,Power,Vector(0,0));
			LifeCounter = 0;
		}
	}
}
void XLaser::Visualize(const VisualizeParam &param) const
{
	param.Screen.Blit(Position.x,Position.y,
		param.Sprite,Rect(176,0,Size(16,16)),Direction.radian());
}


